import java.io.File;
import java.util.Scanner;
import java.io.FileWriter;
import java.text.SimpleDateFormat;
import java.util.Date;

public class TopoGenerator {

	static int threads = 1; //
	static int nodes = 8;
	static int iterations = -1;
	static int edges = 0;
	static double edgeProb = 0.5;

	static String selectionMode = "first"; // first, uniform
	static String filename;
	static FileWriter fw;

	static boolean recordRuntimes = false;
	static int j = 0;

	public static void main(String[] args) {

		Scanner scan = new Scanner(System.in); // Create a Scanner object
		System.out.print("Which task should be performed? [1 = Topology Generation, 2 = Comparison]");

		int task = Integer.parseInt(scan.nextLine());

		switch (task) {
		case 1:
			String[] param = new String[6];
			String temp = "";

			System.out.print("Threads = ");
			param[0] = "-t" + scan.nextLine();

			System.out.print("Nodes = ");
			param[1] = "-n" + scan.nextLine();

			System.out.print("Iterations (Default All) = ");
			temp = scan.nextLine();
			param[2] = "-i" + (temp.length() == 0 ? "-1" : temp);
			
			System.out.print("Edges (Default Random) = ");
			temp = scan.nextLine();
			param[3] = "-e" + (temp.length() == 0 ? "0" : temp);
			
			System.out.print("EdgeSelectionMode [first (default), uniform] = ");
			temp = scan.nextLine();
			param[4] = "-sm" + (temp.length() == 0 ? "first" : temp);
			
			System.out.print("EdgeProbability [only for Iterations>0, Default 0.5] = ");
			temp = scan.nextLine();
			param[5] = "-p" + (temp.length() == 0 ? "0.5" : temp);

			runExp(param);
			break;
		case 2:
			for (int i = 4; i <= 30; i++) {
				for (j = 0; j < 100; j++) {
					String[] parm = { "-n" + i };
					runComparison(parm);
				}

			}
			break;
			
		case 3: 
			String [] p = new String[4];
			p[0]="-t32";
			p[1]="-i15";
			p[2]="-p0.3";
			for(int i = 9; i<=50; i++) {
				p[3] = "-n" + i;
				runExp(p);
			}
			break;
		default:
			break;
		}
	}

	public static void runComparison(String[] args) {
		for (String s : args) {
			if (s.startsWith("-t")) {
				threads = Integer.parseInt(s.substring(2));
			} else if (s.startsWith("-n")) {
				nodes = Integer.parseInt(s.substring(2));
			} else if (s.startsWith("-i")) {
				iterations = Integer.parseInt(s.substring(2));
			} else if (s.startsWith("-p")) {
				selectionMode = s.substring(2);
			} else if (s.startsWith("-e")) {
				edges = Integer.parseInt(s.substring(2));
			} else if (s.startsWith("-sm")) {
				selectionMode = s.substring(3);
			}
		}

		filename = "N" + String.format("%03d", nodes) + (iterations >= 0 ? "R" : "C") + (edges > 0 ? "_E" + edges : "")
				+ "__" + new SimpleDateFormat("yyyy-MM-dd_HH-mm-ss").format(new Date()) + "__Deterministisch";

		TopologyIteratorWrapper topo = new TopologyIteratorWrapper(nodes);

		topo.nextRandom();
		topo.generateTopology();
		while (!topo.proveFaban("first")) {
			topo.nextRandom();
			topo.generateTopology();
		}

		topo.writeStatisticsToFile(filename + "_T" + (1));

		filename = "N" + String.format("%03d", nodes) + (iterations >= 0 ? "R" : "C") + (edges > 0 ? "_E" + edges : "")
				+ "__" + new SimpleDateFormat("yyyy-MM-dd_HH-mm-ss").format(new Date()) + "__Zufall";
		topo.clearStatistics();

		System.out.print(String.format("[%03d/100] Progress: |", j + 1));
		int p = 0;
		for (int i = 0; i < 100000; i++) {
			topo.generateTopology();
			topo.proveFaban("uniform");
			if ((i + 1) % 1000 == 0)
				System.out.print("=");
		}
		System.out.println("| Finished n=" + nodes);

		topo.writeStatisticsToFile(filename + "_T" + (1));

	}

	public static void runExp(String[] args) {
		for (String s : args) {
			if (s.startsWith("-t")) {
				threads = Integer.parseInt(s.substring(2));
			} else if (s.startsWith("-n")) {
				nodes = Integer.parseInt(s.substring(2));
			} else if (s.startsWith("-i")) {
				iterations = Integer.parseInt(s.substring(2));
			} else if (s.startsWith("-e")) {
				edges = Integer.parseInt(s.substring(2));
			} else if (s.startsWith("-sm")) {
				selectionMode = s.substring(3);
			}
		}

		filename = "N" + String.format("%03d", nodes) + (iterations >= 0 ? "R" : "C") + (edges > 0 ? "_E" + edges : "")
				+ "__" + new SimpleDateFormat("yyyy-MM-dd_HH-mm-ss").format(new Date()) + "__AlgMode" + selectionMode;

		System.out.println("Starting with the following parameters: " + nodes + " nodes and " + threads + " threads\n");
		System.out.println("\t" + "Threads: \t" + threads);
		System.out.println("\t" + "Nodes: \t\t" + nodes);
		System.out.println("\t" + "Iterations: \t" + (iterations < 0 ? "All" : "2^" + iterations));
		System.out.println(
				"\t" + "Edges: \t\t" + (edges >= nodes && edges <= nodes * (nodes - 1) * 0.5 ? "=" + edges : "random"));
		System.out.println("\t" + "SelectionMode: \t" + selectionMode);
		System.out.println("");

		String[] s = new String[threads];
		int last = (int) (Math.log(threads) / Math.log(2));
		boolean[] arr = new boolean[last];
		for (int i = 0; i < s.length; i++) {
			for (int j = last - 1; j >= 0; j--) {
				if (i == 0) {
					arr[j] = false;
				} else {
					if (!arr[j]) {
						arr[j] = true;
						break;
					} else
						arr[j] = false;
				}
			}
			s[i] = "";
			for (boolean b : arr) {
				s[i] += (b ? 1 : 0);
			}
		}

		long startTime = System.currentTimeMillis();

		Thread[] thr = new Thread[threads];

		try {
			fw = new FileWriter(new File(filename + "_Summary.csv"));
			fw.write("Nodes\t" + "Max. Edges\t" + "ValidTopologies(abs) [x]\t" + "ValidTopologies(abs) [x*2^31]\t"
					+ "ValidTopologies(rel)\t" + "TotalTopologiesChecked [2^x]\n");
		} catch (Exception e) {
			e.printStackTrace();
		}

		for (int i = 0; i < threads; i++) {
			final int j = i;

			thr[i] = new Thread(new Runnable() {

				@Override
				public void run() {
					Thread.currentThread().setName("T" + j);
					System.out.println("Thread " + j + " started!");
					TopologyIteratorWrapper topo;

					switch (threads) {
					case 1:
						topo = new TopologyIteratorWrapper(nodes);
						break;

					default:
						topo = new TopologyIteratorWrapper(nodes, s[j]);
						break;

					}

					// Iterate all combinations
					if (iterations < 0) {
						while (!topo.end()) {
							topo.generateTopology();
							topo.proveFaban(selectionMode);
							topo.next();
						}

						topo.writeStatisticsToFile(filename + "_T" + (j + 1));
						synchronized (fw) {
							try {
								fw.write(topo.getNodes() + "\t");
								fw.write(topo.getEdges() + "\t");
								fw.write(topo.getNoOfValidTopologies() + "\t");
								fw.write((int) topo.getNoOfValidTopologies_2pow31() + "\t");
								fw.write(topo.getValidTopoPercentage() + "\t");
								fw.write(topo.getLnNumberOfIterations() + "\n");
							} catch (Exception e) {
								e.printStackTrace();
							}
						}

						System.out.println("Thread " + j + " finished! [" + topo.getValidTopoPercentage() + "]");
					}
					// Random iteration with 2^iterations repititions
					else {
						topo.setLnNumberOfIterations(iterations);
						int mod = (int) Math.pow(2, Math.max(0, iterations - 7));
						for (int i = 0, j = 0; i < Math.pow(2, iterations); i++) {
							if (edges >= nodes && edges <= nodes * (nodes - 1) * 0.5)
								topo.nextRandom(edges);
							else
								topo.nextRandom();
							topo.generateTopology();
							topo.proveFaban(selectionMode);
							j = (j + 1) % mod;
							if (j == 0) {
								System.out.println("[" + Thread.currentThread().getName() + "] "
										+ (int) (100 * (i / Math.pow(2, iterations))) + "%");
							}
						}

						topo.writeStatisticsToFile(filename + "_T" + (j + 1));
						synchronized (fw) {
							try {
								fw.write(topo.getNodes() + "\t");
								fw.write(topo.getEdges() + "\t");
								fw.write(topo.getNoOfValidTopologies() + "\t");
								fw.write((int) topo.getNoOfValidTopologies_2pow31() + "\t");
								fw.write(topo.getValidTopoPercentage() + "\t");
								fw.write(iterations + "\n");
							} catch (Exception e) {
								e.printStackTrace();
							}
						}

						System.out.println("Thread " + j + " finished! [" + topo.getValidTopoPercentage() + "]");
					}

				}
			});
			thr[i].start();
		}

		for (Thread t : thr) {
			try {
				t.join();
			} catch (Exception e) {
			}
		}

		try {
			fw.write("\nMode: " + (iterations <= 0 ? "Complete" : "Random") + "\n");
			fw.write("Elapsed Time: " + (System.currentTimeMillis() - startTime) / 1000.0 + "s");
			fw.close();
		} catch (Exception e) {
			e.printStackTrace();
		}

		System.out.println("\nFinished at " + (System.currentTimeMillis() - startTime) / 1000.0 + "s\n");
	}

}
