import topology.FabanWaveTool;

public class CliMain {

	public static void main(String[] args) {
		String inputfile = "";
		String outputfile = "";
		String command = "";
		boolean override = false;

		// Reading arguments
		for (int i = 0; i < args.length; i++) {
			switch (args[i]) {
			case "-i":
				if (i + 1 >= args.length || args[i + 1].startsWith("-")) {
					System.out.println("No input path specified");
					return;
				}
				inputfile = args[i + 1];
				i++;
				break;
			case "-o":
				if (i + 1 >= args.length || args[i + 1].startsWith("-")) {
					System.out.println("No output path specified");
					return;
				}
				outputfile = args[i + 1];
				i++;
				break;
			case "-O":
				override = true;
				break;
			case "-create":
				command = "create";
				break;
			case "-validate":
				command = "validate";
				break;
			case "-help":
				command = "help";
				break;
			default:
				System.out.println(
						"Unknown argument: \"" + args[i] + "\". Use \"-help\" as argument for more information.");
				return;
			}
		}

		FabanWaveTool tool = new FabanWaveTool();

		switch (command) {
		case "create":
			if ((outputfile.isEmpty() || outputfile.equals(inputfile)) && !override) {
				System.out.println(
						"Path for output is equal to the input path. Specify a different path or set the override flag -O to override the input file.");
				return;
			}
			if (tool.loadJsonFile(inputfile) && tool.parseTopology() && tool.generateWaves()) {
				if (outputfile.length() > 0) {
					tool.saveJsonFile(outputfile);
				} else {
					tool.saveJsonFile(inputfile);
				}
			}
			break;
			
		case "validate":
			if (tool.loadJsonFile(inputfile) && tool.parseTopology() && tool.parseFabanConfiguration()) {
				tool.validateFabanConfiguration();
			}
			break;

		case "help":
			System.out.println("Option \t\tMeaning\n");
			System.out.println("-i <path> \tSpecifies the path of the input json file.");
			System.out.println("-o <path> \tSpecifies the path of the output json file. (optional)");
			System.out.println("-create \tReads the topology information stored in the input "
					+ "json file\n\t\tand creates a valid Faban Configuration.");
			System.out.println("-validate \tReads the topology and Faban Configuration information "
					+ "stored\n\t\tin the input json file and validates the correctness of the last.");
			System.out.println("-O \t\tOverrides the inputfile if no outputfile was specified.");
			System.out.println("-help \t\tShows this output.");
			break;

		default:
			System.out.println("Could not detect a command");
			break;
		}
	}

}
