package configgui.fi_jn;

import java.awt.Button;
import java.io.File;
import java.io.FileReader;
import java.io.PrintWriter;
import java.nio.file.Paths;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Optional;

import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonParser;

import configgui.java.Rcc;
import configgui.java.commands.Command;
import configgui.java.commands.simple.SendRemoteCommand;
import configgui.java.model.configuration.BananaInformation;
import javafx.collections.FXCollections;
import javafx.collections.ListChangeListener;
import javafx.collections.ObservableList;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.scene.control.Alert;
import javafx.scene.control.Alert.AlertType;
import javafx.scene.control.ButtonType;
import javafx.scene.control.ChoiceBox;
import javafx.scene.control.ContentDisplay;
import javafx.scene.control.SelectionMode;
import javafx.scene.control.TableCell;
import javafx.scene.control.TableColumn;
import javafx.scene.control.TableColumn.CellEditEvent;
import javafx.scene.control.TableView;
import javafx.scene.control.TextField;
import javafx.scene.control.cell.PropertyValueFactory;
import javafx.scene.control.cell.TextFieldTableCell;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyCodeCombination;
import javafx.scene.input.KeyEvent;
import javafx.scene.paint.Color;
import javafx.stage.FileChooser;
import javafx.stage.FileChooser.ExtensionFilter;
import javafx.util.Callback;
import javafx.util.converter.DoubleStringConverter;
import javafx.util.converter.IntegerStringConverter;

public class FaultInjectionController {

	JsonObject current;

	@FXML
	ChoiceBox<String> cbProtocolList;

	@FXML
	Button button_jn;

	@FXML
	Button button_save;
	@FXML
	Button button_load;

	@FXML
	TableView<FaultInjectionTableEntry> table;

	@FXML
	TableColumn<FaultInjectionTableEntry, String> BananaIPColumn;

	@FXML
	TableColumn<FaultInjectionTableEntry, Double> BFPercentageColumn;
	@FXML
	TableColumn<FaultInjectionTableEntry, String> BFDistributionColumn;

	@FXML
	TableColumn<FaultInjectionTableEntry, Double> WFPercentageColumn;
	@FXML
	TableColumn<FaultInjectionTableEntry, String> WFPortSet;

	@FXML
	TableColumn<FaultInjectionTableEntry, Double> DuplicationPercentageColumn;
	@FXML
	TableColumn<FaultInjectionTableEntry, String> DuplicationDistributionColumn;

	@FXML
	TableColumn<FaultInjectionTableEntry, Double> OmissionPercentageColumn;

	@FXML
	TableColumn<FaultInjectionTableEntry, Double> CrashPercentageColumn;

	@FXML
	TableColumn<FaultInjectionTableEntry, Double> DelayPercentageColumn;
	@FXML
	TableColumn<FaultInjectionTableEntry, String> DelayDistributionColumn;

	@FXML
	TableColumn<FaultInjectionTableEntry, Boolean> ByzantineBehaviourColumn;

	String regexUniformDist = "U\\((\\d*)(\\p{Blank})*,(\\p{Blank})*(\\d*)\\)";
	String regexNormalDist = "N\\((\\d+)(\\.\\d+)?(\\p{Blank})*,(\\p{Blank})*(\\d+)(\\.\\d+)?(\\p{Blank})*(,(\\p{Blank})*(\\d+)(\\p{Blank})*,(\\p{Blank})*(\\d+)(\\p{Blank})*)?\\)";
	String regexExpDist = "Exp\\((\\d+)(\\.\\d+)?(\\p{Blank})*,(\\p{Blank})*(\\d+)\\)";

	String regexPortSet = "(\\{1\\})|(\\{(1,)?2\\})|(\\{(1,)?(2,)?3\\})|(\\{(1,)?(2,)?(3,)?4\\})";
	String regex0_100 = ("((^([0-9]{1,2}){1}(\\.[0-9]{1,2})?$)|100|100\\.[0]+)");
	String regexDoubleNumbers = ("(([0-9]+)|[0-9]+\\.[0-9]+)");

	ObservableList<FaultInjectionTableEntry> bananalist = FXCollections.observableArrayList(new FaultInjectionTableEntry("No Node Connected!"));

	Rcc main;

	String fiTargetProtocol = "";

	// Methoden speichern die eingegebenen Werte

	// BF
	public void changeBFPercentageColumnCellEvent(CellEditEvent edditedCell) {
		FaultInjectionTableEntry FI_selected = table.getSelectionModel().getSelectedItem();
		FI_selected.setBFPercentage((double) edditedCell.getNewValue());

	}

	public void changeBFDistributionColumnCellEvent(CellEditEvent edditedCell) {
		FaultInjectionTableEntry FI_selected = table.getSelectionModel().getSelectedItem();
		String input = edditedCell.getNewValue().toString();
		if (input.matches(regexUniformDist)) {
			input = input.substring(1);
			input = input.replace("(", "");
			input = input.replace(")", "");
			input.trim();
			String[] inputarray = input.split("(\\p{Blank})*,(\\p{Blank})*");
			FI_selected.setBFDistribution("discrete_equal_distribution");
			FI_selected.setBFRange(inputarray[0] + "-" + inputarray[1]);
		} else if (input.matches(regexNormalDist)) {
			input = input.substring(1);
			input = input.replace("(", "");
			input = input.replace(")", "");
			input.trim();
			String[] inputarray = input.split("(\\p{Blank})*,(\\p{Blank})*");

			FI_selected.setBFDistribution("Gaussian_distribution");
			FI_selected.setBFMean(Integer.parseInt(inputarray[0]));
			FI_selected.setBFStandardDeviation(Double.parseDouble(inputarray[1]));
			FI_selected.setBFRange(((inputarray.length > 2) ? inputarray[2] : 0) + "-" + ((inputarray.length > 2) ? inputarray[3] : 1234567)); // 1234567 = MAX Value
		} else if (input.matches(regexExpDist)) {
			input = input.substring(3);
			input = input.replace("(", "");
			input = input.replace(")", "");
			input.trim();
			String[] inputarray = input.split("(\\p{Blank})*,(\\p{Blank})*");
			FI_selected.setBFDistribution("Exponential_distribution");
			FI_selected.setBFLambda(Double.parseDouble(inputarray[0]));
			FI_selected.setBFRange("0-" + inputarray[1]);

		}
		FI_selected.setBFDistributionInput(edditedCell.getNewValue().toString());
	}

	// WF
	public void changeWFPercentageColumnCellEvent(CellEditEvent edditedCell) {
		FaultInjectionTableEntry FI_selected = table.getSelectionModel().getSelectedItem();
		FI_selected.setWFPercentage((double) edditedCell.getNewValue());
	}

	public void changeWFPortSetCellEvent(CellEditEvent edditedCell) {

		FaultInjectionTableEntry FI_selected = table.getSelectionModel().getSelectedItem();
		String input = edditedCell.getNewValue().toString();
		if (input.matches(regexPortSet)) {
			input = input.replace("{", "");
			input = input.replace("}", "");
			String[] inputarray = input.split(",");
			int tempResult = 0;
			for (String string : inputarray) {
				tempResult += Math.pow(2, Double.parseDouble(string) - 1);
			}
			FI_selected.setWFPortSetInt(tempResult);
//			FI_selected.setWFPortSet(edditedCell.getNewValue().toString());
		}
		FI_selected.setWFPortSet(edditedCell.getNewValue().toString());
	}

	// Duplication
	public void changeDuplicationPercentageColumnCellEvent(CellEditEvent edditedCell) {
		FaultInjectionTableEntry FI_selected = table.getSelectionModel().getSelectedItem();
		FI_selected.setDuplicationPercentage((double) edditedCell.getNewValue());
	}

	public void changeDuplicationDistributionColumnCellEvent(CellEditEvent edditedCell) {
		FaultInjectionTableEntry FI_selected = table.getSelectionModel().getSelectedItem();
		String input = edditedCell.getNewValue().toString();
		if (input.matches(regexUniformDist)) {
			input = input.substring(1);
			input = input.replace("(", "");
			input = input.replace(")", "");
			input.trim();
			String[] inputarray = input.split("(\\p{Blank})*,(\\p{Blank})*");
			FI_selected.setDuplicationDistribution("discrete_equal_distribution");
			FI_selected.setDuplicationRange(inputarray[0] + "-" + inputarray[1]);
		} else if (input.matches(regexNormalDist)) {
			input = input.substring(1);
			input = input.replace("(", "");
			input = input.replace(")", "");
			input.trim();
			String[] inputarray = input.split("(\\p{Blank})*,(\\p{Blank})*");

			FI_selected.setDuplicationDistribution("Gaussian_distribution");
			FI_selected.setDuplicationMean(Integer.parseInt(inputarray[0]));
			FI_selected.setDuplicationStandardDeviation(Double.parseDouble(inputarray[1]));
			FI_selected.setDuplicationRange(((inputarray.length > 2) ? inputarray[2] : 0) + "-" + ((inputarray.length > 2) ? inputarray[2] : 1234567)); // 1234567 = MAX Value
		} else if (input.matches(regexExpDist)) {
			input = input.substring(3);
			input = input.replace("(", "");
			input = input.replace(")", "");
			input.trim();
			String[] inputarray = input.split("(\\p{Blank})*,(\\p{Blank})*");
			FI_selected.setDuplicationDistribution("Exponential_distribution");
			FI_selected.setDuplicationLambda(Double.parseDouble(inputarray[0]));
			FI_selected.setDuplicationRange("0-" + inputarray[1]);

		}
		FI_selected.setDuplicationDistributionInput(edditedCell.getNewValue().toString());
	}

	// Omission
	public void changeOmissionPercentageColumnCellEvent(CellEditEvent edditedCell) {
		FaultInjectionTableEntry FI_selected = table.getSelectionModel().getSelectedItem();
		FI_selected.setOmissionPercentage((double) edditedCell.getNewValue());
	}

	// Crash
	public void changeCrashPercentageColumnCellEvent(CellEditEvent edditedCell) {
		FaultInjectionTableEntry FI_selected = table.getSelectionModel().getSelectedItem();
		FI_selected.setCrashPercentage((double) edditedCell.getNewValue());
	}

	// Delay
	public void changeDelayPercentageColumnCellEvent(CellEditEvent edditedCell) {
		// FaultInjectionTableEntry FI_selected =
		// table.getSelectionModel().getSelectedItem();
		double temp = (double) edditedCell.getNewValue();
		// FI_selected.setDelayPercentage();

		for (FaultInjectionTableEntry selectedFI : table.getSelectionModel().getSelectedItems()) {
			selectedFI.setDelayPercentage(temp);
		}
	}

	public void changeDelayDistributionColumnCellEvent(CellEditEvent edditedCell) {
		FaultInjectionTableEntry FI_selected = table.getSelectionModel().getSelectedItem();
		String input = edditedCell.getNewValue().toString();
		if (input.matches(regexUniformDist)) {
			input = input.substring(1);
			input = input.replace("(", "");
			input = input.replace(")", "");
			input.trim();
			String[] inputarray = input.split("(\\p{Blank})*,(\\p{Blank})*");
			FI_selected.setDelayDistribution("discrete_equal_distribution");
			FI_selected.setDelayRange(inputarray[0] + "-" + inputarray[1]);
		} else if (input.matches(regexNormalDist)) {
			input = input.substring(1);
			input = input.replace("(", "");
			input = input.replace(")", "");
			input.trim();
			String[] inputarray = input.split("(\\p{Blank})*,(\\p{Blank})*");

			FI_selected.setDelayDistribution("Gaussian_distribution");
			FI_selected.setDelayMean(Integer.parseInt(inputarray[0]));
			FI_selected.setDelayStandardDeviation(Double.parseDouble(inputarray[1]));
			FI_selected.setDelayRange(((inputarray.length > 2) ? inputarray[2] : 0) + "-" + ((inputarray.length > 2) ? inputarray[2] : 1234567)); // 1234567 = MAX Value
		} else if (input.matches(regexExpDist)) {
			input = input.substring(3);
			input = input.replace("(", "");
			input = input.replace(")", "");
			input.trim();
			String[] inputarray = input.split("(\\p{Blank})*,(\\p{Blank})*");
			FI_selected.setDelayDistribution("Exponential_distribution");
			FI_selected.setDelayLambda(Double.parseDouble(inputarray[0]));
			FI_selected.setDelayRange("0-" + inputarray[1]);

		}
		FI_selected.setDelayDistributionInput(edditedCell.getNewValue().toString());
	}

	public void initialize() {

		// Lies Protokollliste aus der Config aus oder erstelle einen neuen leeren
		// Eintrag
		if (Rcc.configuration.getConfigEntry("protocollist") == null) {
			Rcc.configuration.setConfigEntry("protocollist", "");
		} else {
			for (String p : Rcc.configuration.getConfigEntry("protocollist").split(",")) {
				cbProtocolList.getItems().add(p);
			}
			cbProtocolList.getSelectionModel().select(0);
			fiTargetProtocol = cbProtocolList.getSelectionModel().getSelectedItem();
		}
		cbProtocolList.getSelectionModel().selectedIndexProperty().addListener((obj, oldIndex, newIndex) -> {
			fiTargetProtocol = cbProtocolList.getItems().get(newIndex.intValue());
		});

		// sorgt für die Zuordnung der Werte zu der Tabellenzellen
		BananaIPColumn.setCellValueFactory(new PropertyValueFactory<FaultInjectionTableEntry, String>("BananaIP"));

		BFPercentageColumn.setCellValueFactory(new PropertyValueFactory<FaultInjectionTableEntry, Double>("BFPercentage"));
		BFDistributionColumn.setCellValueFactory(new PropertyValueFactory<FaultInjectionTableEntry, String>("BFDistributionInput"));

		WFPercentageColumn.setCellValueFactory(new PropertyValueFactory<FaultInjectionTableEntry, Double>("WFPercentage"));
		WFPortSet.setCellValueFactory(new PropertyValueFactory<FaultInjectionTableEntry, String>("WFPortSet"));

		DuplicationPercentageColumn.setCellValueFactory(new PropertyValueFactory<FaultInjectionTableEntry, Double>("DuplicationPercentage"));
		DuplicationDistributionColumn.setCellValueFactory(new PropertyValueFactory<FaultInjectionTableEntry, String>("DuplicationDistributionInput"));

		OmissionPercentageColumn.setCellValueFactory(new PropertyValueFactory<FaultInjectionTableEntry, Double>("OmissionPercentage"));

		CrashPercentageColumn.setCellValueFactory(new PropertyValueFactory<FaultInjectionTableEntry, Double>("CrashPercentage"));

		DelayPercentageColumn.setCellValueFactory(new PropertyValueFactory<FaultInjectionTableEntry, Double>("DelayPercentage"));
		DelayDistributionColumn.setCellValueFactory(new PropertyValueFactory<FaultInjectionTableEntry, String>("DelayDistributionInput"));

		ByzantineBehaviourColumn.setCellValueFactory(new PropertyValueFactory<FaultInjectionTableEntry, Boolean>("ByzantineBehaviour"));

		table.setItems(bananalist);
		table.getSelectionModel().setSelectionMode(SelectionMode.MULTIPLE);

		// sorgt dafür, dass die Spalten bearbeitet werden können
		table.setEditable(true);

		BFPercentageColumn.setCellFactory(new Callback<TableColumn<FaultInjectionTableEntry, Double>, TableCell<FaultInjectionTableEntry, Double>>() {

			@Override
			public TableCell<FaultInjectionTableEntry, Double> call(TableColumn<FaultInjectionTableEntry, Double> param) {

				TextField tf = new TextField();
				return new TableCell<FaultInjectionTableEntry, Double>() {

					@Override
					protected void updateItem(Double item, boolean empty) {
						super.updateItem(item, empty);
						if (item == null || empty) {
							setText(null);
							setStyle("");
						} else {
							setText(item.toString());
							setTextFill(item.toString().matches(regex0_100) ? Color.BLACK : Color.RED);
						}
					}

					@Override
					public void startEdit() {
						super.startEdit();
						tf.setOnKeyPressed(e -> {
							if (e.getCode().equals(KeyCode.TAB)) {
								if (tf.getText().matches(regexDoubleNumbers)) {
									commitEdit(Double.valueOf(tf.getText()));
								}
								setContentDisplay(ContentDisplay.TEXT_ONLY);
								getTableView().edit(getIndex(), BFDistributionColumn);
							} else if (e.getCode().equals(KeyCode.ENTER)) {
								if (tf.getText().matches(regexDoubleNumbers)) {
									commitEdit(Double.valueOf(tf.getText()));
								}
								setContentDisplay(ContentDisplay.TEXT_ONLY);
							}
						});

						setGraphic(tf);
						setContentDisplay(ContentDisplay.GRAPHIC_ONLY);
						tf.setText(String.valueOf(getItem()));
						tf.requestFocus();
					}

					@Override
					public void cancelEdit() {
						super.cancelEdit();
						setText(String.valueOf(getItem()));
						setContentDisplay(ContentDisplay.TEXT_ONLY);
					}

				};
			}
		});

		BFDistributionColumn.setCellFactory(new Callback<TableColumn<FaultInjectionTableEntry, String>, TableCell<FaultInjectionTableEntry, String>>() {

			@Override
			public TableCell<FaultInjectionTableEntry, String> call(TableColumn<FaultInjectionTableEntry, String> param) {

				TextField tf = new TextField();
				return new TableCell<FaultInjectionTableEntry, String>() {

					@Override
					protected void updateItem(String item, boolean empty) {
						super.updateItem(item, empty);
						if (item == null || empty) {
							setText(null);
							setStyle("");
						} else {
							setText(item);
							setTextFill((item.matches(regexExpDist) || item.matches(regexNormalDist) || item.matches(regexUniformDist)) ? Color.BLACK : Color.RED);
						}
					}

					@Override
					public void startEdit() {
						super.startEdit();
						tf.setOnKeyPressed(e -> {
							if (e.getCode().equals(KeyCode.TAB)) {
								commitEdit(tf.getText());
								setContentDisplay(ContentDisplay.TEXT_ONLY);
								getTableView().edit(getIndex(), WFPercentageColumn);
							} else if (e.getCode().equals(KeyCode.ENTER)) {
								commitEdit(tf.getText());
								setContentDisplay(ContentDisplay.TEXT_ONLY);
							}
						});

						setGraphic(tf);
						setContentDisplay(ContentDisplay.GRAPHIC_ONLY);
						tf.setText(getItem());
						tf.requestFocus();
					}

					@Override
					public void cancelEdit() {
						super.cancelEdit();
						setText(String.valueOf(getItem()));
						setContentDisplay(ContentDisplay.TEXT_ONLY);
					}

				};
			}
		});

		WFPercentageColumn.setCellFactory(new Callback<TableColumn<FaultInjectionTableEntry, Double>, TableCell<FaultInjectionTableEntry, Double>>() {

			@Override
			public TableCell<FaultInjectionTableEntry, Double> call(TableColumn<FaultInjectionTableEntry, Double> param) {
				TextField tf = new TextField();
				return new TableCell<FaultInjectionTableEntry, Double>() {

					@Override
					protected void updateItem(Double item, boolean empty) {
						super.updateItem(item, empty);
						if (item == null || empty) {
							setText(null);
							setStyle("");
						} else {
							setText(item.toString());
							setTextFill(item.toString().matches(regex0_100) ? Color.BLACK : Color.RED);
						}
					}

					@Override
					public void startEdit() {
						super.startEdit();
						tf.setOnKeyPressed(e -> {
							if (e.getCode().equals(KeyCode.TAB)) {
								if (tf.getText().matches(regexDoubleNumbers)) {
									commitEdit(Double.valueOf(tf.getText()));
								}
								setContentDisplay(ContentDisplay.TEXT_ONLY);
								getTableView().edit(getIndex(), WFPortSet);
							} else if (e.getCode().equals(KeyCode.ENTER)) {
								if (tf.getText().matches(regexDoubleNumbers)) {
									commitEdit(Double.valueOf(tf.getText()));
								}
								setContentDisplay(ContentDisplay.TEXT_ONLY);
							}
						});

						setGraphic(tf);
						setContentDisplay(ContentDisplay.GRAPHIC_ONLY);
						tf.setText(String.valueOf(getItem()));
						tf.requestFocus();
					}

					@Override
					public void cancelEdit() {
						super.cancelEdit();
						setText(String.valueOf(getItem()));
						setContentDisplay(ContentDisplay.TEXT_ONLY);
					}

				};
			}
		});

		WFPortSet.setCellFactory(new Callback<TableColumn<FaultInjectionTableEntry, String>, TableCell<FaultInjectionTableEntry, String>>() {

			@Override
			public TableCell<FaultInjectionTableEntry, String> call(TableColumn<FaultInjectionTableEntry, String> param) {

				TextField tf = new TextField();
				return new TableCell<FaultInjectionTableEntry, String>() {

					@Override
					protected void updateItem(String item, boolean empty) {
						super.updateItem(item, empty);
						if (item == null || empty) {
							setText(null);
							setStyle("");
						} else {
							setText(item);
							setTextFill((item.matches(regexPortSet)) ? Color.BLACK : Color.RED);
						}
					}

					@Override
					public void startEdit() {
						super.startEdit();
						tf.setOnKeyPressed(e -> {
							if (e.getCode().equals(KeyCode.TAB)) {
								commitEdit(tf.getText());
								setContentDisplay(ContentDisplay.TEXT_ONLY);
								getTableView().edit(getIndex(), DuplicationPercentageColumn);
							} else if (e.getCode().equals(KeyCode.ENTER)) {
								commitEdit(tf.getText());
								setContentDisplay(ContentDisplay.TEXT_ONLY);
							}
						});

						setGraphic(tf);
						setContentDisplay(ContentDisplay.GRAPHIC_ONLY);
						tf.setText(getItem());
						tf.requestFocus();
					}

					@Override
					public void cancelEdit() {
						super.cancelEdit();
						setText(String.valueOf(getItem()));
						setContentDisplay(ContentDisplay.TEXT_ONLY);
					}

				};
			}
		});

		DuplicationPercentageColumn.setCellFactory(new Callback<TableColumn<FaultInjectionTableEntry, Double>, TableCell<FaultInjectionTableEntry, Double>>() {

			@Override
			public TableCell<FaultInjectionTableEntry, Double> call(TableColumn<FaultInjectionTableEntry, Double> param) {

				TextField tf = new TextField();
				return new TableCell<FaultInjectionTableEntry, Double>() {

					@Override
					protected void updateItem(Double item, boolean empty) {
						super.updateItem(item, empty);
						if (item == null || empty) {
							setText(null);
							setStyle("");
						} else {
							setText(item.toString());
							setTextFill(item.toString().matches(regex0_100) ? Color.BLACK : Color.RED);
						}
					}

					@Override
					public void startEdit() {
						super.startEdit();
						tf.setOnKeyPressed(e -> {
							if (e.getCode().equals(KeyCode.TAB)) {
								if (tf.getText().matches(regexDoubleNumbers)) {
									commitEdit(Double.valueOf(tf.getText()));
								}
								setContentDisplay(ContentDisplay.TEXT_ONLY);
								getTableView().edit(getIndex(), DuplicationDistributionColumn);
							} else if (e.getCode().equals(KeyCode.ENTER)) {
								if (tf.getText().matches(regexDoubleNumbers)) {
									commitEdit(Double.valueOf(tf.getText()));
								}
								setContentDisplay(ContentDisplay.TEXT_ONLY);
							}
						});

						setGraphic(tf);
						setContentDisplay(ContentDisplay.GRAPHIC_ONLY);
						tf.setText(String.valueOf(getItem()));
						tf.requestFocus();
					}

					@Override
					public void cancelEdit() {
						super.cancelEdit();
						setText(String.valueOf(getItem()));
						setContentDisplay(ContentDisplay.TEXT_ONLY);
					}

				};
			}
		});

		DuplicationDistributionColumn.setCellFactory(new Callback<TableColumn<FaultInjectionTableEntry, String>, TableCell<FaultInjectionTableEntry, String>>() {

			@Override
			public TableCell<FaultInjectionTableEntry, String> call(TableColumn<FaultInjectionTableEntry, String> param) {

				TextField tf = new TextField();
				return new TableCell<FaultInjectionTableEntry, String>() {

					@Override
					protected void updateItem(String item, boolean empty) {
						super.updateItem(item, empty);
						if (item == null || empty) {
							setText(null);
							setStyle("");
						} else {
							setText(item);
							setTextFill((item.matches(regexExpDist) || item.matches(regexNormalDist) || item.matches(regexUniformDist)) ? Color.BLACK : Color.RED);
						}
					}

					@Override
					public void startEdit() {
						super.startEdit();
						tf.setOnKeyPressed(e -> {
							if (e.getCode().equals(KeyCode.TAB)) {
								commitEdit(tf.getText());
								setContentDisplay(ContentDisplay.TEXT_ONLY);
								getTableView().edit(getIndex(), OmissionPercentageColumn);
							} else if (e.getCode().equals(KeyCode.ENTER)) {
								commitEdit(tf.getText());
								setContentDisplay(ContentDisplay.TEXT_ONLY);
							}
						});

						setGraphic(tf);
						setContentDisplay(ContentDisplay.GRAPHIC_ONLY);
						tf.setText(getItem());
						tf.requestFocus();
					}

					@Override
					public void cancelEdit() {
						super.cancelEdit();
						setText(String.valueOf(getItem()));
						setContentDisplay(ContentDisplay.TEXT_ONLY);
					}

				};
			}
		});

		OmissionPercentageColumn.setCellFactory(new Callback<TableColumn<FaultInjectionTableEntry, Double>, TableCell<FaultInjectionTableEntry, Double>>() {

			@Override
			public TableCell<FaultInjectionTableEntry, Double> call(TableColumn<FaultInjectionTableEntry, Double> param) {

				TextField tf = new TextField();
				return new TableCell<FaultInjectionTableEntry, Double>() {

					@Override
					protected void updateItem(Double item, boolean empty) {
						super.updateItem(item, empty);
						if (item == null || empty) {
							setText(null);
							setStyle("");
						} else {
							setText(item.toString());
							setTextFill(item.toString().matches(regex0_100) ? Color.BLACK : Color.RED);
						}
					}

					@Override
					public void startEdit() {
						super.startEdit();
						tf.setOnKeyPressed(e -> {
							if (e.getCode().equals(KeyCode.TAB)) {
								if (tf.getText().matches(regexDoubleNumbers)) {
									commitEdit(Double.valueOf(tf.getText()));
								}
								setContentDisplay(ContentDisplay.TEXT_ONLY);
								getTableView().edit(getIndex(), CrashPercentageColumn);
							} else if (e.getCode().equals(KeyCode.ENTER)) {
								if (tf.getText().matches(regexDoubleNumbers)) {
									commitEdit(Double.valueOf(tf.getText()));
								}
								setContentDisplay(ContentDisplay.TEXT_ONLY);
							}
						});

						setGraphic(tf);
						setContentDisplay(ContentDisplay.GRAPHIC_ONLY);
						tf.setText(String.valueOf(getItem()));
						tf.requestFocus();
					}

					@Override
					public void cancelEdit() {
						super.cancelEdit();
						setText(String.valueOf(getItem()));
						setContentDisplay(ContentDisplay.TEXT_ONLY);
					}

				};
			}
		});

		CrashPercentageColumn.setCellFactory(new Callback<TableColumn<FaultInjectionTableEntry, Double>, TableCell<FaultInjectionTableEntry, Double>>() {

			@Override
			public TableCell<FaultInjectionTableEntry, Double> call(TableColumn<FaultInjectionTableEntry, Double> param) {

				TextField tf = new TextField();
				return new TableCell<FaultInjectionTableEntry, Double>() {

					@Override
					protected void updateItem(Double item, boolean empty) {
						super.updateItem(item, empty);
						if (item == null || empty) {
							setText(null);
							setStyle("");
						} else {
							setText(item.toString());
							setTextFill(item.toString().matches(regex0_100) ? Color.BLACK : Color.RED);
						}
					}

					@Override
					public void startEdit() {
						super.startEdit();
						tf.setOnKeyPressed(e -> {
							if (e.getCode().equals(KeyCode.TAB)) {
								if (tf.getText().matches(regexDoubleNumbers)) {
									commitEdit(Double.valueOf(tf.getText()));
								}
								setContentDisplay(ContentDisplay.TEXT_ONLY);
								getTableView().edit(getIndex(), DelayPercentageColumn);
							} else if (e.getCode().equals(KeyCode.ENTER)) {
								if (tf.getText().matches(regexDoubleNumbers)) {
									commitEdit(Double.valueOf(tf.getText()));
								}
								setContentDisplay(ContentDisplay.TEXT_ONLY);
							}
						});

						setGraphic(tf);
						setContentDisplay(ContentDisplay.GRAPHIC_ONLY);
						tf.setText(String.valueOf(getItem()));
						tf.requestFocus();
					}

					@Override
					public void cancelEdit() {
						super.cancelEdit();
						setText(String.valueOf(getItem()));
						setContentDisplay(ContentDisplay.TEXT_ONLY);
					}

				};
			}
		});

		DelayPercentageColumn.setCellFactory(new Callback<TableColumn<FaultInjectionTableEntry, Double>, TableCell<FaultInjectionTableEntry, Double>>() {

			@Override
			public TableCell<FaultInjectionTableEntry, Double> call(TableColumn<FaultInjectionTableEntry, Double> param) {

				TextField tf = new TextField();
				return new TableCell<FaultInjectionTableEntry, Double>() {

					@Override
					protected void updateItem(Double item, boolean empty) {
						super.updateItem(item, empty);
						if (item == null || empty) {
							setText(null);
							setStyle("");
						} else {
							setText(item.toString());
							setTextFill(item.toString().matches(regex0_100) ? Color.BLACK : Color.RED);
						}
					}

					@Override
					public void startEdit() {
						super.startEdit();
						tf.setOnKeyPressed(e -> {
							if (e.getCode().equals(KeyCode.TAB)) {
								if (tf.getText().matches(regexDoubleNumbers)) {
									commitEdit(Double.valueOf(tf.getText()));
								}
								setContentDisplay(ContentDisplay.TEXT_ONLY);
								getTableView().edit(getIndex(), DelayDistributionColumn);
							} else if (e.getCode().equals(KeyCode.ENTER)) {
								if (tf.getText().matches(regexDoubleNumbers)) {
									commitEdit(Double.valueOf(tf.getText()));
								}
								setContentDisplay(ContentDisplay.TEXT_ONLY);
							}
						});

						setGraphic(tf);
						setContentDisplay(ContentDisplay.GRAPHIC_ONLY);
						tf.setText(String.valueOf(getItem()));
						tf.requestFocus();
					}

					@Override
					public void cancelEdit() {
						super.cancelEdit();
						setText(String.valueOf(getItem()));
						setContentDisplay(ContentDisplay.TEXT_ONLY);
					}

				};
			}
		});
		DelayDistributionColumn.setCellFactory(new Callback<TableColumn<FaultInjectionTableEntry, String>, TableCell<FaultInjectionTableEntry, String>>() {

			@Override
			public TableCell<FaultInjectionTableEntry, String> call(TableColumn<FaultInjectionTableEntry, String> param) {

				TextField tf = new TextField();
				return new TableCell<FaultInjectionTableEntry, String>() {

					@Override
					protected void updateItem(String item, boolean empty) {
						super.updateItem(item, empty);
						if (item == null || empty) {
							setText(null);
							setStyle("");
						} else {
							setText(item);
							setTextFill((item.matches(regexExpDist) || item.matches(regexNormalDist) || item.matches(regexUniformDist)) ? Color.BLACK : Color.RED);
						}
					}

					@Override
					public void startEdit() {
						super.startEdit();
						tf.setOnKeyPressed(e -> {
							if (e.getCode().equals(KeyCode.TAB)) {
								commitEdit(tf.getText());
								setContentDisplay(ContentDisplay.TEXT_ONLY);
								getTableView().edit(getIndex(), BFPercentageColumn);
							} else if (e.getCode().equals(KeyCode.ENTER)) {
								commitEdit(tf.getText());
								setContentDisplay(ContentDisplay.TEXT_ONLY);
							}
						});

						setGraphic(tf);
						setContentDisplay(ContentDisplay.GRAPHIC_ONLY);
						tf.setText(getItem());
						tf.requestFocus();
					}

					@Override
					public void cancelEdit() {
						super.cancelEdit();
						setText(String.valueOf(getItem()));
						setContentDisplay(ContentDisplay.TEXT_ONLY);
					}

				};
			}
		});

		table.setOnKeyPressed(e -> {
			if (new KeyCodeCombination(KeyCode.S, KeyCodeCombination.CONTROL_DOWN).match(e)) {
				try {
					saveTable();
				} catch (Exception e1) {
				}
			}
		});

	}

	public void ButtonpressTableReset(ActionEvent event) throws Exception {
		ObservableList<FaultInjectionTableEntry> FI_selected = FXCollections.observableArrayList(table.getSelectionModel().getSelectedItems());
		bananalist.removeAll(FI_selected);
		for (FaultInjectionTableEntry faultInjectionTableEntry : FI_selected) {
			bananalist.add(new FaultInjectionTableEntry(faultInjectionTableEntry.getBananaIP()));
		}
	}

	public void ButtonpressSave(ActionEvent event) throws Exception {
		saveTable();
	}

	public void ButtonpressLoad(ActionEvent event) throws Exception {
		if (Rcc.out.getTerminal().getInput() == null) {
			return;
		} else if (Rcc.out.getTerminal().getInput().length() > 0) {
			Rcc.out.getTerminal().deleteInput();
		}

		FileChooser fileChooser = new FileChooser();
		fileChooser.setTitle("Open Json FaultInjection Table file");
		fileChooser.getExtensionFilters().add(new ExtensionFilter("json", "*.json"));
		fileChooser.setInitialDirectory(Paths.get("FISavedFiles").toFile());
		File filename = fileChooser.showOpenDialog(null);

		if (filename != null) {

			Rcc.out.terminalPrintln("Info: Loading local saved table entries from json file \"" + filename.getName() + "\"");

			try {
				FileReader reader = new FileReader(filename);
				JsonParser parser = new JsonParser();

				JsonArray jsonObject = (JsonArray) parser.parse(reader);
				ObservableList<FaultInjectionTableEntry> ListToAdd = FXCollections.observableArrayList();
				ObservableList<FaultInjectionTableEntry> ListToDelete = FXCollections.observableArrayList();
				boolean listIsConsistent = true;
				for (JsonElement jsonElement : jsonObject) {
					boolean isInList = false;
					for (FaultInjectionTableEntry bananalistentry : bananalist) {
						if (bananalistentry.getBananaIP().equals(jsonElement.getAsJsonObject().get("bannanaip").getAsString())) {
							ListToDelete.add(bananalistentry);
							isInList = true;
						}
					}
					if (!isInList) {
						listIsConsistent = false;
					}
					ListToAdd.add(new FaultInjectionTableEntry(jsonElement.getAsJsonObject().get("bannanaip").getAsString(), jsonElement.getAsJsonObject().get("bfpercentage").getAsDouble(),
							jsonElement.getAsJsonObject().get("bfdistribution").getAsString(), jsonElement.getAsJsonObject().get("bfrange").getAsString(),
							jsonElement.getAsJsonObject().get("bfMean").getAsInt(), jsonElement.getAsJsonObject().get("bfstandarddeviation").getAsDouble(),
							jsonElement.getAsJsonObject().get("wfpercentage").getAsDouble(), jsonElement.getAsJsonObject().get("wfportset").getAsString(),
							jsonElement.getAsJsonObject().get("duplicationpercentage").getAsDouble(), jsonElement.getAsJsonObject().get("duplicationdistribution").getAsString(),
							jsonElement.getAsJsonObject().get("duplicationrange").getAsString(), jsonElement.getAsJsonObject().get("duplicationMean").getAsInt(),
							jsonElement.getAsJsonObject().get("duplicationstandarddeviation").getAsDouble(), jsonElement.getAsJsonObject().get("omissionPercentage").getAsDouble(),
							jsonElement.getAsJsonObject().get("crashpercentage").getAsDouble(), jsonElement.getAsJsonObject().get("delaypercentage").getAsDouble(),
							jsonElement.getAsJsonObject().get("delaydistribution").getAsString(), jsonElement.getAsJsonObject().get("delayrange").getAsString(),
							jsonElement.getAsJsonObject().get("delayMean").getAsInt(), jsonElement.getAsJsonObject().get("delaystandarddeviation").getAsDouble()));
				}
				if (!listIsConsistent || ListToDelete.size() != bananalist.size()) {
					Alert alert = new Alert(AlertType.CONFIRMATION);
					alert.setTitle("Warning");
					alert.setHeaderText("The selected file and the current connected nodes do not match! \nThe upload will cause the Table to be in a inconsistent state.");
					alert.setContentText("Do you want to upload anyway?");

					Optional<ButtonType> result = alert.showAndWait();

					if (result.get() == ButtonType.OK) {
						bananalist.removeAll(ListToDelete);
						bananalist.addAll(ListToAdd);
						Rcc.out.terminalPrintln("File loaded successfully!", 1);
					} else {
						Rcc.out.terminalPrintln("Loading the file failed! Please choose a correct file!", 1);
					}
				} else {
					bananalist.removeAll(ListToDelete);
					bananalist.addAll(ListToAdd);
					Rcc.out.terminalPrintln("File loaded successfully!", 1);
				}

			} catch (Exception e) {
				Rcc.out.addException(e);
				Rcc.out.terminalPrintln(" Please check if the JSON File has the correct format! \n For additional information see the Exceptions Tab.");
			}
		} else {
			Rcc.out.terminalPrintln("No File has been choosen!");
		}

		Rcc.out.startCommand();
	}

	public void Buttonpressapply(ActionEvent event) throws Exception {

		Thread t = new Thread() {
			@Override
			public void run() {
				if (Rcc.out.getTerminal().getInput() == null) {
					return;
				} else if (Rcc.out.getTerminal().getInput().length() > 0) {
					Rcc.out.getTerminal().deleteInput();
				}

				Rcc.out.terminalPrintln("Injecting faults into nodes ...");
				Rcc.out.lockTerminal();

				for (FaultInjectionTableEntry fi_temp : bananalist) {

					Rcc.out.terminalPrintln("");
					Rcc.out.terminalPrintln("INJECTING INTO " + fi_temp.getBananaIP(), 1);

					ObservableList<BananaInformation> sendToBananaInformation = FXCollections.observableArrayList();

					// BananaInformation sendToBananaInformation = null;

					for (BananaInformation bananaInformation : Rcc.connectedNodes) {
						if (bananaInformation.getIP().equals(fi_temp.getBananaIP())) {
							sendToBananaInformation.add(bananaInformation);
						}

					}

					// Sending the commands to the BananaPI's
					Command c;

					// ###### Bitflips #######//
					Command.appendNextCommandToCommandchain();
					if (fi_temp.getBFDistribution().equals("discrete_equal_distribution")) {
						c = new SendRemoteCommand((fiTargetProtocol + " fi_jn_bf discrete_equal_distribution " + fi_temp.getBFPercentage() + " " + fi_temp.getBFRange()), sendToBananaInformation).exec();
						c.waitForTermination();
					} else if (fi_temp.getBFDistribution().equals("Gaussian_distribution")) {
						c = new SendRemoteCommand((fiTargetProtocol + " fi_jn_bf Gaussian_distribution " + fi_temp.getBFPercentage() + " " + fi_temp.getBFRange() + " " + fi_temp.getBFMean() + " "
								+ fi_temp.getBFStandardDeviation()), sendToBananaInformation).exec();
						c.waitForTermination();
					} else if (fi_temp.getBFDistribution().equals("Exponential_distribution")) {
						c = new SendRemoteCommand((fiTargetProtocol + " fi_jn_bf Exponential_distribution " + fi_temp.getBFPercentage() + " " + fi_temp.getBFLambda() + " " + fi_temp.getBFRange()),
								sendToBananaInformation).exec();
						c.waitForTermination();
					}

					try {
						Thread.currentThread().sleep(50);
					} catch (Exception e) {
						Rcc.out.addException(e);
					}

					//// ###### Wrong Forwarding #######//
					Command.appendNextCommandToCommandchain();
					c = new SendRemoteCommand((fiTargetProtocol + " fi_jn_wf " + fi_temp.getWFPercentage() + " " + fi_temp.getWFPortSetInt()), sendToBananaInformation).exec();
					c.waitForTermination();

					try {
						Thread.currentThread().sleep(50);
					} catch (Exception e) {
						Rcc.out.addException(e);
					}

					//// ###### Duplication #######//
					Command.appendNextCommandToCommandchain();
					if (fi_temp.getDuplicationDistribution().equals("discrete_equal_distribution")) {
						c = new SendRemoteCommand((fiTargetProtocol + " fi_jn_duplication discrete_equal_distribution " + fi_temp.getDuplicationPercentage() + " " + fi_temp.getDuplicationRange()),
								sendToBananaInformation).exec();
						c.waitForTermination();
					} else if (fi_temp.getDuplicationDistribution().equals("Gaussian_distribution")) {
						c = new SendRemoteCommand((fiTargetProtocol + " fi_jn_duplication Gaussian_distribution " + fi_temp.getDuplicationPercentage() + " " + fi_temp.getDuplicationRange() + " "
								+ fi_temp.getDuplicationMean() + " " + fi_temp.getDuplicationStandardDeviation()), sendToBananaInformation).exec();
						c.waitForTermination();
					} else if (fi_temp.getDuplicationDistribution().equals("Exponential_distribution")) {
						c = new SendRemoteCommand((fiTargetProtocol + " fi_jn_duplication Exponential_distribution " + fi_temp.getDuplicationPercentage() + " " + fi_temp.getDuplicationLambda() + " "
								+ fi_temp.getDuplicationRange()), sendToBananaInformation).exec();
						c.waitForTermination();
					}

					try {
						Thread.currentThread().sleep(50);
					} catch (Exception e) {
						Rcc.out.addException(e);
					}

					// ###### Omission #######//
					Command.appendNextCommandToCommandchain();
					c = new SendRemoteCommand((fiTargetProtocol + " fi_jn_omissionPercentage " + fi_temp.getOmissionPercentage()), sendToBananaInformation).exec();
					c.waitForTermination();

					try {
						Thread.currentThread().sleep(50);
					} catch (Exception e) {
						Rcc.out.addException(e);
					}

					// ###### Crash #######//
					Command.appendNextCommandToCommandchain();
					c = new SendRemoteCommand((fiTargetProtocol + " fi_jn_crashPercentage " + fi_temp.getCrashPercentage()), sendToBananaInformation).exec();
					c.waitForTermination();

					try {
						Thread.currentThread().sleep(50);
					} catch (Exception e) {
						Rcc.out.addException(e);
					}

					// ###### Delay #######//
					Command.appendNextCommandToCommandchain();
					if (fi_temp.getDelayDistribution().equals("discrete_equal_distribution")) {
						c = new SendRemoteCommand((fiTargetProtocol + " fi_jn_delay discrete_equal_distribution " + fi_temp.getDelayPercentage() + " " + fi_temp.getDelayRange()), sendToBananaInformation)
								.exec();
						c.waitForTermination();
					} else if (fi_temp.getDelayDistribution().equals("Gaussian_distribution")) {
						c = new SendRemoteCommand((fiTargetProtocol + " fi_jn_delay Gaussian_distribution " + fi_temp.getDelayPercentage() + " " + fi_temp.getDelayRange() + " " + fi_temp.getDelayMean()
								+ " " + fi_temp.getDelayStandardDeviation()), sendToBananaInformation).exec();
						c.waitForTermination();
					} else if (fi_temp.getDelayDistribution().equals("Exponential_distribution")) {
						c = new SendRemoteCommand(
								(fiTargetProtocol + " fi_jn_delay Exponential_distribution " + fi_temp.getDelayPercentage() + " " + fi_temp.getDelayLambda() + " " + fi_temp.getDelayRange()),
								sendToBananaInformation).exec();
						c.waitForTermination();
					}
					try {
						Thread.currentThread().sleep(50);
					} catch (Exception e) {
						Rcc.out.addException(e);
					}

					// ###### ByzantineBehavior ######//
					Command.appendNextCommandToCommandchain();
					c = new SendRemoteCommand((fiTargetProtocol + " fi_jn_byzantineBehavior " + fi_temp.getByzantineBehaviour().isSelected()), sendToBananaInformation).exec();
					c.waitForTermination();

					try {
						Thread.currentThread().sleep(50);
					} catch (Exception e) {
						Rcc.out.addException(e);
					}

					sendToBananaInformation.clear();
				}
				Rcc.out.startCommand();
				Rcc.out.unlockTerminal();
			}
		};
		t.start();

	}

	// set Main and calls function fill table
	public void setMain(Rcc main) {
		this.main = main;

		Rcc.connectedNodes.addListener(new ListChangeListener<BananaInformation>() {
			@Override
			public void onChanged(Change<? extends BananaInformation> c) {
				while (c.next()) {
					if (c.getAddedSize() > 0 || c.getRemovedSize() > 0 || c.wasPermutated()) {

						fillTable(Rcc.connectedNodes);

					}
				}
			}
		});
	}

	// check whether or not a banana is new or has been connected and add them to
	// the table
	void fillTable(ObservableList<BananaInformation> b) {
		ObservableList<FaultInjectionTableEntry> temp = FXCollections.observableArrayList();

		for (BananaInformation bananaInformation : b) {
			boolean InTableList = false;
			for (FaultInjectionTableEntry bananalistentry : bananalist) {
				// prüfe ob die Node aus b bereits in der Liste ist, falls gefunden setzte
				// InList auf true
				if (bananalistentry.getBananaIP().equals(bananaInformation.getIP())) {
					InTableList = true;
					temp.add(bananalistentry);
				}
			}
			// Hinzufügen neuer Node falls noch nicht in der Liste
			if (!InTableList) {
				temp.add(new FaultInjectionTableEntry(bananaInformation.getIP()));
			}
		}

		if (b.size() != 0) {
			bananalist.clear();
			bananalist.addAll(temp);
		}

	}

	void saveTable() throws Exception {

		JsonArray JsonEntireArray = new JsonArray();
		for (FaultInjectionTableEntry fitemp : bananalist) {
			JsonObject jsonObject = new JsonObject();

			jsonObject.addProperty("bannanaip", fitemp.getBananaIP());
			jsonObject.addProperty("bfpercentage", Double.toString(fitemp.getBFPercentage()));
			jsonObject.addProperty("bfdistribution", fitemp.getBFDistribution());
			jsonObject.addProperty("bfrange", fitemp.getBFRange());
			jsonObject.addProperty("bfMean", Integer.toString(fitemp.getBFMean()));
			jsonObject.addProperty("bfstandarddeviation", Double.toString(fitemp.getBFStandardDeviation()));
			jsonObject.addProperty("wfpercentage", Double.toString(fitemp.getWFPercentage()));
			jsonObject.addProperty("wfportset", fitemp.getWFPortSet());
			jsonObject.addProperty("duplicationpercentage", Double.toString(fitemp.getDuplicationPercentage()));
			jsonObject.addProperty("duplicationdistribution", fitemp.getDuplicationDistribution());
			jsonObject.addProperty("duplicationrange", fitemp.getDuplicationRange());
			jsonObject.addProperty("duplicationMean", Integer.toString(fitemp.getDuplicationMean()));
			jsonObject.addProperty("duplicationstandarddeviation", Double.toString(fitemp.getDuplicationStandardDeviation()));
			jsonObject.addProperty("omissionPercentage", Double.toString(fitemp.getOmissionPercentage()));
			jsonObject.addProperty("crashpercentage", Double.toString(fitemp.getCrashPercentage()));
			jsonObject.addProperty("delaypercentage", Double.toString(fitemp.getDelayPercentage()));
			jsonObject.addProperty("delaydistribution", fitemp.getDelayDistribution());
			jsonObject.addProperty("delayrange", fitemp.getDelayRange());
			jsonObject.addProperty("delayMean", Integer.toString(fitemp.getDelayMean()));
			jsonObject.addProperty("delaystandarddeviation", Double.toString(fitemp.getDelayStandardDeviation()));

			JsonEntireArray.add(jsonObject);
		}

		FileChooser fileChooser = new FileChooser();
		fileChooser.setTitle("Save FaultInjection Table json File");
		fileChooser.setInitialDirectory(Paths.get("FISavedFiles").toFile());
		fileChooser.getExtensionFilters().add(new ExtensionFilter("json", "*.json"));
		fileChooser.setInitialFileName(new SimpleDateFormat("yyyyMMDD_HH_mm_ss").format(new Date()) + "_fi_jn_table" + ".json");
		File file = fileChooser.showSaveDialog(null);

		if (file != null) {
			PrintWriter writer;
			writer = new PrintWriter(file);
			writer.println(JsonEntireArray);
			writer.close();

			Rcc.out.terminalPrintln("Info: \"" + file.getName() + "\"" + " has been saved successfully");
		} else {
			Rcc.out.terminalPrintln("Saving the table failed!");
		}

		Rcc.out.startCommand();
	}

}
