package configgui.java;

import java.io.FileNotFoundException;
import java.util.ArrayList;

import configgui.java.controller.BananaController;
import configgui.java.model.configuration.BananaInformation;
import configgui.java.model.configuration.Configuration;
import configgui.java.util.network.ControlMsgListener;
import configgui.java.util.network.ServerThread;
import configgui.java.util.terminal.GeneralOutputInterface;
import javafx.application.Application;
import javafx.beans.Observable;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.fxml.FXMLLoader;
import javafx.scene.Parent;
import javafx.scene.Scene;
import javafx.scene.image.Image;
import javafx.stage.Stage;

public class Rcc extends Application {

	static public GeneralOutputInterface out;
	static public ControlMsgListener ctrlMsgListener;
	static public Configuration configuration;
	static public ObservableList<BananaInformation> connectedNodes;

	private BananaController bananaController;
	private Stage primaryStage;
	private ArrayList<Exception> initialTemporaryExceptionList;
	private ServerThread server;
	private Thread serverThread, ctrlMsgListenerThread;

	public static void main(String[] args) {
		launch(args);
	}

	@Override
	public void init() {
		initialTemporaryExceptionList = new ArrayList<>();
		connectedNodes = FXCollections.observableArrayList(node -> new Observable[] { node.state(), node.load() });

		ctrlMsgListener = new ControlMsgListener(this);
		ctrlMsgListenerThread = new Thread(ctrlMsgListener);
		ctrlMsgListenerThread.start();

		server = new ServerThread(this);
		serverThread = new Thread(server);
		serverThread.start();

		configuration = new Configuration();
		try {
			configuration.loadOrCreateConfiguration();
		} catch (Exception e) {
			if (e instanceof FileNotFoundException)
				configuration.createDefaultConfiguration();
			initialTemporaryExceptionList.add(e);
		}
	}

	@Override
	public void start(Stage stage) throws Exception {
		// Determine windowsize
		String windowsize = Rcc.configuration.getConfigEntry("windowsize");
		int width = 1024;
		int height = 600;
		if (windowsize != null) {
			String[] confRes = windowsize.split("x");
			if (confRes.length != 2) {
				Rcc.configuration.setConfigEntry("windowsize", width + "x" + height);
			} else {
				try {
					width = Integer.parseInt(confRes[0]);
					height = Integer.parseInt(confRes[1]);
				} catch (Exception e) {
					Rcc.configuration.setConfigEntry("windowsize", width + "x" + height);
				}
			}
		} else {
			Rcc.configuration.setConfigEntry("windowsize", width + "x" + height);
		}
		stage.setMinWidth(1600);
		stage.setMinHeight(900);
		stage.setWidth(width);
		stage.setHeight(height);

		// Determine position
		String position = Rcc.configuration.getConfigEntry("position");
		int xPos = 0;
		int yPos = 0;
		if (position != null) {
			String[] confPos = position.split(",");
			if (confPos.length != 2) {
				Rcc.configuration.setConfigEntry("position", xPos + "," + yPos);
			} else {
				try {
					xPos = Integer.parseInt(confPos[0]);
					yPos = Integer.parseInt(confPos[1]);
				} catch (Exception e) {
					Rcc.configuration.setConfigEntry("position", xPos + "," + yPos);
				}
			}
		} else {
			Rcc.configuration.setConfigEntry("position", xPos + "," + yPos);
		}
		stage.setX(xPos);
		stage.setY(yPos);

		stage.getIcons().add(new Image(getClass().getResourceAsStream("/configgui/resources/icon.png")));

		FXMLLoader loader = new FXMLLoader();
		loader.setLocation(getClass().getResource("../view/ConfigTool.fxml"));
		Parent root = loader.load();
		primaryStage = stage;
		bananaController = loader.getController();

		Scene scene = new Scene(root);
		scene.getStylesheets().add("/configgui/resources/css/mainstyle.css");
		scene.getStylesheets().add("/configgui/resources/css/syntax.css");

		stage.setTitle("Remote Control Central");
		stage.setScene(scene);
		stage.show();
		
		bananaController.initController(this, initialTemporaryExceptionList);

	}

	@Override
	public void stop() {
		for (BananaInformation banana : connectedNodes) {
			banana.stopClientThread();
		}
		server.stopServer();
		ctrlMsgListener.active = false;
		ctrlMsgListenerThread.interrupt();
		out.active = false;
	}

	public Stage getPrimaryStage() {
		return primaryStage;
	}

	public BananaController getBananaController() {
		return bananaController;
	}

}
