package configgui.java.commands;

import java.util.Vector;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;

import configgui.java.Rcc;
import configgui.java.model.configuration.BananaInformation;
import javafx.application.Platform;
import javafx.collections.ObservableList;

public abstract class Command {

	static private boolean APPENDTOCOMMANDCHAIN = false;
	static private int GLOBALCMDCOUNTER = 0;
	static public String[] COMMANDLIST = { "help", "get", "scan", "bpi", "put", "discovertopology", "set", "update", "updatelib" };

	protected boolean startNewComand = true;
	protected String cmdId;
	protected Vector<String[]> responseBuffer;
	protected Semaphore responseBufferCounter;
	protected Thread t;
	protected boolean success = true;

	public Command() {
		startNewComand = !APPENDTOCOMMANDCHAIN;
		APPENDTOCOMMANDCHAIN = false;

		cmdId = String.format("ID%03d", GLOBALCMDCOUNTER);
		if (startNewComand)
			GLOBALCMDCOUNTER = (GLOBALCMDCOUNTER + 1) % 1000;

		responseBufferCounter = new Semaphore(0);
		responseBuffer = new Vector<>();
	}

	public Command exec() {
		Rcc.out.statusClear();

		if (!validate())
			return null;

		t = new Thread() {
			@Override
			public void run() {
				executeCommand();
				if (startNewComand)
					Platform.runLater(() -> Rcc.out.startCommand());
			};
		};
		t.start();

		return this;
	}

	protected abstract void executeCommand();

	protected abstract boolean validate();

	public static void appendNextCommandToCommandchain() {
		APPENDTOCOMMANDCHAIN = true;
	}

	public String getCmdId() {
		return cmdId;
	}

	public boolean addResponse(String[] s) {
		if (s.length == 2) {
			responseBuffer.add(s);
			responseBufferCounter.release();
			return true;
		}
		return false;
	}

	public Boolean awaitStat(ObservableList<BananaInformation> bananas, long timeout_ms, boolean print) {
		Vector<String> busyBananaList = new Vector<>();
		Vector<BananaInformation> busyBananaInfoList = new Vector<>();

		Rcc.out.activateProgressBar();
		Rcc.out.updateProgressBar(0, bananas.size());

		for (BananaInformation banana : bananas) {
			busyBananaList.add(banana.getIP());
			busyBananaInfoList.add(banana);
			Rcc.out.setState(banana, BananaInformation.state_busy);
		}

		boolean result = true;

		long timeout = System.currentTimeMillis() + timeout_ms;
		while (System.currentTimeMillis() < timeout && busyBananaList.size() > 0) {
			try {
				if (responseBufferCounter.tryAcquire(timeout - System.currentTimeMillis(), TimeUnit.MILLISECONDS)) {
					String[] msg = responseBuffer.get(0);
					responseBuffer.remove(0);

					int index = busyBananaList.indexOf(msg[0]);
					if (index >= 0) {

						String[] items = msg[1].split(";");
						switch (items[2]) {
						case "\"error\"":
							Rcc.out.setState(busyBananaInfoList.get(index), BananaInformation.state_error);
							result = false;
							break;
						case "\"fail\"":
							Rcc.out.setState(busyBananaInfoList.get(index), BananaInformation.state_error);
							result = false;
							break;
						case "\"success\"":
							Rcc.out.setState(busyBananaInfoList.get(index), BananaInformation.state_idle);
							result = result & true;
							break;
						}

						if (print)
							Rcc.out.terminalPrintln("RX (" + msg[0] + ") " + msg[1], 2);
						Rcc.out.setState(busyBananaInfoList.get(index), BananaInformation.state_idle);

						busyBananaInfoList.remove(index);
						busyBananaList.remove(index);

						Rcc.out.updateProgressBar(bananas.size() - busyBananaInfoList.size(), bananas.size());
					} else {
						System.out.println("Unknown ctrl message received: " + msg[1]);
					}
				} else {
					return null;
				}
			} catch (Exception e) {
				Rcc.out.addException(e);
			}
		}
		return result;
	}

	public boolean waitForTermination() {
		try {
			t.join();
			return success;
		} catch (Exception e) {
			Rcc.out.addException(e);
			return false;
		}
	}
}
