package configgui.java.commands.chains;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Random;
import java.util.Vector;
import java.util.concurrent.TimeUnit;

import configgui.java.Rcc;
import configgui.java.commands.Command;
import configgui.java.controller.FabanCompanionController;
import configgui.java.model.configuration.BananaInformation;
import javafx.collections.ObservableList;

public class ConfigureFabanNodes extends Command {

	ArrayList<String[]> routingTable;
	ObservableList<BananaInformation> bananas;
	FabanCompanionController fabCtrl;

	public ConfigureFabanNodes(ObservableList<BananaInformation> bananas, ArrayList<String[]> routingTable, FabanCompanionController fabCtrl) {
		this.bananas = bananas;
		this.routingTable = routingTable;
		this.fabCtrl = fabCtrl;
	}

	@Override
	public void executeCommand() {
		Vector<String> busyBananaList = new Vector<>();
		Rcc.ctrlMsgListener.registerCommandThread(getCmdId(), this);

		// 1. Retrieve Faban status
		for (BananaInformation banana : bananas) {
			Rcc.out.bananaOutPrintln(banana, "> " + "bpi Faban status");
			banana.sendMessage("Faban status" + " " + getCmdId());
			busyBananaList.add(banana.getIP());
			busyBananaList.add(banana.getIP());
		}
		Rcc.out.terminalPrint("Faban running?\t\t\t", 1);
		busyBananaList = awaitResponses(2, busyBananaList, new ArrayList<String>(Arrays.asList("active")), false,
				false);
		Rcc.out.terminalPrintln((busyBananaList.size() == 0 ? "[OK]" : "[NO]"));

		// 1b. Start Faban (optional)
		if (busyBananaList.size() > 0) {
			busyBananaList.clear();
			for (BananaInformation banana : bananas) {
				Rcc.out.bananaOutPrintln(banana, "> " + "bpi pstart Faban");
				banana.sendMessage("pstart Faban" + " " + getCmdId());
				busyBananaList.add(banana.getIP());
			}
			Rcc.out.terminalPrint("Start Faban\t\t\t\t", 1);
			busyBananaList = awaitResponses(2, busyBananaList, new ArrayList<String>());
			Rcc.out.terminalPrintln((busyBananaList.size() == 0 ? "[OK]" : "[FAILED]"));
			if (busyBananaList.size() > 0) {
				Rcc.out.terminalPrintln("Starting of Faban failed. Abort...", 1);
				return;
			}

			try {
				Thread.currentThread();
				Thread.sleep(1000);
			} catch (Exception e) {
				Rcc.out.addException(e);
			}
			busyBananaList.clear();
			for (BananaInformation banana : bananas) {
				Rcc.out.bananaOutPrintln(banana, "> " + "bpi Faban status");
				banana.sendMessage("Faban status" + " " + getCmdId());
				busyBananaList.add(banana.getIP());
				busyBananaList.add(banana.getIP());
			}
			Rcc.out.terminalPrint("Faban running?\t\t\t", 1);
			busyBananaList = awaitResponses(5, busyBananaList, new ArrayList<String>(Arrays.asList("active")));
			Rcc.out.terminalPrintln((busyBananaList.size() == 0 ? "[OK]" : "[NO]"));

			if (busyBananaList.size() > 0) {
				Rcc.out
						.terminalPrintln("Faban still not running on each node. Abort... " + busyBananaList.size(), 1);
				return;
			}
		}

		// 2. Reset possible existent routing information on all nodes
		busyBananaList.clear();
		for (BananaInformation banana : bananas) {
			Rcc.out.bananaOutPrintln(banana, "> " + "bpi Faban resetrouting");
			banana.sendMessage("Faban resetrouting" + " " + getCmdId());
			busyBananaList.add(banana.getIP());
			busyBananaList.add(banana.getIP());
		}
		Rcc.out.terminalPrint("Reset routing information\t\t", 1);
		busyBananaList = awaitResponses(2, busyBananaList, new ArrayList<String>(Arrays.asList("resetted")));
		Rcc.out.terminalPrintln((busyBananaList.size() == 0 ? "[OK]" : "[FAILED]"));
		if (busyBananaList.size() > 0) {
			Rcc.out.terminalPrintln("Abort: Resetting routing information failed!", 1);
			return;
		}

		// 2b. Initialize Protocol
		busyBananaList.clear();
		for (BananaInformation banana : bananas) {
			Rcc.out.bananaOutPrintln(banana, "> " + "bpi Faban init " + banana.getId());
			banana.sendMessage("Faban init " + banana.getId() + " " + getCmdId());
			busyBananaList.add(banana.getIP());
		}
		Rcc.out.terminalPrint("Initializing Faban\t\t\t", 1);
		busyBananaList = awaitResponses(2, busyBananaList, new ArrayList<String>());
		Rcc.out.terminalPrintln((busyBananaList.size() == 0 ? "[OK]" : "[FAILED]"));
		if (busyBananaList.size() > 0) {
			Rcc.out.terminalPrintln("Abort: Faban initialization!", 1);
			return;
		}

		// 3. Send routing information to all nodes
		Rcc.out.terminalPrint("Configure nodes\t\t\t", 1);
		busyBananaList.clear();
		routingTable.sort(new Comparator<String[]>() {
			public int compare(String[] o1, String[] o2) {
				return o1[0].compareTo(o2[0]);
			};
		});

		BananaInformation curBan = null;
		String curBanStr = "";
		for (String[] s : routingTable) {
			if (!s[0].equals(curBanStr)) {
				for (BananaInformation b : bananas) {
					if (b.getIP().equals(s[0])) {
						curBanStr = s[0];
						curBan = b;
						break;
					}
				}
			}
			System.out.println("[" + curBan.getIP() + "]  Faban addroutinginfo " + s[1] + " " + getCmdId());
			curBan.sendMessage("Faban addroutinginfo " + s[1] + " " + getCmdId());
			busyBananaList.add(curBan.getIP());
		}

		busyBananaList = awaitResponses(2, busyBananaList, new ArrayList<>());
		Rcc.out.terminalPrintln((busyBananaList.size() == 0 ? "[OK]" : "[FAILED]"));

		// 4. Calculate and send signature configuration to all nodes
		Rcc.out.terminalPrint("Configure signature information\t", 1);
		busyBananaList.clear();
		Random rand = new Random();
		for (BananaInformation b : bananas) {
			long p = 1 + 2 * (long) Math.abs(rand.nextInt());
			long q = 1 + 2 * (long) Math.abs(rand.nextInt());
			long r = oddMultiplikation(p, q, 32);
			long max = (long) (Math.pow(2, 32) - 1);
			if (q > max || p > max || r > max) {
				System.err.println(String.format("SGN NUmber too large: p=%l q=%l r=%l", p, q, r));
			}
			for (BananaInformation b2 : bananas) {
				if (b == b2) {
					b2.sendMessage("Faban addsgnprivate " + p + " " + getCmdId());
					busyBananaList.add(b2.getIP());
				}
				b2.sendMessage("Faban addsgnpublic " + b.getId() + " " + q + " " + r + " " + getCmdId());
				busyBananaList.add(b2.getIP());
			}
		}
		busyBananaList = awaitResponses(2, busyBananaList, new ArrayList<>());
		Rcc.out.terminalPrintln((busyBananaList.size() == 0 ? "[OK]" : "[FAILED]"));

		// 5. Generate and commit Faban masks
		Rcc.out.terminalPrint("Generate Faban Masks\t\t\t", 1);
		busyBananaList.clear();
		long dMask = (long) rand.nextInt();
		long cMask = (long) rand.nextInt();
		dMask = (dMask >= 0 ? dMask : Integer.MAX_VALUE - dMask);
		cMask = (cMask >= 0 ? cMask : Integer.MAX_VALUE - cMask);
		long rMask = rotate(dMask, 1, 32) ^ rotate(cMask, 2, 32);
		for (BananaInformation b : bananas) {
			b.sendMessage("Faban setmasks " + dMask + " " + cMask + " " + rMask + " " + getCmdId());
			busyBananaList.add(b.getIP());
		}
		busyBananaList = awaitResponses(2, busyBananaList, new ArrayList<>());
		Rcc.out.terminalPrintln((busyBananaList.size() == 0 ? "[OK]" : "[FAILED]"));

		// Finish Configuration
		fabCtrl.switchLabel(fabCtrl.labelNodesConfigured, true);
		fabCtrl.switchLabel(fabCtrl.labelReadyForExecution, true);

		Rcc.ctrlMsgListener.deregisterCommandThread(getCmdId(), this);
	}

	private long rotate(long value, int pos, int size) {
		for (int i = 0; i < pos; i++) {
			long incBit = (long) (value % 2);
			value = (value >>> 1) | (incBit << 31);
		}
		return value;
	}

	private long oddMultiplikation(long x, long y, int u) {
		int xPot = (int) (Math.log(x) / Math.log(2));
		int yPot = (int) (Math.log(y) / Math.log(2));
		long xRest = x - (long) (Math.pow(2, xPot));
		long yRest = y - (long) (Math.pow(2, yPot));

		long mod = (long) Math.pow(2, u);

		long result = 0;
		result = (result + (xPot + yPot >= 32 ? 0 : (long) Math.pow(2, xPot + yPot))) % mod;
		result = (result + xRest * (long) Math.pow(2, yPot)) % mod;
		result = (result + yRest * (long) Math.pow(2, xPot)) % mod;
		result = (result + xRest * yRest);
		result = (result * 2 + x + y) % mod;
		result = (result < 0 ? (result + mod) % mod : result);
//		System.out.println(x + " == " + (long)(Math.pow(2, xPot) + xRest) + " = 2^" + xPot + " + " + xRest);
//		System.out.println(y + " == " + (long)(Math.pow(2, yPot) + yRest) + " = 2^" + yPot + " + " + yRest);
//		System.out.println("x**y = " + result);
		return result;
	}

	public Vector<String> awaitResponses(int timeoutSec, Vector<String> busyBananaList, ArrayList<String> keywords) {
		return awaitResponses(timeoutSec, busyBananaList, keywords, false, true);
	}

	public Vector<String> awaitResponses(int timeoutSec, Vector<String> busyBananaList, ArrayList<String> keywords,
			boolean debugOutput, boolean printError) {
		long timeout = System.currentTimeMillis() + timeoutSec * 1000;
		while (System.currentTimeMillis() < timeout && busyBananaList.size() > 0) {
			try {
				if (responseBufferCounter.tryAcquire(timeout - System.currentTimeMillis(), TimeUnit.MILLISECONDS)) {
					String[] msg = responseBuffer.get(0);
					responseBuffer.remove(0);
					String[] items = msg[1].split(";");

					if (debugOutput)
						System.out.println("[" + msg[0] + "] " + msg[1]);

					int index = busyBananaList.indexOf(msg[0]);
					if (index >= 0) {
						if (items.length >= 2 && keywords.contains(items[1].substring(1, items[1].length() - 1))) {
							busyBananaList.remove(msg[0]);
						} else if (items.length > 2 && items[1].equals("\"S\"")
								&& (items[2].equals("\"success\"") || items[2].equals("\"error\""))) {
							busyBananaList.remove(msg[0]);
							if (printError && items[2].equals("\"error\""))
								System.out.println("Error message received: " + msg[0] + " " + msg[1]);
						} else {
							System.out.println("Unexpected ctrl msg: " + msg[0] + " " + msg[1]);
						}
					}
				}
			} catch (Exception e) {
				Rcc.out.addException(e);
			}
		}
		return busyBananaList;
	}

	@Override
	public boolean validate() {
		return true;
	}
}
