package configgui.java.commands.chains;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.concurrent.TimeUnit;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonArray;
import com.google.gson.JsonObject;

import configgui.java.Rcc;
import configgui.java.commands.Command;
import configgui.java.controller.FabanCompanionController;
import configgui.java.model.configuration.BananaInformation;
import javafx.collections.ObservableList;

public class DiscoverTopologyCommand extends Command {

	ObservableList<BananaInformation> bananas;
	FabanCompanionController fabCompCtrl;

	public DiscoverTopologyCommand(ObservableList<BananaInformation> bananas, FabanCompanionController fabCompCtrl) {
		this.bananas = bananas;
		this.fabCompCtrl = fabCompCtrl;
	}

	@Override
	public void executeCommand() {
		Rcc.out.terminalPrintln("Start topology discovery. Waiting for responses.\n", 1);
		Rcc.ctrlMsgListener.registerCommandThread(getCmdId(), this);
		for (BananaInformation banana : bananas) {
			Rcc.out.bananaOutPrintln(banana, "> " + "bpi pstart DiscoveryProtocol");
			banana.sendMessage("pstart DiscoveryProtocol" + " " + getCmdId());
		}
		try {
			Thread.sleep(1000);
		} catch (Exception e) {
			Rcc.out.addException(e);
		}

		for (BananaInformation banana : bananas) {
			Rcc.out.bananaOutPrintln(banana, "> " + "bpi DiscoveryProtocol discover");
			banana.sendMessage("DiscoveryProtocol discover" + " " + getCmdId());
		}

		JsonObject file = new JsonObject();
		JsonObject mapping = new JsonObject();
		JsonObject topology = new JsonObject();
		topology.add("endstations", new JsonArray());
		JsonArray bridges = new JsonArray();
		JsonArray links = new JsonArray();

		HashMap<String, String> macIpTable = new HashMap<>();
		ArrayList<String[]> routing = new ArrayList<>();

		long timeout = System.currentTimeMillis() + 5 * 1000;
		while (System.currentTimeMillis() < timeout) {
			try {
				if (responseBufferCounter.tryAcquire(5, TimeUnit.SECONDS)) {
					String[] msg = responseBuffer.get(0);
					responseBuffer.remove(0);
					String[] items = msg[1].split(";");
					if (items[1].substring(1, items[1].length() - 1).equals("myMac")) {
						macIpTable.put(items[2].substring(1, items[2].length() - 1), msg[0]);
						mapping.addProperty(msg[0], items[2].substring(1, items[2].length() - 1));
						bridges.add(msg[0]);
						Rcc.out.terminalPrintln("RX (" + msg[0] + ") " + msg[1], 2);
					} else if (items[1].substring(1, items[1].length() - 1).equals("S")) {
					} else if (items.length == 4) {
						routing.add(msg);
						Rcc.out.terminalPrintln("RX (" + msg[0] + ") " + msg[1], 2);
					}
				}
			} catch (Exception e) {
				Rcc.out.addException(e);
			}
		}

		Rcc.out.terminalPrintln("Topology composition started...", 1);

		ArrayList<String> linkList = new ArrayList<>();
		for (String[] r : routing) {
			String b1 = r[0];
			String[] split = r[1].split(";");
			String b2 = macIpTable.get(split[2].substring(1, split[2].length() - 1));
			if (b2 == null) {
				Rcc.out.terminalPrintln(
						"Topology composition failed. Unknown component with mac " + split[2] + " appeared.", 2);
				return;
			}
			int p1 = Integer.parseInt(split[1].substring(("rx_").length() + 1, split[1].length() - 1));
			int p2 = Integer.parseInt(split[3].substring(("tx_").length() + 1, split[3].length() - 1));

			String link;
			if (b1.compareTo(b2) > 0) {
				String temp = b1;
				b1 = b2;
				b2 = temp;

				int t = p1;
				p1 = p2;
				p2 = t;
			}

			link = b1 + "-" + b2;

			if (!linkList.contains(link)) {
				linkList.add(link);

				JsonObject joLink = new JsonObject();
				JsonObject a = new JsonObject();
				JsonObject b = new JsonObject();
				a.addProperty("name", b1);
				a.addProperty("port", p1);
				b.addProperty("name", b2);
				b.addProperty("port", p2);
				joLink.add("A", a);
				joLink.add("B", b);

				links.add(joLink);
			}
		}

		file.add("mapping", mapping);

		topology.add("bridges", bridges);
		topology.add("links", links);
		file.add("topology", topology);

		String filename = new SimpleDateFormat("yyyyMMDD_HH_mm_ss").format(new Date()) + ".json";

		fabCompCtrl.setTopology(file, filename);

		try {
			if (!(new File("topologies").exists())) {
				new File("topologies").mkdirs();
			}
			PrintWriter out = new PrintWriter(new File("topologies/" + filename));
			Gson g = new GsonBuilder().setPrettyPrinting().create();
			g.toJson(file, out);
			out.close();
		} catch (FileNotFoundException e) {
			Rcc.out.addException(e);
		}

		Rcc.out.terminalPrintln("Topology successfully discovered. Saved to \"" + filename + "\"", 1);

		for (BananaInformation banana : bananas) {
			Rcc.out.bananaOutPrintln(banana, "> " + "bpi pstop DiscoveryProtocol");
			banana.sendMessage("pstop DiscoveryProtocol" + " " + getCmdId());
		}

		awaitStat(bananas, 5 * 1000, false);
		Rcc.ctrlMsgListener.deregisterCommandThread(getCmdId(), this);
	}

	@Override
	public boolean validate() {
		return true;
	}
}
