package configgui.java.commands.chains;

import java.util.Optional;
import java.util.concurrent.Callable;
import java.util.concurrent.FutureTask;

import configgui.java.Rcc;
import configgui.java.commands.Command;
import configgui.java.commands.simple.UploadCommand;
import configgui.java.model.configuration.BananaInformation;
import javafx.application.Platform;
import javafx.collections.ObservableList;
import javafx.geometry.Insets;
import javafx.scene.control.ButtonBar.ButtonData;
import javafx.scene.control.ButtonType;
import javafx.scene.control.Dialog;
import javafx.scene.control.Label;
import javafx.scene.control.TextField;
import javafx.scene.layout.GridPane;
import javafx.util.Pair;

public class UpdateProtocol extends Command {
	ObservableList<BananaInformation> bananas;
	String parameter;
	boolean startAfterCompilation;

	public UpdateProtocol(ObservableList<BananaInformation> bananas, String parameter, boolean start) {
		this.bananas = bananas;
		this.parameter = parameter;
		this.startAfterCompilation = start;
	}

	@Override
	public void executeCommand() {
		String[] items = parameter.split("/");
		String protocolName = items[items.length - 1];

		Rcc.ctrlMsgListener.registerCommandThread(getCmdId(), this);

		Rcc.out.terminalPrintln("Start updating Protocol \"" + protocolName + "\".\n", 1);

		Command.appendNextCommandToCommandchain();
		UploadCommand upCmd = new UploadCommand(bananas, parameter);
		upCmd.exec();
		Rcc.out.terminalPrint("Uploading \"" + protocolName + "\" ... ", 1);

		if (upCmd.waitForTermination()) {

			Boolean retValue;

			// Terminate Protocol
			for (BananaInformation banana : bananas) {
				Rcc.out.bananaOutPrintln(banana, "> " + "bpi pstop " + protocolName);
				banana.sendMessage("pstop " + protocolName + " " + getCmdId());
			}
			Rcc.out.terminalPrint("Stopping " + protocolName + " ... ", 1);
			retValue = awaitStat(bananas, 10 * 1000, false);
			Rcc.out.terminalPrintln((retValue == null ? "[Timeout]" : (retValue ? " [OK]" : " [OK]")), 0);

			// Check whether a config entry exists
			for (BananaInformation banana : bananas) {
				Rcc.out.bananaOutPrintln(banana, "> " + "bpi getc " + protocolName);
				banana.sendMessage("getc " + protocolName + " " + getCmdId());
			}
			Rcc.out.terminalPrint("Check for config entries", 1);
			retValue = awaitStat(bananas, 180 * 1000, false);
			Rcc.out.terminalPrintln((retValue == null ? "[Timeout]" : (retValue ? " [OK]" : " [NO]")), 0);

			// Propose to create a config entry
			if (!retValue) {
				final FutureTask<Optional<Pair<String, String>>> dialog = new FutureTask<>(new Callable<Optional<Pair<String, String>>>() {
					@Override
					public Optional<Pair<String, String>> call() throws Exception {
						Dialog<Pair<String, String>> setConfigDialog = new Dialog<>();
						setConfigDialog.setTitle("");
						setConfigDialog.setHeaderText("Create a protocol config entry ...");
						ButtonType createButtonType = new ButtonType("Create!", ButtonData.OK_DONE);
						ButtonType dontCreateButtonType = new ButtonType("No! Don't create config entry!", ButtonData.CANCEL_CLOSE);
						setConfigDialog.getDialogPane().getButtonTypes().addAll(createButtonType, dontCreateButtonType);

						GridPane grid = new GridPane();
						grid.setHgap(10);
						grid.setVgap(10);
						grid.setPadding(new Insets(20, 150, 10, 10));

						TextField executable = new TextField();
						executable.setText(protocolName);
						executable.setPrefWidth(500);
						TextField path = new TextField();
						path.setPrefWidth(500);
						path.setText(Rcc.configuration.getDestinationPath() + "/" + parameter + "/Release/");
						path.setPromptText("Path to executable: ");

						grid.add(new Label("Protocol executable name: "), 0, 0);
						grid.add(executable, 1, 0);
						grid.add(new Label("Path to executable: "), 0, 1);
						grid.add(path, 1, 1);

						setConfigDialog.getDialogPane().setContent(grid);

						Platform.runLater(() -> executable.requestFocus());

						setConfigDialog.setResultConverter(dialogButton -> {
							if (dialogButton == createButtonType) {
								return new Pair<>(executable.getText(), path.getText());
							}
							return null;
						});

						Optional<Pair<String, String>> result = setConfigDialog.showAndWait();

						return result;
					}
				});

				Platform.runLater(dialog);

				try {
					Optional<Pair<String, String>> result = dialog.get();
					// Set configuration
					for (BananaInformation banana : bananas) {
						Rcc.out.bananaOutPrintln(banana, "> " + "bpi setc " + result.get().getKey() + " " + result.get().getValue());
						banana.sendMessage("setc " + result.get().getKey() + " " + result.get().getValue() + " " + getCmdId());
					}
					Rcc.out.terminalPrint("Set config entry ... ", 1);
					retValue = awaitStat(bananas, 10 * 1000, false);
					Rcc.out.terminalPrintln((retValue == null ? "[Timeout]" : (retValue ? " [OK]" : " [Error]")), 0);
				} catch (Exception e) {
				}

			}

			if (retValue) {
				// Compile protocol
				for (BananaInformation banana : bananas) {
					Rcc.out.bananaOutPrintln(banana, "> " + "bpi pcompile " + protocolName);
					banana.sendMessage("pcompile " + protocolName + " " + getCmdId());
				}
				Rcc.out.terminalPrint("Compiling " + protocolName + " ... ", 1);
				retValue = awaitStat(bananas, 180 * 1000, false);
				Rcc.out.terminalPrintln((retValue == null ? "[Timeout]" : (retValue ? " [OK]" : " [Error]")), 0);

				if (startAfterCompilation) {
					// Start protocol
					for (BananaInformation banana : bananas) {
						Rcc.out.bananaOutPrintln(banana, "> " + "bpi pstart " + protocolName);
						banana.sendMessage("pstart " + protocolName + " " + getCmdId());
					}
					Rcc.out.terminalPrint("Starting " + protocolName + " ... ", 1);
					retValue = awaitStat(bananas, 5 * 1000, false);
					Rcc.out.terminalPrintln((retValue == null ? "[Timeout]" : (retValue ? " [OK]" : " [Error]")), 0);
				}

				Rcc.ctrlMsgListener.deregisterCommandThread(getCmdId(), this);
			}
		}
	}

	@Override
	public boolean validate() {
		return true;
	}
}
