package configgui.java.commands.simple;

import java.util.Vector;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;

import configgui.java.Rcc;
import configgui.java.commands.Command;
import configgui.java.model.configuration.BananaInformation;
import configgui.java.util.network.ControlMsgListener;
import configgui.java.util.terminal.GeneralOutputInterface;
import javafx.application.Platform;
import javafx.collections.ObservableList;

public class SendRemoteCommand extends Command {

	private String command;
	private ObservableList<BananaInformation> bananas;

	public SendRemoteCommand(String cmd, ObservableList<BananaInformation> bananas) {
		command = cmd;
		this.bananas = bananas;
	}

	@Override
	protected void executeCommand() {
		if (responseBufferCounter == null)
			responseBufferCounter = new Semaphore(0);
		if (responseBuffer == null)
			responseBuffer = new Vector<>();

		Vector<String> busyBananaList = new Vector<>();
		Vector<BananaInformation> busyBananaInfoList = new Vector<>();

		Rcc.ctrlMsgListener.registerCommandThread(getCmdId(), this);

		Rcc.out.terminalPrintln("Sending remote command \"" + command + "\" to nodes", 1);
		for (BananaInformation banana : bananas) {
			if (banana.getState() != BananaInformation.state_disconnected) {
				Rcc.out.bananaOutPrintln(banana, "> " + command);
				banana.sendMessage(command + " " + getCmdId());
				busyBananaList.add(banana.getIP());
				busyBananaInfoList.add(banana);
			}
		}
		for (BananaInformation banana : busyBananaInfoList) {
			Rcc.out.setState(banana, BananaInformation.state_busy);
		}
		long timeout = System.currentTimeMillis() + 120 * 1000;
		Rcc.out.activateProgressBar();
		int totalNumberOfBananas = busyBananaList.size();
		Rcc.out.updateProgressBar(0, totalNumberOfBananas);
		while (System.currentTimeMillis() < timeout && busyBananaList.size() > 0) {
			try {
				if (responseBufferCounter.tryAcquire(timeout - System.currentTimeMillis(), TimeUnit.MILLISECONDS)) {
					String[] msg = responseBuffer.get(0);
					responseBuffer.remove(0);
					String[] items = msg[1].split(";");

					int index = busyBananaList.indexOf(msg[0]);
					if (index >= 0) {

						switch (items[2]) {
						case "\"error\"":
							Rcc.out.setState(busyBananaInfoList.get(index), BananaInformation.state_error);
							busyBananaInfoList.remove(index);
							busyBananaList.remove(index);
							Rcc.out.updateProgressBar(totalNumberOfBananas - busyBananaList.size(), busyBananaList.size());
							break;
						case "\"fail\"":
							Rcc.out.setState(busyBananaInfoList.get(index), BananaInformation.state_error);
							busyBananaInfoList.remove(index);
							busyBananaList.remove(index);
							Rcc.out.updateProgressBar(totalNumberOfBananas - busyBananaList.size(), busyBananaList.size());
							break;
						case "\"success\"":
							Rcc.out.setState(busyBananaInfoList.get(index), BananaInformation.state_idle);
							busyBananaInfoList.remove(index);
							busyBananaList.remove(index);
							Rcc.out.updateProgressBar(totalNumberOfBananas - busyBananaList.size(), busyBananaList.size());
							break;
						}
					}
				}
			} catch (Exception e) {
				Rcc.out.addException(e);
			}
		}
		Rcc.out.unlockTerminal();
		Rcc.ctrlMsgListener.deregisterCommandThread(getCmdId(), this);
	}

	@Override
	public boolean validate() {
		if (Rcc.out == null || bananas == null)
			return false;
		if (command == null) {
			Rcc.out.terminalPrint("Error: no remote command to be sent.");
			return false;
		}
		return true;
	}
}
