package configgui.java.commands.simple;

import java.util.Optional;
import java.util.concurrent.Semaphore;

import configgui.java.Rcc;
import configgui.java.commands.Command;
import javafx.application.Platform;
import javafx.scene.control.TextInputDialog;

public class SetOptionCommand extends Command {

	private String key;
	private String stringValue;

	public SetOptionCommand( String opt, String val) {
		key = opt;
		stringValue = val;
	}

	@Override
	protected void executeCommand() {
		if (stringValue == null) {
			Semaphore s = new Semaphore(0);
			Platform.runLater(() -> {
				String value = openValueDialog(key);
				if (value == null || value.isEmpty()) {
					Rcc.out.terminalPrintln("Please specify a value", 1);
				}
				Rcc.configuration.setConfigEntry(key, value);
				Rcc.out.terminalPrintln("Configuration saved: \"" + key + "\": \"" + value + "\"", 1);
				s.release();
			});
			try {
				s.acquire();
			} catch (Exception e) {
				Rcc.out.addException(e);
			}
		} else {
			Rcc.out.terminalPrintln("Configuration saved: \"" + key + "\": \"" + stringValue + "\"", 1);
			Rcc.configuration.setConfigEntry(key, stringValue);
		}
	}

	@Override
	public boolean validate() {
		if (Rcc.out == null || Rcc.configuration == null || key == null)
			return false;
		return true;
	}

	private String openValueDialog(String key) {
		TextInputDialog dialog = new TextInputDialog(Rcc.configuration.getConfigEntry(key));
		dialog.setTitle("Set Value for \"" + key + "\"");
		dialog.setHeaderText("Specify a value for the option \"" + key + "\"");
		dialog.setContentText("Value:");

		Optional<String> result = dialog.showAndWait();
		return result.orElse(null);
	}

}
