package configgui.java.commands.simple;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;

import configgui.java.Rcc;
import configgui.java.commands.Command;
import configgui.java.model.configuration.BananaInformation;
import configgui.java.time.TimeC;
import javafx.collections.ObservableList;

public class StartClockSyncCommand extends Command{
	ObservableList<BananaInformation> bananas;
	
	public StartClockSyncCommand(ObservableList<BananaInformation> bananas) {
		this.bananas = bananas;
	}
//	
//	public StartClockSyncCommand() {
//	
//	}
	@Override
	protected void executeCommand() {
		// TODO Auto-generated method stub
		Boolean retValue;
		String output= null;
		String mode = Rcc.configuration.getConfigEntry("clocksyncmode");
		Rcc.out.terminalPrintln("Start Clock Synchronisation in Mode: " + mode , 1);
		

		for (BananaInformation banana : bananas) {
			//Set Config
			Rcc.out.bananaOutPrintln(banana, "> " + "bpi setconfig clocksync "+ mode);
			banana.sendMessage("setconfig clocksync "+ mode  + " " + getCmdId());
		
			//Kill NTP
			Rcc.out.bananaOutPrintln(banana, "> " + "bpi ssh service ntp stop");
			banana.sendMessage("ssh service ntp stop"+ " " + getCmdId());				
			
			//Kill PTP
			Rcc.out.bananaOutPrintln(banana, "> " + "bpi ssh killall -e /home/user/Banananetwork/PTP/ptpd2");
			banana.sendMessage("ssh killall -e /home/user/Banananetwork/PTP/ptpd2" + " " + getCmdId());		
			Rcc.out.bananaOutPrintln(banana, "> " + "bpi ssh killall -e /home/user/Banananetwork/PTP/ptpd");
			banana.sendMessage("ssh killall -e /home/user/Banananetwork/PTP/ptpd" + " " + getCmdId());		
		}
		
		//set the time of the RCC Laptop as system time on the BPiR's	
		TimeC sendingTheTime = new TimeC();
		sendingTheTime.gettingCTime();
		long currenttime[] = sendingTheTime.gettingCTime();
		long currenttimeNS = (long) currenttime[0]*1000000000 + currenttime[1];
		for (BananaInformation banana : bananas) {	
			
			Rcc.out.bananaOutPrintln(banana, "> " + "bpi setsystemclock " + currenttimeNS);
			banana.sendMessage("setsystemclock " + currenttimeNS + " " + getCmdId());
		}
		
		switch (mode) {
		
		case "CUSTOM":
			
			
			
		case "SYSTEM":
			
			
			break;
		case "PTP":	
			
			//Copy data to the nodes
			for (BananaInformation banana : bananas) {		
				Rcc.out.bananaOutPrintln(banana, "> " + "put PTP");
			}
			new UploadCommand(bananas, "PTP").exec().waitForTermination();
			
			//change permissions to execute ptpd
			for (BananaInformation banana : bananas) {	
				Rcc.out.bananaOutPrintln(banana, "> " + "bpi ssh chmod 744 /home/user/Banananetwork/PTP/ptpd");
				Rcc.out.bananaOutPrintln(banana, "> " + "bpi ssh chmod 744 /home/user/Banananetwork/PTP/ptpd2");
				banana.sendMessage("ssh chmod 744 /home/user/Banananetwork/PTP/ptpd"+ " " + getCmdId());
				banana.sendMessage("ssh chmod 744 /home/user/Banananetwork/PTP/ptpd2"+ " " + getCmdId());
			}
			//pause for the permission change to take effect
			try {
				Thread.sleep(1000);
			} catch (Exception e) {
				Rcc.out.addException(e);
			}
			
			//start PTP
			for (BananaInformation banana : bananas){
				Rcc.out.bananaOutPrintln(banana, "> " + "bpi ssh /home/user/Banananetwork/PTP/ptpd2 -c /home/user/Banananetwork/PTP/bpi_ptp2.conf");
				banana.sendMessage("ssh /home/user/Banananetwork/PTP/ptpd2 -c /home/user/Banananetwork/PTP/bpi_ptp2.conf"+ " " + getCmdId());			
//				Rcc.out.bananaOutPrintln(banana, "> " + "bpi ssh /home/user/Banananetwork/PTP/ptpd -c /home/user/Banananetwork/PTP/bpi_ptp.conf");
//				banana.sendMessage("ssh /home/user/Banananetwork/PTP/ptpd -c /home/user/Banananetwork/PTP/bpi_ptp.conf"+ " " + getCmdId());			
			}
			
			//done
			Rcc.out.terminalPrintln("Clock Synchronisation is now active!", 1);
			break;
			
		case "NTP":
			
			//choose the NTP-Server...
			BananaInformation bpi_server; 
			try{
			bpi_server = 	bananas.get(0);
			}catch(Exception e){
				Rcc.out.addException(e);
				break;
			}
			List<String> newLines = new ArrayList<>();
			try {
				for (String line : Files.readAllLines(Paths.get("/home/user/Banananetwork/NTP/ntp_client.conf"),
						StandardCharsets.UTF_8)) {
					if (line.contains("#BpiServerIP")) {
						newLines.add(line.replace(line , "server " + bpi_server.getIP() + " #BpiServerIP"));
					} else {
						newLines.add(line);
					}
				}
			} catch (IOException e) {
				Rcc.out.addException(e);
			}
			try {
				Files.write(Paths.get("/home/user/Banananetwork/NTP/ntp_client.conf"), newLines,
						StandardCharsets.UTF_8);
			} catch (IOException e) {
				Rcc.out.addException(e);
			}
			
			// transfer config files to BPiR's
			new UploadCommand(bananas, "NTP").exec().waitForTermination();
			
			//Change ntp.conf based on server or client node
			for (BananaInformation banana : bananas) {
				String conf= (banana.getIP().compareTo(bpi_server.getIP())==0 )? "ntp_server.conf" : "ntp_client.conf";
				banana.sendMessage("ssh cp /home/user/Banananetwork/NTP/"+conf + " /etc/ntp.conf" +  " " + getCmdId());
			}
			
			//start ntp service on all nodes
			for (BananaInformation banana : bananas) {
				banana.sendMessage("ssh service ntp start"+  " " + getCmdId());
			}
					
			break;

		default:
			
			break;
		}
	}

	@Override
	protected boolean validate() {
		return Rcc.out != null;
	}

}
