package configgui.java.controller;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileWriter;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.Optional;
import java.util.Vector;
import java.util.concurrent.Semaphore;

import configgui.fi_jn.DistributionCommand;
import configgui.fi_jn.FaultInjectionController;
import configgui.java.Rcc;
import configgui.java.commands.Command;
import configgui.java.commands.chains.DiscoverTopologyCommand;
import configgui.java.commands.chains.UpdateLib;
import configgui.java.commands.chains.UpdateProtocol;
import configgui.java.commands.simple.GetOptionCommand;
import configgui.java.commands.simple.HelpCommand;
import configgui.java.commands.simple.RunExternalBashScript;
import configgui.java.commands.simple.ScanCommand;
import configgui.java.commands.simple.SendRemoteCommand;
import configgui.java.commands.simple.SetOptionCommand;
import configgui.java.commands.simple.StartClockSyncCommand;
import configgui.java.commands.simple.UploadCommand;
import configgui.java.model.cells.BananaInformationListViewCell;
import configgui.java.model.configuration.BananaInformation;
import configgui.java.util.tab.DockingTabContent;
import configgui.java.util.terminal.BpiOutputModule;
import configgui.java.util.terminal.GeneralOutputInterface;
import configgui.java.util.terminal.Terminal;
import javafx.application.Platform;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.collections.ListChangeListener;
import javafx.collections.ObservableList;
import javafx.event.EventHandler;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.geometry.Insets;
import javafx.scene.Node;
import javafx.scene.Parent;
import javafx.scene.Scene;
import javafx.scene.control.Alert;
import javafx.scene.control.Alert.AlertType;
import javafx.scene.control.Button;
import javafx.scene.control.ButtonBar;
import javafx.scene.control.ButtonBar.ButtonData;
import javafx.scene.control.ButtonType;
import javafx.scene.control.CheckBox;
import javafx.scene.control.Dialog;
import javafx.scene.control.Label;
import javafx.scene.control.ListView;
import javafx.scene.control.Menu;
import javafx.scene.control.MenuItem;
import javafx.scene.control.ProgressBar;
import javafx.scene.control.RadioButton;
import javafx.scene.control.SelectionMode;
import javafx.scene.control.SplitPane;
import javafx.scene.control.TabPane;
import javafx.scene.control.TextField;
import javafx.scene.control.ToggleGroup;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyEvent;
import javafx.scene.layout.ColumnConstraints;
import javafx.scene.layout.GridPane;
import javafx.scene.layout.RowConstraints;
import javafx.scene.layout.VBox;
import javafx.stage.Stage;

public class BananaController {
	@FXML
	private Label uiStatusBar;
	@FXML
	private ProgressBar progressBar;
	@FXML
	private MenuItem fileConfig;
	@FXML
	private MenuItem fileExit;
	@FXML
	private MenuItem viewOutputGrid;
	@FXML
	private Menu runMenu;
	@FXML
	private MenuItem runRefresh;
	@FXML
	private MenuItem runFind;
	@FXML
	private TextField ipfield;
	@FXML
	private ListView<BananaInformation> connectedNodeList;
	@FXML
	private SplitPane ttSplitPane;

	/*
	 * Reference to the main class
	 */
	private Rcc main;

	/*
	 * Primary components of the UI, sonsting of a Terminal and a TabPane.
	 */
	private Terminal terminal;
	private TabPane ovTabPane;
	private boolean mouseDragOnDivider = false;
	private double terminalWidthPortion = 0.4;

	/*
	 * Container for Script references
	 */
	ArrayList<Menu> scriptList;

	/*
	 * Tab 1: Detailed output for all connected BananaPi clients.
	 */
	private DockingTabContent bpiIndividualOutput = null;
	private GridPane fxGridBpiOut;
//	private BpiOutputModule[][] bpiOutGrid;
	private ArrayList<BpiOutputModule> bpiOutModContainer;
	private int lastRows = 1, lastColumns = 1;

	/*
	 * Tab 2: Exception Tab
	 */
	private DockingTabContent excTab = null;
	private ExceptionTabController exceptionTabController;
	private int newExceptions = 0;

	FabanCompanionController fabanCompanionController;

	/*
	 * Tab 4: Fi Tab
	 */
	FaultInjectionController faultInjectionController;
	private DockingTabContent fiTab = null;

	/*
	 * Separate Scene 1: Configuration window
	 */
	private Scene configWindowScene;
	private ConfigWindowController configWindowCtrl;

	@FXML
	private void initialize() {
		scriptList = new ArrayList<>();

		// Ui tweaks
		uiStatusBar.setMaxWidth(Double.MAX_VALUE);
		connectedNodeList.getSelectionModel().setSelectionMode(SelectionMode.MULTIPLE);
		connectedNodeList.setCellFactory(lv -> new BananaInformationListViewCell());

		// Setting up MenuBar
		fileExit.setOnAction((event) -> System.exit(0));
		fileConfig.setOnAction((event) -> showConfiguration());
		viewOutputGrid.setOnAction((event) -> selectGridSize());
		runRefresh.setOnAction((event) -> runRefresh());
		runFind.setOnAction((event) -> runFind());

		// Setting up ip field logic
		ipfield.setOnKeyPressed(new EventHandler<KeyEvent>() {
			@Override
			public void handle(KeyEvent ke) {
				if (ke.getCode().equals(KeyCode.ENTER)) {
					scanButtonPressed();
				}
			}
		});

		// Setting up terminal
		terminal = new Terminal();
		ttSplitPane.getItems().add(terminal);
		terminal.setController(this);
		terminal.setCompleteCmds(getAutocomplitionCommandList());

		// Setting up Tab Container
		ovTabPane = new TabPane();
		ttSplitPane.getItems().add(ovTabPane);
		ttSplitPane.setDividerPositions(terminalWidthPortion);
		ttSplitPane.getDividers().get(0).positionProperty().addListener((obs, oldVal, newVal) -> terminalWidthPortion = newVal.doubleValue());

		ovTabPane.setOnKeyPressed(e -> {
			if (e.getCode().equals(KeyCode.F5)) {
				for (BpiOutputModule bpimod : bpiOutModContainer) {
					bpimod.refresh();
				}
			}
		});

		// Tab: BpPi Output
		bpiIndividualOutput = new DockingTabContent(ovTabPane, "BPi Ouput");
		bpiOutModContainer = new ArrayList<>();

		// Tab: Exceptions
		try {
			FXMLLoader loader = new FXMLLoader();
			loader.setLocation(getClass().getResource("../../view/subscenes/ExceptionTab.fxml"));
			Parent root = loader.load();
			exceptionTabController = loader.getController();
			excTab = new DockingTabContent(ovTabPane, "Exceptions", root);
		} catch (Exception e) {
			e.printStackTrace();
		}

		ovTabPane.getSelectionModel().selectedIndexProperty().addListener(new ChangeListener<Number>() {
			@Override
			public void changed(ObservableValue<? extends Number> ov, Number oldValue, Number newValue) {
				if (ovTabPane.getTabs().get(newValue.intValue()) == excTab.getTab())
					newExceptions = 0;
				Platform.runLater(() -> {
					excTab.getTab().setStyle(null);
					excTab.getTab().setLabel("Exceptions");
				});
			}
		});

		// Tab: Faban Companion
		try {
			FXMLLoader loader = new FXMLLoader();
			loader.setLocation(getClass().getResource("../../view/subscenes/FabanCompanion.fxml"));
			Parent root = loader.load();
			fabanCompanionController = loader.getController();
			DockingTabContent fabanCompanionTab = new DockingTabContent(ovTabPane, "Faban Companion", root);
		} catch (Exception e) {
			e.printStackTrace();
		}

		// Tab: Fault Injection TODO
		try {
			FXMLLoader loader = new FXMLLoader();
			loader.setLocation(getClass().getResource("../../fi_jn/FI_JN.fxml"));
			Parent root = loader.load();
			faultInjectionController = loader.getController();
			DockingTabContent fiTab = new DockingTabContent(ovTabPane, "Fault Injection", root);
		} catch (Exception e) {
			e.printStackTrace();
		}

	}

	public void initController(Rcc main, ArrayList<Exception> exceps) {
		this.main = main;
		ipfield.setText(Rcc.configuration.getTargetNetwork());
		fabanCompanionController.setMain(main);
		faultInjectionController.setMain(main);
		connectedNodeList.setItems(Rcc.connectedNodes);

		String c = Rcc.configuration.getConfigEntry("columns");
		String r = Rcc.configuration.getConfigEntry("rows");
		r = (r == null ? "1" : r);
		c = (c == null ? "1" : c);

		generateBpiOutputGrid(Rcc.connectedNodes, Integer.parseInt(r), Integer.parseInt(c));

		Rcc.connectedNodes.addListener(new ListChangeListener<BananaInformation>() {
			@Override
			public void onChanged(Change<? extends BananaInformation> c) {
				while (c.next())
					if (c.getAddedSize() > 0 || c.getRemovedSize() > 0 || c.wasPermutated()) {
						connectedNodeList.getSelectionModel().selectAll();
						generateBpiOutputGrid(Rcc.connectedNodes, lastRows, lastColumns);
					}
			}
		});

		ttSplitPane.getDividers().get(0).positionProperty().addListener(d -> {
			for (BpiOutputModule bpimod : bpiOutModContainer) {
				bpimod.refresh();
			}
		});

		main.getPrimaryStage().widthProperty().addListener((obs, oldVal, newVal) -> {
			ttSplitPane.setDividerPositions(terminalWidthPortion);
			Rcc.configuration.setConfigEntry("windowsize", (int) main.getPrimaryStage().getWidth() + "x" + (int) main.getPrimaryStage().getHeight());
		});

		main.getPrimaryStage().heightProperty().addListener((obs, oldVal, newVal) -> {
			ttSplitPane.setDividerPositions(terminalWidthPortion);
			Rcc.configuration.setConfigEntry("windowsize", (int) main.getPrimaryStage().getWidth() + "x" + (int) main.getPrimaryStage().getHeight());
		});

		main.getPrimaryStage().xProperty().addListener((obs, oldVal, newVal) -> {
			Rcc.configuration.setConfigEntry("position", (int) main.getPrimaryStage().xProperty().get() + "," + (int) main.getPrimaryStage().yProperty().get());
		});

		main.getPrimaryStage().yProperty().addListener((obs, oldVal, newVal) -> {
			Rcc.configuration.setConfigEntry("position", (int) main.getPrimaryStage().xProperty().get() + "," + (int) main.getPrimaryStage().yProperty().get());
		});

		ttSplitPane.requestLayout();
		ttSplitPane.applyCss();
		for (Node node : ttSplitPane.lookupAll(".split-pane-divider")) {
			node.setOnDragDetected(evMousePressed -> mouseDragOnDivider = true);
			node.setOnMouseReleased(evMouseReleased -> mouseDragOnDivider = false);
		}

		Rcc.out = new GeneralOutputInterface(terminal, main.getBananaController());
		new Thread(Rcc.out).start();
		Rcc.out.terminalPrintln("Type 'help' to see the options \n\n" + "Initial scan ...", 1);

		// Initial command execution
		new ScanCommand(Rcc.connectedNodes).exec();

		for (Exception excep : exceps)
			if (excep != null)
				exceptionTabController.addException(excep);

		runRefresh();
	}

	public Label getStatusTextField() {
		return uiStatusBar;
	}

	public ProgressBar getProgressBar() {
		return progressBar;
	}

	public ExceptionTabController getExceptionTabController() {
		return exceptionTabController;
	}

	public void reportNewException() {
		newExceptions++;
		Platform.runLater(() -> {
			if (ovTabPane.getSelectionModel().getSelectedItem() != excTab.getTab()) {
				excTab.getTab().setStyle("-fx-background-color: rgb(255,148,148);");
				excTab.getTab().setLabel("Exceptions (" + newExceptions + ")");
			}
		});

	}

	private void initConfigWindow() {
		try {
			FXMLLoader loader = new FXMLLoader();
			loader.setLocation(getClass().getResource("../../view/subscenes/ConfigurationWindow.fxml"));
			Parent root = loader.load();
			configWindowCtrl = loader.getController();
			configWindowScene = new Scene(root, 640, 640);
		} catch (Exception e) {
			Rcc.out.addException(e);
		}

	}

	private void showConfiguration() {
		if (configWindowScene == null)
			initConfigWindow();
		Stage cStage = new Stage();
		cStage.setTitle("Configuration");
		cStage.setScene(configWindowScene);
		cStage.show();
		configWindowCtrl.refresh();
	}

	private void selectGridSize() {
		Stage gridselection = new Stage();
		gridselection.setTitle("Set Grid Size...");
		VBox root = new VBox();
		GridPane gr = new GridPane();
		root.setPadding(new Insets(24, 12, 24, 12));
		root.setSpacing(8);
		TextField rNumber = new TextField(Rcc.configuration.getConfigEntry("rows"));
		TextField cNumber = new TextField(Rcc.configuration.getConfigEntry("columns"));
		RadioButton rBtn = new RadioButton("Save additionally to config file");
		RadioButton autoGrid = new RadioButton("Automatically detect best alignment");
		rBtn.scaleXProperty().set(0.8);
		rBtn.scaleYProperty().set(0.8);
		autoGrid.scaleXProperty().set(0.8);
		autoGrid.scaleYProperty().set(0.8);

		autoGrid.setOnAction(event -> {
			boolean clicked = autoGrid.isSelected();
			rNumber.setDisable(clicked);
			cNumber.setDisable(clicked);
		});

		ButtonBar btnBar = new ButtonBar();
		Button btnCancel = new Button("Cancel");
		btnCancel.setOnAction((event) -> gridselection.close());
		Button btnConfirm = new Button("Confirm");
		btnConfirm.setOnAction((event) -> {
			try {
				int r = -1, c = -1;
				if (!autoGrid.isSelected()) {
					r = Integer.parseInt(rNumber.getText());
					c = Integer.parseInt(cNumber.getText());
				}

				generateBpiOutputGrid(Rcc.connectedNodes, r, c);
				if (rBtn.isSelected()) {
					Rcc.configuration.setConfigEntry("rows", "" + r);
					Rcc.configuration.setConfigEntry("columns", "" + c);
					Rcc.out.statusSetText("Grid configuration saved!");
				}
				gridselection.close();
			} catch (Exception e) {
				Alert alert = new Alert(AlertType.ERROR);
				alert.setTitle("Error");
				alert.setHeaderText("Could not parse number from input");
				alert.showAndWait();
			}
		});
		btnBar.getButtons().addAll(btnCancel, btnConfirm);

		gr.add(new Label("#rows:  "), 0, 0);
		gr.add(rNumber, 1, 0);
		gr.add(new Label("#columns:  "), 0, 1);
		gr.add(cNumber, 1, 1);
		gr.setVgap(8);
		gr.setHgap(16);
		root.getChildren().add(gr);
		root.getChildren().add(rBtn);
		root.getChildren().add(autoGrid);
		root.getChildren().add(btnBar);
		Scene scene = new Scene(root, 300, 200);
		gridselection.setScene(scene);
		gridselection.show();
	}

	private void runRefresh() {
		String p = Rcc.configuration.getConfigEntry("scriptpath");

		runMenu.getItems().removeAll(scriptList);
		scriptList.clear();

		File scriptfolder = new File(p);

		if (Rcc.configuration.getConfigEntry("ipfiles") == null) {
			System.out.println("Set at least one ipfiles entry in configuration");
			return;
		}

		if (scriptfolder.exists()) {
			for (File file : scriptfolder.listFiles()) {
				if (file.getName().contains(".sh")) {
					Menu m = new Menu(file.getName().replace('_', ' ').replace(".sh", ""));
					for (String ipfile : Rcc.configuration.getConfigEntry("ipfiles").split(",")) {
						MenuItem item = new MenuItem(ipfile);
						m.getItems().add(item);
						item.setOnAction((event) -> (new RunExternalBashScript(file, ipfile)).exec());
					}
					scriptList.add(m);

				}
			}
			scriptList.sort(new Comparator<Menu>() {
				@Override
				public int compare(Menu o1, Menu o2) {
					return o1.getText().compareTo(o2.getText());
				}
			});
			runMenu.getItems().addAll(0, scriptList);
		}
	}

	private void runFind() {

		Rcc.out.getTerminal().deleteInput();
		Rcc.out.terminalPrintln("Blind client search started");
		ArrayList<String> foundNetworks = new ArrayList<>();
		// Find available networks
		try {
			ProcessBuilder pb = new ProcessBuilder("ifconfig");
			pb.directory(new File("/home/user/"));

			Process p = pb.start();
			BufferedReader out = new BufferedReader(new InputStreamReader(p.getInputStream()));
			String line = null;

			while ((line = out.readLine()) != null) {
				if (line.contains("inet ") && line.contains("netmask") && line.contains("broadcast")) {
					for (final String l : line.split("  ")) {
						if (l.contains("inet")) {
							String[] items = l.split(" ");
							if (items.length == 2) {
								String[] ipItems = items[1].split("\\.");
								if (ipItems.length == 4)
									foundNetworks.add(String.format("%03d.%03d.%03d.%s", Integer.parseInt(ipItems[0]), Integer.parseInt(ipItems[1]), Integer.parseInt(ipItems[2]), "xxx"));
							}
						}
					}
				}
			}
		} catch (Exception e) {
			Rcc.out.addException(e);
		}

		// Show a network choice dioalog box
		Dialog<String> networkChoiceDialog = new Dialog<>();
		networkChoiceDialog.setTitle("Subnet Selection");
		networkChoiceDialog.setHeaderText("Where to search for nodes?");
		ButtonType searchButtonType = new ButtonType("Search", ButtonData.OK_DONE);

		VBox choiceVbox = new VBox();
		choiceVbox.setSpacing(10);
		choiceVbox.setPadding(new Insets(10, 70, 10, 70));
		ToggleGroup tg = new ToggleGroup();
		if (foundNetworks.size() > 0) {
			networkChoiceDialog.getDialogPane().getButtonTypes().addAll(searchButtonType, ButtonType.CANCEL);
			RadioButton[] buttons = new RadioButton[foundNetworks.size()];
			int i = 0;
			for (String ip : foundNetworks) {
				buttons[i] = new RadioButton(ip);
				buttons[i++].setToggleGroup(tg);
			}
			choiceVbox.getChildren().addAll(buttons);
			networkChoiceDialog.getDialogPane().setContent(choiceVbox);
			buttons[0].setSelected(true);
			networkChoiceDialog.setResultConverter(dialogButton -> {
				if (dialogButton == searchButtonType) {
					return ((RadioButton) tg.getSelectedToggle()).getText();
				}
				return "";
			});
		} else {

			networkChoiceDialog.getDialogPane().getButtonTypes().addAll(ButtonType.CANCEL);
			choiceVbox.getChildren().add(new Label("No networks found"));
			networkChoiceDialog.getDialogPane().setContent(choiceVbox);
			networkChoiceDialog.setResultConverter(dialogButton -> {
				return "";
			});
		}

		String chosenNetwork = "";
		try {
			chosenNetwork = networkChoiceDialog.showAndWait().get();
		} catch (Exception e) {
			Rcc.out.addException(e);
		}

		String[] network = chosenNetwork.split("\\.");
		// Abort
		if (network.length != 4) {
			Rcc.out.terminalPrintln("No network chosen or no network available. Abort...", 1);
			Rcc.out.startCommand();
		}
		// Search
		else {
			final String net = chosenNetwork;
			// Create Joiner Thread
			Thread multiping = new Thread(new Runnable() {
				@Override
				public void run() {
					Rcc.out.terminalPrintln("Start searching clients in subnet " + net, 1);
					String subnet = String.format("%s.%s.%s.", network[0], network[1], network[2]);
					Semaphore pingThreads = new Semaphore(0);
					Semaphore maxThreads = new Semaphore(300);
					Vector<String> foundNodes = new Vector<>();
					Platform.runLater(() -> {
						Rcc.out.activateProgressBar();
						Rcc.out.updateProgressBar(0, 254);
					});
					for (int i = 1; i <= 254; i++) {
						try {
							maxThreads.acquire();
							final int j = i;
							new Thread(new Runnable() {
								@Override
								public void run() {
									try {
										String addr = subnet + j;
										ProcessBuilder pb = new ProcessBuilder("ping", "-w 3", addr);
										pb.directory(new File("/home/user/"));

										Process p = pb.start();
										BufferedReader out = new BufferedReader(new InputStreamReader(p.getInputStream()));
										String line = null;

										StringBuffer s = new StringBuffer();
										while ((line = out.readLine()) != null) {
											s.append(line + "\n");
										}
										if (!s.toString().contains("100% packet loss")) {
											foundNodes.add(addr);
											Platform.runLater(() -> Rcc.out.terminalPrintln("Found " + addr, 2));
										}
									} catch (Exception e) {
										Rcc.out.addException(e);
									}
									pingThreads.release();
									maxThreads.release();
									Platform.runLater(() -> Rcc.out.updateProgressBar(pingThreads.availablePermits(), 254));
								}
							}).start();
						} catch (Exception e) {
							Rcc.out.addException(e);
						}
					}
					try {
						pingThreads.acquire(254);
					} catch (Exception e) {
						Rcc.out.addException(e);
					}

					// Save to ipfile dialog
					Platform.runLater(() -> {
						if (foundNodes.size() > 0) {
							Dialog<Boolean> ipSelectionDialog = new Dialog<>();
							ipSelectionDialog.setTitle("Save Ip File");
							ipSelectionDialog.setHeaderText("Select adresses which shall be written to the specified destination file");
							ButtonType okButtonType = new ButtonType("Ok", ButtonData.OK_DONE);

							VBox ipContainer = new VBox();
							ipContainer.setPadding(new Insets(10, 200, 30, 200));
							ipContainer.setSpacing(10);
							VBox dialogContainer = new VBox();
							dialogContainer.setPadding(new Insets(20, 20, 20, 20));
							dialogContainer.setSpacing(5);

							ipSelectionDialog.getDialogPane().getButtonTypes().addAll(okButtonType, ButtonType.CANCEL);
							CheckBox[] ipCheckBox = new CheckBox[foundNodes.size()];
							int i = 0;
							for (String ip : foundNodes) {
								ipCheckBox[i] = new CheckBox(ip);
								ipCheckBox[i++].setSelected(true);
							}
							ipContainer.getChildren().addAll(ipCheckBox);
							Label saveLabel = new Label("Save selected IPs to following file:");
							TextField savePath = new TextField(Rcc.configuration.getConfigEntry("scriptpath") + "/newipfile");

							savePath.setPrefWidth(500);
							dialogContainer.getChildren().add(ipContainer);
							dialogContainer.getChildren().add(saveLabel);
							dialogContainer.getChildren().add(savePath);
							ipSelectionDialog.getDialogPane().setContent(dialogContainer);
							ipSelectionDialog.setResultConverter(dialogButton -> {
								if (dialogButton == okButtonType) {
									String filename = savePath.getText();
									try {
										FileWriter fw = new FileWriter(filename);
										PrintWriter pw = new PrintWriter(fw);
										for (CheckBox ipCb : ipCheckBox) {
											if (ipCb.isSelected())
												pw.println("root@" + ipCb.getText());
										}
										pw.close();
										String[] filenameSplitted = filename.split("/");
										String ipfilename = filenameSplitted[filenameSplitted.length - 1];
										if (Rcc.configuration.getConfigEntry("ipfiles") == null) {
											Rcc.configuration.setConfigEntry("ipfiles", ipfilename);
										} else {
											if (!Rcc.configuration.getConfigEntry("ipfiles").contains(ipfilename)) {
												Rcc.configuration.setConfigEntry("ipfiles", Rcc.configuration.getConfigEntry("ipfiles") + "," + ipfilename);
											}
										}
									} catch (Exception e) {
										Platform.runLater(() -> Rcc.out.terminalPrintln("Error occured while writing to file: " + filename, 1));
										return false;
									}
									return true;
								}
								return false;
							});

							try {
								if (!ipSelectionDialog.showAndWait().get())
									Platform.runLater(() -> Rcc.out.terminalPrintln("Nothing saved", 1));
								else
									Platform.runLater(() -> Rcc.out.terminalPrintln("Successfully saved selected IPs to file", 1));
							} catch (Exception e) {
								Rcc.out.addException(e);
							}

							Platform.runLater(() -> Rcc.out.startCommand());
						}
					});

				}
			});
			multiping.start();
		}

	}

	private ArrayList<String> getAutocomplitionCommandList() {
		ArrayList<String> completeCmds = new ArrayList<String>();
		for (String s : Command.COMMANDLIST) {
			completeCmds.add(s);
		}
		completeCmds.sort(new Comparator<String>() {
			public int compare(String o1, String o2) {
				return o1.compareTo(o2);
			};
		});
		return completeCmds;
	}

	public void generateBpiOutputGrid(ObservableList<BananaInformation> bananas, int rows, int columns) {

		lastRows = rows;
		lastColumns = columns;

		if (rows == 1 || columns == -1) {
			int N = Rcc.connectedNodes.size();
			int x = (int) Math.sqrt(N);
			if ((x - 1) * (x - 1 + 3) >= N)
				rows = columns = Math.max(1, x - 1);
			else
				rows = columns = x;

			while (rows * columns < N)
				rows++;
		}

		fxGridBpiOut = new GridPane();

		for (int n = 0; n < columns; n++) {
			ColumnConstraints c = new ColumnConstraints();
			c.setPercentWidth(100.0 / columns);
			fxGridBpiOut.getColumnConstraints().add(c);
		}
		for (int m = 0; m < rows; m++) {
			RowConstraints c = new RowConstraints();
			c.setPercentHeight(100.0 / rows);
			fxGridBpiOut.getRowConstraints().add(c);
		}

		int newBpiOutMods = rows * columns - bpiOutModContainer.size();

		for (int i = 0; i < newBpiOutMods; i++) {
			bpiOutModContainer.add(new BpiOutputModule(bananas, 1));
		}

		for (int n = 0; n < columns; n++) {
			for (int m = 0; m < rows; m++) {
				if (n + m * columns < Rcc.connectedNodes.size()) {
					bpiOutModContainer.get(n + m * columns).selectedNode.getSelectionModel().select(n + m * columns);
					fxGridBpiOut.add(bpiOutModContainer.get(n + m * columns).detPair, n, m, 1, 1);
				}
			}
		}
		bpiIndividualOutput.refreshContent(fxGridBpiOut);
	}

	public void dispatchCommand(String cmd) {
		if (cmd == null)
			return;
		String[] array = cmd.split(" ", 2);
		String command = array[0];
		String param = null;
		if (array.length > 1 && !array[1].isEmpty())
			param = array[1];
		switch (command.toLowerCase()) {
		case "":
			Rcc.out.startCommand();
			break;

		case "bpi":
			new SendRemoteCommand(param, connectedNodeList.getSelectionModel().getSelectedItems()).exec();
			break;

		case "scan":
			if (param != null) {
				new SetOptionCommand("targetNetwork", param).exec();
			}
			Command scan = new ScanCommand(Rcc.connectedNodes);
			scan.exec();
			break;

		case "set":
			if (param == null) {
				Rcc.out.terminalPrintln("Error: please specify an option name.", 1);
				Rcc.out.startCommand();
				return;
			}
			String[] options = param.split(" ", 2);
			String value = null;
			if (options.length > 1)
				value = options[1];
			new SetOptionCommand(options[0].toLowerCase(), value).exec();
			break;

		case "get":
			new GetOptionCommand(param).exec();
			break;

		case "help":
			new HelpCommand(param).exec();
			break;

		case "put":
			new UploadCommand(connectedNodeList.getSelectionModel().getSelectedItems(), param).exec();
			break;
		case "updatelib":
			if (Rcc.connectedNodes.size() > connectedNodeList.getSelectionModel().getSelectedItems().size()) {
				Alert alert = new Alert(AlertType.CONFIRMATION);
				alert.setTitle("Warning");
				alert.setHeaderText("Only a subset of Nodes is selected.");
				alert.setContentText("Continue with upload to the selected subset?");

				Optional<ButtonType> result = alert.showAndWait();
				if (result.get() == ButtonType.OK) {
					new UpdateLib(connectedNodeList.getSelectionModel().getSelectedItems()).exec();
				} else {
					Rcc.out.startCommand();
				}
			} else
				new UpdateLib(connectedNodeList.getSelectionModel().getSelectedItems()).exec();

			break;
		case "update":
			if (array.length != 2 && array.length != 3) {
				Rcc.out.terminalPrint("False parameter", 1);
				Rcc.out.startCommand();
			} else {
				boolean start = false;
				if (array.length == 3 && array[1].equals("-s")) {
					start = true;
				}
				if (Rcc.connectedNodes.size() > connectedNodeList.getSelectionModel().getSelectedItems().size()) {
					Alert alert = new Alert(AlertType.CONFIRMATION);
					alert.setTitle("Warning");
					alert.setHeaderText("Only a subset of Nodes is selected.");
					alert.setContentText("Continue with upload to the selected subset?");

					Optional<ButtonType> result = alert.showAndWait();
					if (result.get() == ButtonType.OK) {
						new UpdateProtocol(connectedNodeList.getSelectionModel().getSelectedItems(), array[array.length - 1], start).exec();
					} else {
						Rcc.out.startCommand();
					}
				} else
					new UpdateProtocol(connectedNodeList.getSelectionModel().getSelectedItems(), array[array.length - 1], start).exec();
			}
			break;
		case "discovertopology":
			new DiscoverTopologyCommand(connectedNodeList.getSelectionModel().getSelectedItems(), fabanCompanionController).exec();
			break;
		case "show_distributions":
			new DistributionCommand(param).exec();
			break;

		/* FIXME JN */
		case "startclocksync":
			new StartClockSyncCommand(connectedNodeList.getItems()).exec();
			break;
		/* END JN */
		default:
			Rcc.out.terminalPrintln("Unknown Command", 1);
			Rcc.out.terminalPrintln("Type 'help' to get an insight into the options.", 1);
			Rcc.out.startCommand();
		}
	}

	public Stage getPrimaryStage() {
		return main.getPrimaryStage();
	}

	@FXML
	private void scanButtonPressed() {
		Rcc.configuration.setTargetNetwork(ipfield.getText());
		new ScanCommand(Rcc.connectedNodes).exec();
	}

}
