package configgui.java.controller;

import java.io.IOException;
import java.util.HashMap;
import java.util.Optional;

import configgui.java.Rcc;
import configgui.java.model.configuration.Configuration;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.Node;
import javafx.scene.Parent;
import javafx.scene.Scene;
import javafx.scene.control.Alert;
import javafx.scene.control.Alert.AlertType;
import javafx.scene.control.Button;
import javafx.scene.control.ButtonType;
import javafx.scene.control.Label;
import javafx.scene.control.ScrollPane;
import javafx.scene.control.TextField;
import javafx.scene.layout.Background;
import javafx.scene.layout.BackgroundFill;
import javafx.scene.layout.GridPane;
import javafx.scene.paint.Color;
import javafx.stage.Stage;

public class ConfigWindowController {

	private int configNumber;

	@FXML
	private GridPane configPane;

	@FXML
	private ScrollPane scrollPane;

	@FXML
	private void initialize() {
		scrollPane.setBackground(new Background(new BackgroundFill(Color.TRANSPARENT, null, null)));
	}

	/**
	 * Gets the saved configuration and displays it with Labels and TextFields. Also
	 * adds "X"-Buttons to delete them.
	 */
	private void setTableContent() {
		Configuration config = Rcc.configuration;
		int i = 0;
		for (String key : config.getMap().keySet()) {
			showConfiguration(key, config.getMap().get(key), i);
			++i;
		}
		configNumber = config.getMap().size();
	}

	/**
	 * Saves the changes to the configurations in the object for the current runtime
	 * and the file.
	 */
	@FXML
	private void save() {
		ObservableList<Node> children = configPane.getChildren();
		HashMap<String, String> configuration = new HashMap<String, String>();
		for (int i = 0; i < children.size(); i = i + 3) {
			configuration.put(((Label) children.get(i)).getText(), ((TextField) children.get(i + 1)).getText());
		}
		Rcc.configuration.setMap(configuration);
		Rcc.out.statusSetText("Coniguration saved!");
	}

	@FXML
	public void refresh() {
		configPane.getChildren().remove(0, configPane.getChildren().size());
		setTableContent();
	}

	/**
	 * Creates a new stage in which you could create a new key-value-pair and add it
	 * to your configuration.
	 */
	@FXML
	private void add() {
		Stage addStage = new Stage();
		addStage.setTitle("Add Configuration");
		FXMLLoader loader = new FXMLLoader();
		loader.setLocation(getClass().getResource("../../view/fragments/CreateConfiguration.fxml"));
		try {
			Parent root = loader.load();
			addStage.setScene(new Scene(root));
		} catch (IOException e) {
			Rcc.out.addException(e);
		}
		addStage.show();
		CreateConfigController controller = loader.getController();
		controller.setStage(addStage);
		controller.setTabController(this);
	}

	/**
	 * Adds - if not already included - the new key-value-pair and if already
	 * existing asks you whether to overwrite it or not.
	 * 
	 * @param key   The new key to add to the configurations.
	 * @param value The new value to add to the configurations.
	 */
	public void addConfiguration(String key, String value) {
		key = key.toLowerCase();
		boolean valid = checkKey(key);
		if (!valid) {
			Alert alert = new Alert(AlertType.CONFIRMATION);
			alert.setTitle("Replace Configuration");
			alert.setHeaderText("The new configuration already exists");
			alert.setContentText("Do you want to replace the old one?");
			Optional<ButtonType> result = alert.showAndWait();
			if (result.get() == ButtonType.OK) {
				ObservableList<Node> nodes = configPane.getChildren();
				for (Node node : nodes) {
					if (node instanceof Label && ((Label) node).getText().equals(key)) {
						((TextField) configPane.getChildren().get(configPane.getChildren().indexOf(node) + 1))
								.setText(value);
					}
				}
				Rcc.out.statusSetText("Configuration " + key + " set to " + value);
			}
		} else {
			showConfiguration(key, value, configNumber);
			++configNumber;
			Rcc.out.statusSetText("Configuration saved: '" + key + "' '" + value + "'");
		}
	}

	/**
	 * Checks whether a key is already existing.
	 * 
	 * @param key The key which is checked.
	 * @return True if the key is not existing yet.
	 */
	private boolean checkKey(String key) {
		boolean valid = true;
		for (Node label : configPane.getChildren()) {
			if (label instanceof Label) {
				if (key.equals(((Label) label).getText())) {
					valid = false;
				}
			}
		}
		return valid;
	}

	private void showConfiguration(String key, String value, int row) {
		configPane.add(new Label(key), 0, row);
		configPane.add(new TextField(value), 1, row);
		Button destroy = new Button("X");
		destroy.getStyleClass().add("small-cross-button");
		configPane.add(destroy, 2, row);
		final int x = row;
		destroy.setOnAction(e -> {
			configPane.getChildren().remove(x * 3);
			configPane.getChildren().remove(x * 3);
			configPane.getChildren().remove(x * 3);
			save();
			configPane.getChildren().remove(0, configPane.getChildren().size());
			setTableContent();
		});
	}
}
