package configgui.java.controller;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonParser;

import configgui.java.Rcc;
import configgui.java.commands.Command;
import configgui.java.commands.chains.ConfigureFabanNodes;
import configgui.java.commands.chains.DiscoverTopologyCommand;
import javafx.application.Platform;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.fxml.FXML;
import javafx.scene.control.Button;
import javafx.scene.control.ChoiceBox;
import javafx.scene.control.Label;
import javafx.scene.control.ScrollPane;
import javafx.scene.control.Slider;
import javafx.scene.control.TextArea;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.input.ScrollEvent;
import javafx.stage.FileChooser;
import javafx.stage.FileChooser.ExtensionFilter;
import topology.FabanWaveTool;

public class FabanCompanionController {

	@FXML
	Button buttonDiscoverTopology;
	@FXML
	Button buttonCreateFabanConfiguration;
	@FXML
	Button buttonLoadJsonFile;
	@FXML
	Button buttonConfigureNodes;

	@FXML
	public Label labelTopologyScanned;
	@FXML
	public Label labelFabanConfiguration;
	@FXML
	public Label labelReadyForConfiguration;
	@FXML
	public Label labelNodesConfigured;
	@FXML
	public Label labelReadyForExecution;
	@FXML
	public Label labelTopoFile;
	@FXML
	public Label labelFabanFile;
	@FXML
	public Label labelConfigRemark;
	@FXML
	public Label labelReadyRemark;

	@FXML
	TextArea textAreaJsonFile;
	@FXML
	ScrollPane imgPane;
	@FXML
	ImageView image;
	@FXML
	ChoiceBox<String> engine;
	@FXML
	ChoiceBox<String> distributingbridge;
	@FXML
	Slider zoom;

	Rcc main;
	JsonObject jsonFab; // Topology description
	String jsonFileName; // Name of the json topology description file
	FabanWaveTool fwt; // Instance of the FabanWaveTool
	Image graphpic; // Image of the topology

	@FXML
	private void initialize() {
		labelTopologyScanned.setText("No");
		labelFabanConfiguration.setText("No");
		labelNodesConfigured.setText("No");
		labelReadyForExecution.setText("No");
		labelReadyForConfiguration.setText("No");

		fwt = new FabanWaveTool();

		labelReadyForConfiguration.textProperty().addListener(new ChangeListener<String>() {
			@Override
			public void changed(ObservableValue<? extends String> observable, String oldValue, String newValue) {
				buttonConfigureNodes.setDisable((newValue.equals("Yes") ? false : true));
			}
		});

		labelTopologyScanned.textProperty().addListener(new ChangeListener<String>() {
			@Override
			public void changed(ObservableValue<? extends String> observable, String oldValue, String newValue) {
				buttonCreateFabanConfiguration.setDisable((newValue.equals("Yes") ? false : true));
			}
		});

		zoom.valueProperty().addListener(new ChangeListener<Number>() {
			@Override
			public void changed(ObservableValue<? extends Number> observable, Number oldValue, Number newValue) {
				if (graphpic != null)
					image.setFitWidth(newValue.doubleValue() * 0.5 / 100.0 * graphpic.getWidth());
			}
		});

		engine.getItems().add("dot");
		engine.getItems().add("neato");
		engine.getItems().add("fdp");
		engine.getItems().add("sfdp");
		engine.getItems().add("twopi");
		engine.getItems().add("circo");

		engine.getSelectionModel().select(3);

		engine.getSelectionModel().selectedIndexProperty().addListener(new ChangeListener<Number>() {
			@Override
			public void changed(ObservableValue<? extends Number> observable, Number oldValue, Number newValue) {
				generateGraph(engine.getItems().get(newValue.intValue()), distributingbridge.getValue());
			}
		});

		distributingbridge.getItems().add("<None>");
		distributingbridge.getSelectionModel().selectedIndexProperty().addListener(new ChangeListener<Number>() {
			@Override
			public void changed(ObservableValue<? extends Number> observable, Number oldValue, Number newValue) {
				if (distributingbridge.getItems().size() > 0)
					generateGraph(engine.getValue(), distributingbridge.getItems().get(newValue.intValue()));
			}
		});

		distributingbridge.getSelectionModel().select(0);

		imgPane.setOnScroll((ScrollEvent event) -> {
			if (event.getDeltaY() < 0)
				zoom.setValue(zoom.getValue() - 10);
			else
				zoom.setValue(zoom.getValue() + 10);
			event.consume();
		});
	}

	/*
	 * Uses the DiscoveryProtocol to determine the available topology
	 */
	@FXML
	private void buttonDiscoverTopology_Click() {
		if (Rcc.out.getTerminal().getInput() == null) {
			return;
		} else if (Rcc.out.getTerminal().getInput().length() > 0) {
			Rcc.out.getTerminal().deleteInput();
		}

		switchLabel(labelTopologyScanned, false);
		switchLabel(labelFabanConfiguration, false);
		switchLabel(labelReadyForConfiguration, false);
		setLabelText(labelTopoFile, "");
		if (labelReadyForExecution.getText().equals("Yes")) {
			setLabelText(labelReadyRemark, "Config maybe outdated");
		}
		if (labelNodesConfigured.getText().equals("Yes")) {
			setLabelText(labelConfigRemark, "Config maybe outdated");
		}
		Command dt = new DiscoverTopologyCommand(Rcc.connectedNodes, this).exec();
		new Thread(() -> {
			dt.waitForTermination();
			Platform.runLater(() -> {
				initDbChoiceBox();
				generateGraph(engine.getValue());
			});
		}).start();
	}

	@FXML
	private void buttonCreateFabanConfiguration_Click() {
		String filename = jsonFileName;

		if (Rcc.out.getTerminal().getInput() == null) {
			return;
		} else if (Rcc.out.getTerminal().getInput().length() > 0) {
			Rcc.out.getTerminal().deleteInput();
		}

		Rcc.out.terminalPrintln("Info: Generation of Faban Configuration started");

		if (!fwt.loadJsonFile("topologies/" + filename)) {
			Rcc.out.terminalPrintln("Failed to generate a faban configuration (loadJsonFile)", 2);
			Rcc.out.startCommand();
			return;
		}
		if (!fwt.parseTopology()) {
			Rcc.out.terminalPrintln("Failed to generate a faban configuration (parseTopology)", 2);
			Rcc.out.startCommand();
			return;
		}
		if (!fwt.generateWaves()) {
			Rcc.out.terminalPrintln("Failed to generate a faban configuration (generateWaves)", 2);
			Rcc.out.startCommand();
			return;
		}
		if (!fwt.saveJsonFile("topologies/" + filename)) {
			Rcc.out.terminalPrintln("Failed to generate a faban configuration (saveJsonFile)", 2);
			Rcc.out.startCommand();
			return;
		} else {
			jsonFileName = filename;
			setFabanConfiguration();
		}

		switchLabel(labelReadyForConfiguration, true);
		Rcc.out.terminalPrintln("Successfully generated a faban configuration. ", 1);
		Rcc.out.startCommand();
	}

	@FXML
	private void buttonLoadJsonFile_Click() {

		if (Rcc.out.getTerminal().getInput() == null) {
			return;
		} else if (Rcc.out.getTerminal().getInput().length() > 0) {
			Rcc.out.getTerminal().deleteInput();
		}

		FileChooser fileChooser = new FileChooser();
		fileChooser.setTitle("Open Json Topology/Faban description file");
		fileChooser.getExtensionFilters().add(new ExtensionFilter("json", "*.json"));
		File filename = fileChooser.showOpenDialog(null);

		if (filename != null) {
			jsonFileName = filename.getName();

			Rcc.out.terminalPrintln("Info: Loading local json file \"" + jsonFileName + "\"");

			try {
				JsonParser parser = new JsonParser();
				JsonElement jsonElement = parser.parse(new FileReader(filename));
				jsonFab = jsonElement.getAsJsonObject();

				Gson gson = new GsonBuilder().setPrettyPrinting().create();

				Platform.runLater(() -> {
					textAreaJsonFile.setText(gson.toJson(jsonFab));
				});
				switchLabel(labelReadyForConfiguration, false);
				if (jsonElement.getAsJsonObject().has("topology")) {
					switchLabel(labelTopologyScanned, true);
					setLabelText(labelTopoFile, jsonFileName);
					Rcc.out.terminalPrintln("Topology found", 1);
					if (jsonElement.getAsJsonObject().has("fabanconfiguration")) {
						Rcc.out.terminalPrintln("Faban Configuration found", 1);
						switchLabel(labelFabanConfiguration, true);
						setLabelText(labelFabanFile, filename.getName());
						fwt.loadJsonFile(filename.getAbsolutePath());
						fwt.parseTopology();
						fwt.parseFabanConfiguration();
						if (fwt.validateFabanConfiguration()) {
							switchLabel(labelReadyForConfiguration, true);
							Rcc.out.terminalPrintln("Faban Configuration validated", 1);
						} else {
							Rcc.out.terminalPrintln("Faban Configuration invalid. Generate new configuration", 1);
						}
					} else {
						switchLabel(labelFabanConfiguration, false);
						setLabelText(labelFabanFile, "");
					}
					Rcc.out.startCommand();
				} else {
					Rcc.out.terminalPrintln("Invalid file", 1);
					Rcc.out.startCommand();
					switchLabel(labelTopologyScanned, false);
					switchLabel(labelFabanConfiguration, false);
					setLabelText(labelTopoFile, "");
					setLabelText(labelFabanFile, "");
				}
			} catch (FileNotFoundException e) {
				Rcc.out.addException(e);
			}

			initDbChoiceBox();
			generateGraph(engine.getValue());
		}
	}

	@FXML
	private void buttonConfigureNodes_Click() {
		if (Rcc.out.getTerminal().getInput() == null) {
			return;
		} else if (Rcc.out.getTerminal().getInput().length() > 0) {
			Rcc.out.getTerminal().deleteInput();
		}
		Rcc.out.terminalPrintln("Configuring nodes");

		// Calculate Routing tables
		JsonArray fabanconfiguration = jsonFab.getAsJsonArray("fabanconfiguration");

		ArrayList<String[]> fullRoutingTable = new ArrayList<>();

		ArrayList<String> waves = new ArrayList<>();
		waves.add("wave1");
		waves.add("wave2");

		for (JsonElement wavepairE : fabanconfiguration) {
			JsonObject wavepairO = wavepairE.getAsJsonObject();
			String distributingbridge = wavepairO.get("distributingbridge").getAsString();
			String dbRouting = distributingbridge.split("\\.")[3] + " " + (-1) + " ";

			for (String wave : waves) {
				JsonObject jWave = wavepairO.getAsJsonObject(wave);
				String checkingbridge = jWave.get("checkingbridge").getAsString();
				dbRouting += getlinkPorts(distributingbridge, checkingbridge)[0];

				// Read routing information from Faban Description File into a Hashmap
				Map<String, ArrayList<String>> routing = new HashMap<>();
				for (Map.Entry<String, JsonElement> eLink : jWave.getAsJsonObject("routing").entrySet()) {
					ArrayList<String> targetlist = new ArrayList<>();
					for (JsonElement eTarget : eLink.getValue().getAsJsonArray()) {
						targetlist.add(eTarget.getAsString());
					}
					routing.put(eLink.getKey(), targetlist);
				}
				analyzeWave(fullRoutingTable, distributingbridge, routing, distributingbridge, checkingbridge);
			}
			String[] dbEntry = new String[2];
			dbEntry[0] = distributingbridge;
			dbEntry[1] = dbRouting;
			fullRoutingTable.add(dbEntry);
		}
		Rcc.out.terminalPrintln("Generation of routingtable \t\t[OK]", 1);

		// Run configuration of nodes
		new ConfigureFabanNodes(Rcc.connectedNodes, fullRoutingTable, this).exec();
	}

	private int[] getlinkPorts(String A, String B) {
		int[] res = { -1, -1 };

		JsonArray links = jsonFab.getAsJsonObject("topology").getAsJsonArray("links");
		if (links == null) {
			System.out.println("NULL");
			return res;
		}
		for (JsonElement e : links) {
			JsonObject link = e.getAsJsonObject();
			JsonObject A2 = link.getAsJsonObject("A");
			JsonObject B2 = link.getAsJsonObject("B");
			if (A2.get("name").getAsString().equals(A) && B2.get("name").getAsString().equals(B)) {
				res[0] = A2.get("port").getAsInt();
				res[1] = B2.get("port").getAsInt();
			} else if (A2.get("name").getAsString().equals(B) && B2.get("name").getAsString().equals(A)) {
				res[1] = A2.get("port").getAsInt();
				res[0] = B2.get("port").getAsInt();
			}
		}

//		System.out.println(A + ":" + res[0] + " --> " + B + ":" + res[1]);

		return res;
	}

	/**
	 * 
	 * @param fullRoutingTable
	 * @param sender
	 * @param map
	 * @param prevBridge
	 * @param actBridge
	 */
	private void analyzeWave(ArrayList<String[]> fullRoutingTable, String sender, Map<String, ArrayList<String>> map, String prevBridge, String actBridge) {
		if (!actBridge.equals(sender)) {
			int portIn = getlinkPorts(prevBridge, actBridge)[1];
			ArrayList<String> nextNodes = map.get(actBridge);
			String targetPorts = "";
			for (String c : nextNodes) {
				int portOut = getlinkPorts(actBridge, c)[0];
				targetPorts += portOut;
				analyzeWave(fullRoutingTable, sender, map, actBridge, c);
			}

			if (targetPorts.length() > 0) {
				String[] entry = new String[2];
				entry[0] = actBridge;
				entry[1] = sender.split("\\.")[3] + " " + portIn + " " + targetPorts;
				fullRoutingTable.add(entry);
			}
		}
	}

	public void switchLabel(Label label, boolean value) {
		Platform.runLater(() -> {
			if (value) {
				label.setText("Yes");
				label.setStyle("-fx-text-fill: rgb(78,154,6); -fx-font-weight: bold;");
			} else {
				label.setText("No");
				label.setStyle("-fx-text-fill: rgb(204.0,0.0,0.0); -fx-font-weight: bold;");
			}
		});
	}

	private void setLabelText(Label label, String s) {
		Platform.runLater(() -> {
			label.setText(s);
		});
	}

	public void setMain(Rcc main) {
		this.main = main;
	}

	public void setTopology(JsonObject fabJson, String filename) {
		this.jsonFab = fabJson;
		this.jsonFileName = filename;

		Platform.runLater(() -> {
			labelTopoFile.setText(filename);
		});

		Gson gson = new GsonBuilder().setPrettyPrinting().create();
		textAreaJsonFile.setText(gson.toJson(fabJson));

		switchLabel(labelTopologyScanned, true);
	}

	public void setFabanConfiguration() {
		jsonFab = new JsonObject();

		try {
			JsonParser parser = new JsonParser();
			JsonElement jsonElement = parser.parse(new FileReader("topologies/" + jsonFileName));
			jsonFab = jsonElement.getAsJsonObject();

			Gson gson = new GsonBuilder().setPrettyPrinting().create();
			textAreaJsonFile.setText(gson.toJson(jsonFab));

			setLabelText(labelFabanFile, jsonFileName);
			switchLabel(labelFabanConfiguration, true);
		} catch (FileNotFoundException e) {
			System.err.println("File \"" + jsonFileName + "\" not found");
		}

	}

	private String dotBridge(String a) {
		return "\"" + a + "\"";
	}

	private String dotLink(String a, String b) {
		return "\"" + a + "\"" + " -> " + "\"" + b + "\"";
	}

	private boolean modifyDotLink(String oldVal, String newVal, ArrayList<String> links) {
		if (links.remove(oldVal))
			return links.add(newVal);
		return false;
	}

	public void generateGraph(String engine, String db) {
		try {
			if (jsonFab != null) {

				// Create folder if it does not exist
				if (!(new File("topologies").exists())) {
					new File("topologies").mkdirs();
				}

				// Create temporary dot file
				String dotpath = "topologies/.tempgraph.dot";
				PrintWriter dotwriter = new PrintWriter(dotpath);

				// Read bridge and links of the general topology
				ArrayList<String> dotBridgeSet = new ArrayList<>();
				ArrayList<String> dotLinkSet = new ArrayList<>();

				JsonObject topo = jsonFab.getAsJsonObject("topology");
				if (topo != null) {
					JsonArray bridges = topo.getAsJsonArray("bridges");
					if (bridges != null)
						for (JsonElement b : bridges) {
							dotBridgeSet.add(dotBridge(b.getAsString()));
						}

					JsonArray links = topo.getAsJsonArray("links");
					if (links != null) {
						for (JsonElement l : links) {
							JsonObject a = l.getAsJsonObject().getAsJsonObject("A");
							JsonObject b = l.getAsJsonObject().getAsJsonObject("B");
							if (a != null && b != null && a.has("name") && b.has("name"))
								dotLinkSet.add(dotLink(a.get("name").getAsString(), b.get("name").getAsString()));
						}
					}
				}

				ArrayList<String> waveRouteSets = new ArrayList<>();

				// Read wave routing information
				if (db != null && !db.equals("<None>")) {
					dotBridgeSet.remove(db);
					dotBridgeSet.add(dotBridge(db) + "[style=filled]");

					ArrayList<String> waves = new ArrayList<>();
					waves.add("wave1");
					waves.add("wave2");

					JsonArray fabConf = jsonFab.getAsJsonArray("fabanconfiguration");
					if (fabConf != null)
						for (JsonElement fabConfElement : fabConf) {
							JsonObject o = fabConfElement.getAsJsonObject();
							if (o != null && o.get("distributingbridge").getAsString().equals(db)) {
								int i = 1;
								for (String wave : waves) {
									String sub = "subgraph " + wave + "{\nedge [constraint=false colorscheme=set312 color=" + (3 + i) + "]\n";
									JsonObject jWave = o.getAsJsonObject(wave);
									String cb = jWave.get("checkingbridge").getAsString();
									dotBridgeSet.remove(cb);
									dotBridgeSet.add(dotBridge(cb) + "[colorscheme=set312 fillcolor=" + (3 + (i++)) + " style=filled]");
									for (Map.Entry<String, JsonElement> eLink : jWave.getAsJsonObject("routing").entrySet()) {
										for (JsonElement eTarget : eLink.getValue().getAsJsonArray()) {
											sub += dotLink(eLink.getKey(), eTarget.getAsString()) + ";\n";
											modifyDotLink(dotLink(eLink.getKey(), eTarget.getAsString()), dotLink(eLink.getKey(), eTarget.getAsString()) + "[style=invis]",
													dotLinkSet);
											modifyDotLink(dotLink(eTarget.getAsString(), eLink.getKey()), dotLink(eTarget.getAsString(), eLink.getKey()) + "[style=invis]",
													dotLinkSet);
										}
									}
									sub += "}\n";
									waveRouteSets.add(sub);
								}
							}
						}
				}

				String dotBridges = "node [margin=0.1 fontcolor=black width=0.5 shape=rect]\n";
				String dotTopo = "subgraph topo{\n\tedge [dir=none, color=gray50]\n";
				for (String b : dotBridgeSet) {
					dotBridges += b + ";";
				}
				for (String l : dotLinkSet) {
					dotTopo += l + ";\n";
				}
				dotTopo += "}\n";

				dotwriter.println("digraph{\ngraph[dpi=300]");
				dotwriter.println(dotBridges);
				dotwriter.println(dotTopo);
				for (String sub : waveRouteSets)
					dotwriter.println(sub);
				dotwriter.println("}");
				dotwriter.close();

				runCommand("dot -K" + engine + " -Tpng topologies//.tempgraph.dot -o topologies//.tempgraph.png");
				graphpic = new Image(new FileInputStream("topologies/.tempgraph.png"));
				image.setImage(graphpic);
				image.setFitWidth(graphpic.getWidth() * zoom.getValue() * 0.5 / 100.0);
			}
		} catch (Exception e) {
			Rcc.out.addException(e);
		}
	}

	public void generateGraph(String engine) {
		generateGraph(engine, null);
	}

	public void initDbChoiceBox() {
		distributingbridge.getItems().clear();
		distributingbridge.getItems().add("<None>");
		distributingbridge.getSelectionModel().select(0);
		if (jsonFab != null) {
			JsonObject topo = jsonFab.getAsJsonObject("topology");
			if (topo != null) {
				JsonArray bridges = topo.getAsJsonArray("bridges");
				if (bridges != null)
					for (JsonElement b : bridges)
						distributingbridge.getItems().add(b.getAsString());
			}
		}
	}

	private void runCommand(String cmd) {
		try {
			Runtime r = Runtime.getRuntime();
			Process p = r.exec(cmd);
			InputStream in = p.getInputStream();
			BufferedInputStream buf = new BufferedInputStream(in);
			InputStreamReader inread = new InputStreamReader(buf);
			BufferedReader bufferedreader = new BufferedReader(inread);

			String line;
			while ((line = bufferedreader.readLine()) != null) {
				System.out.println(line);
			}

			try {
				if (p.waitFor() != 0) {
					System.err.println("exit value = " + p.exitValue());
				}
			} catch (InterruptedException e) {
				System.err.println(e);
			} finally {
				bufferedreader.close();
				inread.close();
				buf.close();
				in.close();
			}
		} catch (Exception e) {
			Rcc.out.addException(e);
		}

	}

}