package configgui.java.model.cells;

import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.control.Button;
import javafx.scene.control.Label;
import javafx.scene.control.ListCell;
import javafx.scene.layout.GridPane;
import javafx.scene.paint.Color;
import javafx.scene.shape.Circle;

import java.io.IOException;

import configgui.java.model.configuration.BananaInformation;

public class BananaInformationListViewCell extends ListCell<BananaInformation> {
	@FXML
	Label labelIP;

	@FXML
	Circle circleStatus;

	@FXML
	GridPane gpRoot;

	@FXML
	Button disconnect2;

	@FXML
	Label cpuLoad;

	private FXMLLoader fxmlLoader;
	private BananaInformation banana;

	@Override
	protected void updateItem(BananaInformation bananaInformation, boolean empty) {
		super.updateItem(bananaInformation, empty);
		banana = bananaInformation;
		if (empty || bananaInformation == null) {
			setText(null);
			setGraphic(null);
			return;
		}
		if (fxmlLoader == null) {
			fxmlLoader = new FXMLLoader(getClass().getResource("../../../view/fragments/BananaInformationListCell.fxml"));
			fxmlLoader.setController(this);

			try {
				fxmlLoader.load();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		labelIP.setText(bananaInformation.getIP());
		switch (bananaInformation.getState()) {
		case BananaInformation.state_idle:
			circleStatus.setFill(Color.rgb(152, 251, 152));
			break;
		case BananaInformation.state_busy:
			circleStatus.setFill(Color.rgb(255, 232, 142));
			break;
		case BananaInformation.state_error:
			circleStatus.setFill(Color.rgb(240, 128, 128));
			break;
		case BananaInformation.state_disconnected:
			circleStatus.setFill(Color.rgb(75, 75, 75));
			break;
		}
		setText(null);
		setGraphic(gpRoot);
		cpuLoad.setText(bananaInformation.load().intValue() + " %");
	}

	@FXML
	private void disconnect() {
		if (banana.getClientThread() == null) {
			banana.removeFromList();
		} else {
			banana.stopClientThread();
		}
	}

}
