package configgui.java.model.configuration;

import configgui.java.Rcc;
import configgui.java.util.network.ClientThread;
import javafx.application.Platform;
import javafx.beans.Observable;
import javafx.beans.property.IntegerProperty;
import javafx.beans.property.SimpleIntegerProperty;
import javafx.beans.property.SimpleStringProperty;
import javafx.beans.property.StringProperty;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.util.Callback;

public class BananaInformation {

	public static final int state_disconnected = 0;
	public static final int state_idle = 1;
	public static final int state_busy = 2;
	public static final int state_error = 3;

	private StringProperty IP;
	private IntegerProperty state;
	private IntegerProperty cpuLoad;
	private ClientThread clientThread;
	private StringProperty output;
	private StringProperty lastSender;
	private StringProperty lastTimestamp;
	private ObservableList<OutputEvent> tableData;
	private int outputMaxNumberOfLinesConst = 10;

	private String tsRegEx = "\\[([a-zA-Z])*( )*([a-zA-Z])*( )*(\\d)*( )*(\\d)*:(\\d)*:(\\d)*( )*(\\d)*\\]:(.)*";

	public void setIP(String IP) {
		this.IP.set(IP);
	}

	public void setState(int state) {
		this.state.set(state);
	}

	public int getState() {
		return this.state.get();
	}

	public void setClientThread(ClientThread t) {
		clientThread = t;
	}

	public ClientThread getClientThread() {
		return clientThread;
	}

	public String getIP() {
		return this.IP.get();
	}

	public StringProperty getOutputProperty() {
		return output;
	}

	public String getOutput() {
		return output.get();
	}

	public ObservableList<OutputEvent> getTableData() {
		synchronized (tableData) {
			return tableData;
		}
	}

	public BananaInformation(String IP, int state) {
		this.IP = new SimpleStringProperty(IP);
		this.output = new SimpleStringProperty("");
		this.lastSender = new SimpleStringProperty("");
		this.lastTimestamp = new SimpleStringProperty("");
		this.state = new SimpleIntegerProperty(state);
		this.cpuLoad = new SimpleIntegerProperty(0);
		String s = Rcc.configuration.getConfigEntry("TableViewMaxSize");
		if (s == null || s.equals("null")) {
			Rcc.configuration.setConfigEntry("TableViewMaxSize", "" + outputMaxNumberOfLinesConst);
		} else {
			int maxOut = Integer.parseInt(s);
			if (maxOut > 0)
				outputMaxNumberOfLinesConst = maxOut;
		}
		tableData = FXCollections.observableArrayList(new Callback<OutputEvent, Observable[]>() {
			@Override
			public Observable[] call(OutputEvent param) {
				return new Observable[] { param.srcProperty(), param.msgProperty() };
			}
		});
	}

	public void appendText(String s) {
		String[] sLines = s.split("\n");
		for (int i = Math.max(0, sLines.length - outputMaxNumberOfLinesConst); i < sLines.length; i++) {
			String line = sLines[i];
			String sender = "NA";

			// Search for Namestamp
			int posNA = line.indexOf("[N<");
			int posNB = line.indexOf(">]", posNA);
			if (posNA < posNB && posNA >= 0 && posNB < line.length()) {
				sender = "[" + line.substring(posNA + 3, posNB) + "]";
				line = (posNA == 0 ? "" : line.substring(0, posNA)) + (posNB + 2 == line.length() ? "" : line.substring(posNB + 2, line.length()));
			} else {
				posNA = line.indexOf("\u2045N<");
				posNB = line.indexOf(">\u2046", posNA);
				if (posNA < posNB && posNA >= 0 && posNB < line.length()) {
					sender = "\u2045" + line.substring(posNA + 3, posNB) + "\u2046";
					line = (posNA == 0 ? "" : line.substring(0, posNA)) + (posNB + 2 == line.length() ? "" : line.substring(posNB + 2, line.length()));
				}
			}

			// Search for Timestamp
			int posTA = line.indexOf("[T<");
			int posTB = line.indexOf(">]", posTA);
			String timestamp = "";
			if (posTA < posTB && posTA >= 0 && posTB < line.length()) {
				timestamp = line.substring(posTA + 3, posTB);
				line = (posTA == 0 ? "" : line.substring(0, posTA)) + (posTB + 2 == line.length() ? "" : line.substring(posTB + 2, line.length()));
			}

			if (line != null && !line.equals("")) {
				if (sender.equals(lastSender.get()) && timestamp.equals(lastTimestamp.get())) {
					final String l = line;
					synchronized (tableData) {
						Platform.runLater(() -> tableData.get(tableData.size() - 1).appMsg(l));
					}
				} else {
					lastSender.set(sender);
					lastTimestamp.set(timestamp);
					synchronized (tableData) {
						while (tableData.size() >= outputMaxNumberOfLinesConst)
							tableData.remove(0);
						tableData.add(new OutputEvent(sender, timestamp, line));
					}
				}
			}
		}
	}

	public BananaInformation() {
		this(null, 0);
	}

	public void sendMessage(String m) {
		if (clientThread == null)
			return;
		clientThread.sendMessage(m);
	}

	public StringProperty IP() {
		return IP;
	}

	public void stopClientThread() {
		if (getClientThread() != null) {
			clientThread.stopClient();
			clientThread = null;
		}
	}

	public IntegerProperty state() {
		return state;
	}

	public IntegerProperty load() {
		return cpuLoad;
	}

	public String getId() {
		return getIP().split("\\.")[3];
	}

	public void updateCpuLoad(int cpuload) {
		cpuLoad.set(cpuload);
	}

	public void removeFromList() {
		Rcc.connectedNodes.remove(this);
	}

	@Override
	public String toString() {
		return getIP();
	}

}
