package configgui.java.model.configuration;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.PrintWriter;
import java.lang.reflect.Type;
import java.util.HashMap;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.reflect.TypeToken;

public class Configuration {
	private HashMap<String, String> configuration;

	public Configuration() {
		configuration = new HashMap<String, String>();
	}

	public String getConfigurationAsJson() {
		Gson gson = new GsonBuilder().setPrettyPrinting().create();
		return gson.toJson(configuration);
	}

	public void createDefaultConfiguration() {
		setDestinationPath("/home/user/Banananetwork/");
		setSourcePath("/home/user/Banananetwork/");
		setPort((short) 1550);
		setRemotePassword("banana!");
		setRemoteUser("user");
		setTargetNetwork("192.168.178.255");
		configuration.put("ack_mode", "hashes");
		configuration.put("columns", "-1");
		configuration.put("rows", "-1");
		writeConfigFile();
	}

	public void loadOrCreateConfiguration() throws Exception {
		BufferedReader bufferedReader = null;
		Gson gson = new Gson();
		bufferedReader = new BufferedReader(new FileReader("config.json"));
		Type mapType = new TypeToken<HashMap<String, String>>() {
		}.getType();
		configuration = gson.fromJson(bufferedReader, mapType);
	}

	public void setConfigEntry(String option, String value) {
		configuration.put(option.toLowerCase(), value);
		writeConfigFile();
	}

	public String getConfigEntry(String option) {
		return configuration.get(option.toLowerCase());
	}

	public short getPort() {
		return Short.parseShort(configuration.getOrDefault("port", "1550"));
	}

	public void setPort(short port) {
		setConfigEntry("port", "" + port);
	}

	private void writeConfigFile() {
		try {
			PrintWriter out = new PrintWriter(new File("config.json"));
			Gson g = new GsonBuilder().setPrettyPrinting().create();
			g.toJson(configuration, out);
			out.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}

	public String getTargetNetwork() {
		return getConfigEntry("targetnetwork");
	}

	public void setTargetNetwork(String targetNetwork) {
		setConfigEntry("targetnetwork", targetNetwork);
	}

	public String getRemoteUser() {
		return configuration.getOrDefault("remoteuser", "user");
	}

	public void setRemoteUser(String remoteUser) {
		setConfigEntry("remoteuser", remoteUser);
	}

	public String getRemotePassword() {
		return configuration.getOrDefault("remotepassword", "banana");
	}

	public void setRemotePassword(String remotePassword) {
		setConfigEntry("remotepassword", remotePassword);
	}

	public String getSourcePath() {
		return getConfigEntry("sourcepath");
	}

	public void setSourcePath(String sourcePath) {
		setConfigEntry("sourcepath", sourcePath);
	}

	public String getDestinationPath() {
		return configuration.getOrDefault("destinationpath", "/home/user/Banananetwork");
	}

	public void setDestinationPath(String destinationPath) {
		setConfigEntry("destinationpath", destinationPath);
	}

	public HashMap<String, String> getMap() {
		return configuration;
	}

	public void setMap(HashMap<String, String> newMap) {
		this.configuration = newMap;
		writeConfigFile();
	}
}
