package configgui.java.util.network;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Comparator;

import configgui.java.Rcc;
import configgui.java.model.configuration.BananaInformation;
import configgui.java.time.TimeC;
import javafx.application.Platform;

public class ClientThread extends Thread {

	private static final int ALIVEMESSAGEBOUND = 4;

	private boolean stopRequested;
	private Socket socket;
	private BananaInformation bananaInformation;
	private BufferedWriter writer;
	private BufferedReader reader;
	private final Object lock = new Object();
	private int aliveCounter;
	private ArrayList<String> messages;
	private boolean require_acks;
	private boolean ack_hashes;
	private final static String ACK = "" + (char) 0x06;
	private final static String STX = "\u0002";
	private final static String ETX = "\u0003";
	private final static String DLE = "\u0010";
	private String partialMessage = "";
	private boolean restoredOldConnetion = false;
	private TimeC timec;
	long currenttime[];

	public ClientThread(Rcc main, Socket clientSocket) {
		this.socket = clientSocket;
		this.messages = new ArrayList<String>();
		timec= new TimeC();
		for (BananaInformation b : Rcc.connectedNodes) {
			if (b.getIP().equals(socket.getInetAddress().getHostAddress())) {
				if (b.getState() == BananaInformation.state_disconnected) {
					bananaInformation = b;
					Rcc.out.setState(bananaInformation, BananaInformation.state_idle);
					restoredOldConnetion = true;
					break;
				}
			}
		}

		if (!stopRequested) {
			if (bananaInformation == null)
				bananaInformation = new BananaInformation(socket.getInetAddress().getHostAddress(), BananaInformation.state_idle);
			bananaInformation.setClientThread(this);
			stopRequested = false;
			require_acks = Rcc.configuration.getConfigEntry("require_acks") == null || !Rcc.configuration.getConfigEntry("require_acks").equals("false");
			ack_hashes = !(Rcc.configuration.getConfigEntry("ack_mode") == null) && Rcc.configuration.getConfigEntry("ack_mode").equals("hashes");
		}
	}

	public void stopClient() {
		Platform.runLater(() -> bananaInformation.setState(BananaInformation.state_disconnected));
		try {
			socket.close();
		} catch (Exception e) {
			// TODO: handle exception
		}
		stopRequested = true;
	}

	public void run() {
		initialize();

		Thread aliveThread = new Thread(new Runnable() {

			@Override
			public void run() {
				Thread.currentThread().setName("Livenesschecker (" + bananaInformation.getIP() + ")");
				while (!stopRequested) {
					try {
						sleep(1000);
					} catch (Exception e) {
						// TODO: handle exception
					}
					if (aliveCounter == ALIVEMESSAGEBOUND) {
						sendMessage("<alive>");
					} else if (aliveCounter == 2 * ALIVEMESSAGEBOUND) {
						bananaInformation.stopClientThread();
					}
					aliveCounter++;
				}
			}
		});
		aliveThread.start();

		Thread.currentThread().setName("Client (" + bananaInformation.getIP() + ")");
		while (!stopRequested) {
			waitForAndHandleInput();
		}

		cleanUp();

	}

	private void sendMessage(String message, boolean needAck) {
		synchronized (lock) {
			try {
				writer.write(STX + message + ETX);
				writer.flush();
			} catch (IOException e) {
			}
		}
	}

	public void sendMessage(String m) {
		sendMessage(m, true);
	}

	private void waitForAndHandleInput() {
		readLine();
		while (!messages.isEmpty()) {
			String msg = messages.get(0);
			messages.remove(0);
			for (String input : splitIntoLines(msg)) {
				if (input.startsWith(ACK))
					continue;
				if (input.startsWith("<Q>")) {
					continue;
				}
				if (input.contains(DLE)) {
					int begin = input.indexOf(DLE);
					int end = input.indexOf(DLE, begin + 1);
					String control_message = input.substring(begin + 1, end);
					Rcc.ctrlMsgListener.addNewMessage(this.bananaInformation.getIP(), control_message);
				} else if (!input.contains("<alive>") && !input.contains("<csreq>") && input.length() > 0) {
					bananaInformation.appendText(input);
				} else if (input.contains("<alive>")) {
					aliveCounter = 0;
					int load = Integer.parseInt(input.substring(8));
					Platform.runLater(() -> bananaInformation.updateCpuLoad(load));
				} else if (input.contains("<csreq>")) {
					String mode =Rcc.configuration.getConfigEntry("clocksyncmode"); 
					switch (mode) {
					case "NONE":
					case "CUSTOM":
//						sendMessage("_sync " + (System.currentTimeMillis() * 1000 * 1000)); // FIXME *1000
//						System.out.println("_sync " + (System.currentTimeMillis() * 1000 * 1000));
						currenttime = timec.gettingCTime();
						long temp = (long) currenttime[0]*1000000000 + currenttime[1];
						sendMessage("_sync " + temp);
						System.out.println("_sync " + temp);
						break;

					case "SYSTEM":
//						currenttime = timec.gettingCTime();
//						long temp = (long) currenttime[0]*1000000000 + currenttime[1];
//						System.out.println("sending _sync SYSTEM Mode: " + temp);
//						sendMessage("_sync " + temp);
//						break;
					
					case "NTP":
					case "PTP":
						/*Platz für mögliche Behandlung von ankommenden <csreq> Nachrichten,
						 *  obwohl eine "externe Synchronisation ausgewählt wurde*/ 
						
						break;
					default:
						Rcc.out.terminalPrintln("Wrong clocksyncmode entry!");
						break;
					}
				}
			}
		}
	}

	private void wait(int ms) {
		try {
			Thread.sleep(ms);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

	private String[] splitIntoMessages(String s) {
		String[] arr = s.split("(?=\u0002)");
		return arr;
	}

	private String[] splitIntoLines(String s) {
		String[] arr = s.split("\n");
		return arr;
	}

	private void readLine() {
		String res = new String("");
		try {
			res = reader.readLine();
			if (res != null) {
				String[] resArr = splitIntoMessages(res);
				synchronized (lock) {
					for (String s : resArr) {
						if (!s.startsWith(STX) && partialMessage.isEmpty() || s.startsWith(STX) && !partialMessage.isEmpty()) {
						} else if (!s.startsWith(STX) && s.endsWith(ETX)) {
							messages.add((partialMessage + s).substring(1, partialMessage.length() + s.length() - 1));
							partialMessage = "";
						} else if (s.startsWith(STX) && !s.endsWith(ETX)) {
							partialMessage = s + "\n";
						} else if (s.startsWith(STX) && s.endsWith(ETX)) {
							messages.add(s.substring(1, s.length() - 1));
						} else if (!s.startsWith(STX) && !s.startsWith(ETX)) {
							partialMessage += s + "\n";
						}
					}
				}
			} else {
				bananaInformation.stopClientThread();
			}
		} catch (IOException e) {
			bananaInformation.stopClientThread();
			aliveCounter++;
		}
	}

	private void cleanUp() {
		try {
			sendMessage("adieu");
			wait(50);
			socket.close();
		} catch (IOException e) {
			Rcc.out.addException(e);
		}
	}

	private void initialize() {
		aliveCounter = 0;
		if (!restoredOldConnetion)
			Platform.runLater(() -> {
				Rcc.connectedNodes.add(bananaInformation);
				Rcc.connectedNodes.sort(new Comparator<BananaInformation>() {
					public int compare(BananaInformation b1, BananaInformation b2) {
						return b1.getIP().compareTo(b2.getIP());
					}
				});
			});
		try {
			synchronized (lock) {
				this.writer = new BufferedWriter(new OutputStreamWriter(socket.getOutputStream()));
				this.reader = new BufferedReader(new InputStreamReader(socket.getInputStream()));
			}
		} catch (IOException e) {
			Rcc.out.addException(e);
		}
	}

	public boolean isActive() {
		return !stopRequested;
	}
}
