package configgui.java.util.network;

import java.util.HashMap;
import java.util.Vector;
import java.util.concurrent.Semaphore;

import configgui.java.Rcc;
import configgui.java.commands.Command;

public class ControlMsgListener implements Runnable {

	Rcc main;
	HashMap<String, Command> regCommandThreads; // Hashmap of cmdId and command pairs, which are registered, i.e. those
												// which are waiting for data from bananas.
	Vector<String[]> msgBuffer; // First element contains the ip of the Client from which the message was
								// received. Second element contains the data, which itself contain the id of
								// the cmd, which is associated with this message.
	Semaphore msgCounter;
	public boolean active;

	public ControlMsgListener(Rcc main) {
		regCommandThreads = new HashMap<>();
		msgBuffer = new Vector<>();
		msgCounter = new Semaphore(0);
		this.main = main;
		this.active = true;
	}

	public void addNewMessage(String ip, String s) {
		String[] sArr = new String[2];
		sArr[0] = ip;
		sArr[1] = s;
		msgBuffer.add(sArr);
		msgCounter.release();
	}

	public boolean registerCommandThread(String cmdId, Command cmd) {
		if (regCommandThreads.containsKey(cmdId)) {
			System.out.println("Registration failed: " + cmdId);
			return false;
		}
		regCommandThreads.put(cmdId, cmd);
		return true;
	}

	public boolean deregisterCommandThread(String cmdId, Command cmd) {
		if (regCommandThreads.containsKey(cmdId)) {
			regCommandThreads.remove(cmdId);
			return true;
		}
		return false;
	}

	@Override
	public void run() {
		while (true) {
			try {
				msgCounter.acquire();
				String[] msg = msgBuffer.get(0);
				msgBuffer.remove(0);
				String id = msg[1].substring(1, msg[1].indexOf(';') - 1);
				Command cmd = regCommandThreads.get(id);
				if (cmd != null) {
					cmd.addResponse(msg);
				} else {
					System.out.println("Unclassfiable control message received: " + msg[0] + "  " + msg[1]);
				}
			} catch (Exception e) {
//				Rcc.out.addException(e);
				e.printStackTrace();
				if (!active)
					return;
			}
		}
	}
}
