package configgui.java.util.terminal;

import configgui.java.model.configuration.BananaInformation;
import configgui.java.model.configuration.OutputEvent;
import javafx.application.Platform;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.collections.ListChangeListener;
import javafx.collections.ObservableList;
import javafx.scene.control.ComboBox;
import javafx.scene.control.TableCell;
import javafx.scene.control.TableColumn;
import javafx.scene.control.Tooltip;
import javafx.scene.control.cell.PropertyValueFactory;
import javafx.scene.layout.VBox;
import javafx.scene.text.Text;

public class BpiOutputModule {
	public VBox detPair;
	public ComboBox<BananaInformation> selectedNode;
	public CustomTableView eventTable;
	int srcwidth = 80;
	int timewidth = 60;
	int defaultValue = -1;

	public BpiOutputModule() {
		this(null);
	}

	public BpiOutputModule(ObservableList<BananaInformation> list) {
		selectedNode = new ComboBox<>();
		selectedNode.setPrefWidth(10000);
		selectedNode.setItems(list);
		detPair = new VBox();
		detPair.setStyle("-fx-padding: 0;" + "-fx-border-style: solid inside;" + "-fx-border-width: 1;" + "-fx-border-insets: 0;" + "-fx-border-radius: 0;"
				+ "-fx-border-color: rgb(211,215,207);");
		detPair.setFillWidth(true);
		detPair.getChildren().add(selectedNode);

		eventTable = new CustomTableView();
		eventTable.setPrefHeight(10000);

		TableColumn<OutputEvent, String> tcSrc = new TableColumn<OutputEvent, String>("Source");
		TableColumn<OutputEvent, String> tcTim = new TableColumn<OutputEvent, String>("Time");
		TableColumn<OutputEvent, String> tcMsg = new TableColumn<OutputEvent, String>("Message");

		tcTim.setCellFactory(param -> {
			return new TableCell<OutputEvent, String>() {
				@Override
				protected void updateItem(String item, boolean empty) {
					super.updateItem(item, empty);

					if (item == null || empty) {
						setText(null);
						setStyle("");
					} else {
						Text text = new Text(item);
						text.getStyleClass().add("time");
						if (getTableRow() != null && getTableRow().getItem() != null) {
//							setTooltip(new Tooltip(((OutputEvent) getTableRow().getItem()).getTime()));
							setTooltip(new Tooltip(generateTimeTooltip(((OutputEvent) getTableRow().getItem()).getTime())));

						}
						setGraphic(text);
					}
				}
			};
		});

		tcMsg.setCellFactory(param -> {
			return new TableCell<OutputEvent, String>() {
				@Override
				protected void updateItem(String item, boolean empty) {
					super.updateItem(item, empty);

					if (item == null || empty) {
						setText(null);
						setStyle("");
					} else {
						Text text = new Text(item);
//						text.setStyle("-fx-padding: 5px 5px 5px 5px;" + "-fx-text-alignment:left;-fx-fill: rgb(92,53,102);");
						if (getTableRow() != null && getTableRow().getItem() != null && ((OutputEvent) getTableRow().getItem()).getSrc() != null
								&& ((OutputEvent) getTableRow().getItem()).getSrc().length() >= 8) {
							String stCl = "BPI" + ((OutputEvent) (getTableRow().getItem())).getSrc().substring(1, 7).trim().toLowerCase();
							text.getStyleClass().removeIf(s -> s.startsWith("BPI"));
							text.getStyleClass().add(stCl);
						}
						text.setWrappingWidth(param.getPrefWidth() - 35);
						setGraphic(text);
					}
				}
			};
		});

		tcSrc.setPrefWidth(srcwidth);
		tcTim.setPrefWidth(timewidth);
		tcSrc.setResizable(false);
		tcMsg.setResizable(false);
		tcMsg.prefWidthProperty().bind(eventTable.widthProperty().add(-1 * (srcwidth + timewidth + 2)));

		tcSrc.setCellValueFactory(new PropertyValueFactory<OutputEvent, String>("src"));
		tcTim.setCellValueFactory(new PropertyValueFactory<OutputEvent, String>("timeshort"));
		tcMsg.setCellValueFactory(new PropertyValueFactory<OutputEvent, String>("msg"));
		eventTable.getColumns().addAll(tcSrc, tcMsg, tcTim);
		eventTable.setEditable(false);
		detPair.getChildren().add(eventTable);

		selectedNode.getSelectionModel().selectedItemProperty().addListener(new ChangeListener<BananaInformation>() {
			@Override
			public void changed(ObservableValue<? extends BananaInformation> arg0, BananaInformation arg1, BananaInformation arg2) {
				if (arg2 != null) {
					eventTable.setItems(arg2.getTableData());
					eventTable.getItems().addListener(new ListChangeListener<OutputEvent>() {
						@Override
						public void onChanged(javafx.collections.ListChangeListener.Change<? extends OutputEvent> c) {
							while (c.next())
								if (c.wasUpdated() || c.wasAdded() || c.wasRemoved() || c.wasReplaced())
									Platform.runLater(() -> {
										eventTable.refresh();
										eventTable.scrollTo(c.getList().size() - 1);
									});
						}
					});
				}
			}
		});

	}

	public BpiOutputModule(ObservableList<BananaInformation> list, int index) {
		this(list);
		defaultValue = index;
		updateDefaultValue();
	}

	public void updateDefaultValue() {
		selectedNode.getSelectionModel().select(defaultValue);
	}

	public void initBananas(ObservableList<BananaInformation> list) {
		selectedNode.setItems(list);
	}

	public void refresh() {
		eventTable.refresh();
	}

	private String generateTimeTooltip(String full_time) {
		return full_time;
	}
}
