package configgui.java.util.terminal;

import configgui.java.controller.BananaController;
import configgui.java.model.configuration.BananaInformation;
import javafx.application.Platform;
import javafx.scene.control.Label;
import javafx.scene.control.ProgressBar;

public class GeneralOutputInterface implements Runnable {

	public boolean active = true;
	public int charLimit = 10000;
	public int refreshInterval = 10;
	private Terminal terminal;

	StringBuffer stringBuffer = new StringBuffer();

	private Label statusTextField;
	private String statusText;

	private ProgressBar progressBar;
	private double progress;
	private BananaController bc;

	public GeneralOutputInterface(Terminal terminal, BananaController bc) {
		this.terminal = terminal;
		this.statusTextField = bc.getStatusTextField();
		this.progressBar = bc.getProgressBar();
		this.bc = bc;
	}

	public Terminal getTerminal() {
		return terminal;
	}

	public void terminalPrint(String s, int indent) {
		if (indent > 0) {
			String[] line = s.split("\n");
			for (int i = 0; i < line.length; i++) {
				line[i] = getIndentString(indent) + line[i];
				stringBuffer.append(
						line[i] + (i == line.length - 1 ? (s.charAt(s.length() - 1) == '\n' ? "\n" : "") : "\n"));
			}
		} else {
			stringBuffer.append(s);
		}
	}

	public void terminalPrintln(String s, int indent) {
		terminalPrint(s + "\n", indent);
	}

	public void terminalPrintErrln(String s, int indent) {
		terminalPrint("[ERR] " + s + "\n", indent);
	}

	public void terminalPrint(String s) {
		terminalPrint(s, 0);
	}

	public void terminalPrintln(String s) {
		terminalPrintln(s, 0);
	}

	public void statusSetText(String s) {
		statusText = s;
	}

	public void activateProgressBar() {
		progressBar.setVisible(true);
	}

	public void updateProgressBar(double act, double full) {
		progress = act / full;
	}

	public void lockTerminal() {
		terminal.setEditable(false);
	}

	public void unlockTerminal() {
		terminal.setEditable(true);
	}

	public boolean bananaOutPrintln(BananaInformation banana, String s) {
		if (banana == null)
			return false;
		String out = "[N<   RCC>]" + s + "\n";
		banana.appendText(out);
		return true;
	}

	public void statusClear() {
		statusSetText("");
		progressBar.setVisible(false);
	}

	public void addException(Exception e) {
		this.bc.reportNewException();
		this.bc.getExceptionTabController().addException(e);
	}

	public void setState(BananaInformation b, int state) {
		Platform.runLater(() -> b.setState(state));
	}

	private String getStringFromBuffer() {
		String s = stringBuffer.toString();
		stringBuffer.delete(0, stringBuffer.length());
		if (s.length() > charLimit) {
			char lastChar = s.charAt(s.length() - charLimit - 1);
			s = s.substring(Math.max(0, s.length() - charLimit), s.length());
			if (lastChar != '\n')
				s = s.substring(s.indexOf('\n') + 1, s.length());
		}
		return s;
	}

	private String getIndentString(int n) {
		String s = "";
		for (int i = 0; i < n; i++) {
			s += Terminal.CMD_INDENT_CHARS;
		}
		return s;
	}

	public void startCommand() {
		terminalPrint("\n" + Terminal.CMD_START);
	}

	public void startCommand(int indent) {
		terminalPrint("\n" + getIndentString(indent) + Terminal.CMD_START);
	}

	public boolean executeCommand(String cmd) {
		if (terminal.getInput() != null && terminal.getInput().length() > 0) {
			terminal.deleteInput();
		} else if (terminal.getInput() == null) {
			terminalPrintErrln("Wait until command execution terminates", 1);
			return false;
		}
		terminalPrintln(cmd);
		bc.dispatchCommand(cmd);
		return true;
	}

	public void run() {
		Thread.currentThread().setName("GeneralOutputInterface");
		synchronized (this) {
			while (active) {
				try {
					Thread.sleep(refreshInterval);

					String s = this.getStringFromBuffer();
					if (!s.isEmpty()) {
						final String stringFromBuffer = s;
						stringBuffer.delete(0, stringBuffer.length());
						if (!stringFromBuffer.isEmpty())
							Platform.runLater(() -> {
								int free = charLimit - terminal.getText().length();
								int overflow = stringFromBuffer.length() - free;
								char lastChar = '\n';
								if (overflow > 0) {
									lastChar = terminal.getText().charAt(overflow - 1);
								}
								terminal.deleteText(0, Math.max(0, overflow));
								if (lastChar != '\n')
									terminal.deleteText(0, terminal.getText().indexOf('\n') + 1);
								terminal.appendText(stringFromBuffer);
							});
					}

					if (!statusText.isEmpty()) {
						Platform.runLater(() -> {
							statusTextField.setText(statusText);
						});
					}

					if (progressBar.getProgress() != progress) {
						Platform.runLater(() -> {
							progressBar.setProgress(progress);
						});
					}

				} catch (Exception e) {

				}
			}
		}

	}

}
