package configgui.java.util.terminal;

import java.util.Collection;
import java.util.Collections;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.fxmisc.richtext.CodeArea;
import org.fxmisc.richtext.StyleSpans;
import org.fxmisc.richtext.StyleSpansBuilder;

public abstract class ShCodeArea extends CodeArea {

	private static final String[] KEYWORDS = new String[] { "help", "get", "set", "scan", "bpi", "put", "update", "updatelib", "discovertopology" };
	private static final String[] KEYWORDS2 = new String[] { "status", "getconfig", "setconfig", "compile", "compilelib", "clean", "cleanlib", "start", "stop", "ssh", "exit", "shutdown", "adieu" };
	private static final String[] KEYWORDS3 = new String[] { "getc", "setc", "deletec", "pcompile", "pstart", "pstop", "prestart", "pstatus", "locexit" };
	
	public static final String KEYWORD_PATTERN = "\\b(" + String.join("|", KEYWORDS) + ")\\b";
	public static final String KEYWORD_PATTERN2 = "\\b(" + String.join("|", KEYWORDS2) + ")\\b";
	public static final String KEYWORD_PATTERN3 = "\\b(" + String.join("|", KEYWORDS3) + ")\\b";
	public static final String KEY_PATTERN = "\"([^\"])*\":";
	public static final String VALUE_PATTERN = " \"([^\"])*\"(\n|,)";
	public static final String COMMAND_PATTERN = "[\n(: )]> [^\n]*";
	public static final String TIMESTAMP_PATTERN = "\\[([a-zA-Z])*( )*([a-zA-Z])*( )*(\\d)*( )*(\\d)*:(\\d)*:(\\d)*( )*(\\d)*\\]: ";
	public static final String NAMESTAMP_PATTERN = "\\[(.)*\\]:";
	public static final String ERRORNAMESTAMP_PATTERN = "(.)*\u2045(.)*\u2046:";
	public static final String ERRORLINE_PATTERN = ERRORNAMESTAMP_PATTERN; // + "(.|\n )*\n";

	private static final Pattern PATTERN = Pattern
			.compile("(?<KEY>" + KEY_PATTERN + ")" + "|(?<VALUE>" + VALUE_PATTERN + ")" + "|(?<KEYWORD>" + KEYWORD_PATTERN + ")" + "|(?<KEYWORD2>" + KEYWORD_PATTERN2 + ")" + "|(?<KEYWORD3>" + KEYWORD_PATTERN3 + ")"+ "|(?<COMMAND>"
					+ COMMAND_PATTERN + ")" + "|(?<TIMESTAMP>" + TIMESTAMP_PATTERN + ")" + "|(?<NAMESTAMP>" + NAMESTAMP_PATTERN + ")" + "|(?<ERROR>" + ERRORLINE_PATTERN + ")");

	public ShCodeArea() {
		this.richChanges().filter(ch -> !ch.getInserted().equals(ch.getRemoved())) // XXX
				.subscribe(change -> {
					this.setStyleSpans(0, computeHighlighting(this.getText()));
				});
	}

	private static StyleSpans<Collection<String>> computeHighlighting(String text) {
		Matcher matcher = PATTERN.matcher(text);
		int lastKwEnd = 0;
		StyleSpansBuilder<Collection<String>> spansBuilder = new StyleSpansBuilder<>();
		while (matcher.find()) {
			String styleClass = matcher.group("KEY") != null ? "key"
					: matcher.group("VALUE") != null ? "value"
							: matcher.group("COMMAND") != null ? "command"
									: matcher.group("KEYWORD") != null ? "keyword"
											: matcher.group("KEYWORD2") != null ? "keyword2"
													: matcher.group("KEYWORD3") != null ? "keyword3"
													: matcher.group("TIMESTAMP") != null ? "timestamp"
															: matcher.group("ERROR") != null ? "error" : matcher.group("NAMESTAMP") != null ? getDetailedNamestamp(matcher.group("NAMESTAMP")) : null;
			/* never happens */ assert styleClass != null;
			spansBuilder.add(Collections.emptyList(), matcher.start() - lastKwEnd);
			spansBuilder.add(Collections.singleton(styleClass), matcher.end() - matcher.start());
			lastKwEnd = matcher.end();
		}
		spansBuilder.add(Collections.emptyList(), text.length() - lastKwEnd);
		return spansBuilder.create();
	}

	private static String getDetailedNamestamp(String s) {
//		System.out.println(s.substring(1,6).trim().toLowerCase());
		return s.substring(1, 7).trim().toLowerCase();
	}
}
