//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
// 

#include "brDec.h"
#include "../../packets/fabMsg_m.h"

namespace fabansimulator
::bridge {

    Define_Module(BrDec);

    void BrDec::initialize()
    {
        _procQueueLoadPackets.setName("QueueLoad [packets]");
        _procQueueLoadByte.setName("QueueLoad [Byte]");
        _procQueueLoadPackets.record(0);
        _procQueueLoadByte.record(0);
    }

    void BrDec::finish()
    {
        _procQueueLoadPackets.record(_queue.getLength());
        _procQueueLoadByte.record(_queue.getByteLength());
        _procQueueingTime.recordAs("QueueingTime","s");
    }

    void BrDec::refreshDisplay() const
    {
        char buf[40];
        sprintf(buf, "|Q| = %dP / %ldbyte", _queue.getLength(), _queue.getByteLength());
        getDisplayString().setTagArg("t", 0, buf);
    }

    void BrDec::handleMessage(cMessage *msg)
    {
        if(dynamic_cast<fabansimulator::FabMsg*>(msg)!=nullptr) {
            if((double)par("computingtime") == 0) {
                send(msg,"out");
            }
            else {
                msg->setTimestamp();
                _queue.insert(dynamic_cast<fabansimulator::FabMsg*>(msg));
                if(_queue.getLength()==1) {
                    scheduleAt(simTime()+par("computingtime"), new cMessage("popQueue"));
                }
                // Update Statistics
                _procQueueLoadPackets.record(_queue.getLength());
                _procQueueLoadByte.record(_queue.getByteLength());
            }
        }
        else if(msg->isName("popQueue")) {
            FabMsg* fm = dynamic_cast<fabansimulator::FabMsg*>(_queue.pop());
            _procQueueingTime.collect(simTime() - fm->getTimestamp());
            send(fm,"out");
            if(_queue.getBitLength()>0)
                scheduleAt(simTime()+par("computingtime"), msg);
            else
                cancelAndDelete(msg);

            // Update Statistics
            _procQueueLoadPackets.record(_queue.getLength());
            _procQueueLoadByte.record(_queue.getByteLength());
        }
        else {
            throw cRuntimeError("Unknown message received!");
        }

    }

}
//namespace
