//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
// 

#include "bridgeAdmin.h"
#include <algorithm>

namespace fabansimulator
::bridge {

    Define_Module(BridgeAdmin);

    void BridgeAdmin::initialize()
    {
        _networkadmin = dynamic_cast<NetworkAdmin*>(this->getParentModule()->getParentModule()->getSubmodule("networkadmin"));
    }

    void BridgeAdmin::_printPrimaryRoutingTable()
    {
        std::cout << "\n==================================\nPrimaryRoutingTable of Bridge " << this->getParentModule()->getFullName() << "\n=================================="<< std::endl;
        std::cout << "\tDB \tingP \t    egrP" << std::endl;
        std::cout << "\t--------------------------" << std::endl;
        for(auto srcdb : _primaryroutingtable) {
            for(auto portIn : srcdb.second) {
                std::stringstream ss;
                ss << "[";
                for(auto portOut : portIn.second) {
                    ss << portOut;
                }
                ss << "]";
                std::cout <<"\t" << srcdb.first << "\t " << (portIn.first == -1 ? "*" : std::to_string(portIn.first).c_str()) << "\t--> " << ss.str().c_str() << std::endl;
            }
        }

        std::cout << "Nodes at Ports: ";
        for(auto x : _portsToConnectedNodes) {
            std::cout << x;
        }
        std::cout << std::endl;
    }

    void BridgeAdmin::_printRoutingTable()
    {
        std::cout << "\n==================================\nRoutingTable of Bridge " << this->getParentModule()->getFullName() << "\n=================================="<< std::endl;
        std::cout << "\tDB \tingP \t    egrP" << std::endl;
        std::cout << "\t--------------------------" << std::endl;
        for(auto srcdb : _routingtable) {
            for(auto portIn : srcdb.second) {
                std::stringstream ss;
                ss << "[";
                for(auto portOut : portIn.second) {
                    ss << portOut;
                }
                ss << "]";
                std::cout <<"\t" << srcdb.first << "\t " << (portIn.first == -1 ? "*" : std::to_string(portIn.first).c_str()) << "\t--> " << ss.str().c_str() << std::endl;
            }
        }

        std::cout << "Nodes at Ports: ";
        for(auto x : _portsToConnectedNodes) {
            std::cout << x;
        }
        std::cout << std::endl;
    }

    void BridgeAdmin::handleMessage(cMessage *msg)
    {
    }

    void BridgeAdmin::addRoutingInformation(std::string db, int portIn, int portOut)
    {
        std::map<std::string, std::map<int, std::vector<int>>>::iterator it = _routingtable.find(db);
        if(it == _routingtable.end()) {
            std::pair<int, std::vector<int>> newpair(portIn, std::vector<int>());
            newpair.second.push_back(portOut);
            std::map<int, std::vector<int>> newmapentry;
            newmapentry.insert(newpair);
            _routingtable.insert(std::pair<std::string, std::map<int, std::vector<int>>>(db, newmapentry));
        }
        else {
            std::map<int, std::vector<int>>::iterator it2 = it->second.find(portIn);
            if(it2 == it->second.end()) {
                std::pair<int, std::vector<int>> newpair(portIn, std::vector<int>());
                newpair.second.push_back(portOut);
                it->second.insert(newpair);
            }
            else {
                if(std::find(it2->second.begin(),it2->second.end(), portOut) == it2->second.end()) {
                    it2->second.push_back(portOut);}
            }
        }

    }

    void BridgeAdmin::addRoutingInformationNode(std::string db, int portIn)
    {
        std::vector<int> portList = getEgrPortIndexListNode();
        for(int egrPort : portList){
            addRoutingInformation(db, portIn, egrPort);
        }
    }

    void BridgeAdmin::addPrimaryRoutingInformation(std::string db, int portIn, int portOut)
    {
        std::map<std::string, std::map<int, std::vector<int>>>::iterator it = _primaryroutingtable.find(db);
        if(it == _primaryroutingtable.end()) {
            std::pair<int, std::vector<int>> newpair(portIn, std::vector<int>());
            newpair.second.push_back(portOut);
            std::map<int, std::vector<int>> newmapentry;
            newmapentry.insert(newpair);
            _primaryroutingtable.insert(std::pair<std::string, std::map<int, std::vector<int>>>(db, newmapentry));
        }
        else {
            std::map<int, std::vector<int>>::iterator it2 = it->second.find(portIn);
            if(it2 == it->second.end()) {
                std::pair<int, std::vector<int>> newpair(portIn, std::vector<int>());
                newpair.second.push_back(portOut);
                it->second.insert(newpair);
            }
            else {
                if(std::find(it2->second.begin(),it2->second.end(), portOut) == it2->second.end()) {
                    it2->second.push_back(portOut);}
            }
        }
    }

    void BridgeAdmin::addPortToNodeInformation(int port)
    {
        for(auto x : _portsToConnectedNodes) {
            if(x == port) return;
        }
        _portsToConnectedNodes.push_back(port);
    }

    std::vector<int> BridgeAdmin::getEgrPortIndexList(uint16_t senderid, int ingPort) {
        std::vector<int> list;
        std::string db = _networkadmin->getDistributingBridgeName(senderid);

        std::map<std::string, std::map<int, std::vector<int>>>::iterator x = _routingtable.find(db);
        if(x == _routingtable.end()) return list;

        if(x!=_routingtable.end()) {
            std::map<int, std::vector<int>>::iterator y = x->second.find(ingPort);
            if(y!= x->second.end()) {
                for(auto z : y->second) {
                    list.push_back(z);
                }
            }
        }
        return list;
    }

    std::vector<int> BridgeAdmin::getEgrPortIndexList(std::string routingId, int ingPort) {
        std::vector<int> list;
        std::string db = routingId;

        std::map<std::string, std::map<int, std::vector<int>>>::iterator x = _routingtable.find(db);
        if(x == _routingtable.end()) return list;

        if(x!=_routingtable.end()) {
            std::map<int, std::vector<int>>::iterator y = x->second.find(ingPort);
            if(y!= x->second.end()) {
                for(auto z : y->second) {
                    list.push_back(z);
                }
            }
        }
        return list;
    }

    std::vector<int> BridgeAdmin::getEgrPortIndexListNode()
    {
        return _portsToConnectedNodes;
    }

    std::vector<int> BridgeAdmin::getEgrPortIndexListConnectedUnused(uint16_t senderid, int ingPort)
    {
        std::vector<int> egrPortBridgeList = getEgrPortIndexList(senderid, ingPort);
        std::vector<int> egrPortNodeList = getEgrPortIndexListNode();

        std::vector<int> retVector;

        for(int i=0; i<par("noPorts").intValue(); i++) {
            if(!contains(egrPortBridgeList, i) && !contains(egrPortNodeList, i)) {
                if(this->getParentModule()->gate("out", i)->isConnected()) {
                    retVector.push_back(i);
                }
            }
        }

        return retVector;
    }

    bool BridgeAdmin::contains(std::vector<int> vec, int value)
    {
        for(auto x : vec) {
            if(x==value) return true;
        }
        return false;
    }

    bool BridgeAdmin::isActive()
    {
        return _active;
    }

    NetworkAdmin* BridgeAdmin::getNetworkAdmin()
    {
        return _networkadmin;
    }

    void BridgeAdmin::disableBridge()
    {
        _active = false;
    }

    bool BridgeAdmin::failactive()
    {
        return _failactive;
    }

    std::vector<int> BridgeAdmin::getPrimaryEgrPortIndexList(uint16_t senderid, int ingPort)
    {
        std::vector<int> list;
        std::string db = _networkadmin->getDistributingBridgeName(senderid);

        std::map<std::string, std::map<int, std::vector<int>>>::iterator x = _primaryroutingtable.find(db);
        if(x == _primaryroutingtable.end()) return list;

        if(x!=_primaryroutingtable.end()) {
            std::map<int, std::vector<int>>::iterator y = x->second.find(ingPort);
            if(y!= x->second.end()) {
                for(auto z : y->second) {
                    list.push_back(z);
                }
            }
        }
        return list;
    }

    std::vector<int> BridgeAdmin::getPrimaryEgrPortIndexListConnectedUnused(uint16_t senderid, int ingPort)
    {
        std::vector<int> egrPortBridgeList = getPrimaryEgrPortIndexList(senderid, ingPort);
        std::vector<int> egrPortNodeList = getEgrPortIndexListNode();

        std::vector<int> retVector;

        for(int i=0; i<par("noPorts").intValue(); i++) {
            if(!contains(egrPortBridgeList, i) && !contains(egrPortNodeList, i)) {
                if(this->getParentModule()->gate("out", i)->isConnected()) {
                    retVector.push_back(i);
                }
            }
        }

        return retVector;
    }

    void BridgeAdmin::setFailActive()
    {
        _failactive = true;
    }

    bool BridgeAdmin::isDistributingBridge(FabMsg* fm)
    {
        if(par("faultassumption").intValue() <= 1) {
            if(fm->getHopCounter()==0) {
                return true;
            }
            else {
                return false;
            }
        }
        else if(fm->getRoutingPhase()==0) {
            if(getNetworkAdmin()->getDistributingBridgeName(fm->getSenderID()).compare(this->getParentModule()->getFullName())==0) {
                return true;
            }
            return false;
        }
        else {
            return false;
        }
    }

    bool BridgeAdmin::isPrecheckingBridge(FabMsg* fm)
    {
        if(par("faultassumption").intValue() <= 1) {
            if(fm->getHopCounter()>0 && fm->getHopCounter() < (int) par("faultassumption")) {
                return true;
            }
            else {
                return false;
            }
        }
        else if(fm->getRoutingPhase()==0 && !getNetworkAdmin()->isCheckingBridge(getNetworkAdmin()->getDistributingBridgeName(fm->getSenderID()), this->getParentModule()->getFullName())) {
            return true;
        }
        else {
            return false;
        }
    }

    bool BridgeAdmin::isCheckingBridge(FabMsg* fm)
    {
        if(par("faultassumption").intValue() <= 1) {
            if(fm->getHopCounter() == (int)par("faultassumption") && par("checkingBridgeDeliveryTimeVerifactionActive").boolValue()) {
                return true;
            }
            else {
                return false;
            }
        }
        else if(fm->getRoutingPhase()==0 &&getNetworkAdmin()->isCheckingBridge(getNetworkAdmin()->getDistributingBridgeName(fm->getSenderID()), this->getParentModule()->getFullName())) {
            return true;
        }
        else {
            return false;
        }
    }

    uint32_t BridgeAdmin::generateCrc32(FabMsg* fMsg, Data* dMsg) {
            uint32_t crc = 0;

            // Serialize Packet
            uint8_t serializedPaket[fMsg->getByteLength() - (fMsg->getFieldSizeSignature()/8) + dMsg->getByteLength()];
            int i = 0;

            uint16_t id = fMsg->getSenderID();
            serializedPaket[i++] = *((uint8_t*) &id+1);
            serializedPaket[i++] = *((uint8_t*) &id);

            serializedPaket[i++] = fMsg->getSeqNo();

            double delTime = fMsg->getDeliveryTime();
            serializedPaket[i++] = *((uint8_t*) &delTime + 7);
            serializedPaket[i++] = *((uint8_t*) &delTime + 6);
            serializedPaket[i++] = *((uint8_t*) &delTime + 5);
            serializedPaket[i++] = *((uint8_t*) &delTime + 4);
            serializedPaket[i++] = *((uint8_t*) &delTime + 3);
            serializedPaket[i++] = *((uint8_t*) &delTime + 2);
            serializedPaket[i++] = *((uint8_t*) &delTime + 1);
            serializedPaket[i++] = *((uint8_t*) &delTime + 0);

            serializedPaket[i++] = fMsg->getHopCounter();

            for(uint8_t j = 0; j<dMsg->getDataArraySize(); j++) {
                serializedPaket[i++] = dMsg->getData(j);
            }

            for (int p = 0; p < fMsg->getByteLength() - (fMsg->getFieldSizeSignature()/8) + dMsg->getByteLength(); p++) {
                for (int i = 0; i < 7; i++) {
                    if (((crc >> 31) & 1) != ((serializedPaket[p] >> i) & 1))
                    crc = (crc << 1) ^ CRC32GENP;
                    else
                    crc = (crc << 1);
                }
            }
            return crc;
        }

}

//namespace
