//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
// 

#ifndef __FABANSIMULATOR_BRIDGEADMIN_H_
#define __FABANSIMULATOR_BRIDGEADMIN_H_

#include <omnetpp.h>
#include "../../networkadmin/networkadmin.h"
#include "../../packets/data_m.h"
#include "../../packets/fabMsg_m.h"

#define CRC32GENP 0x04C11DB7

using namespace omnetpp;

namespace fabansimulator
::bridge {

    class BridgeAdmin : public cSimpleModule
    {
    private:
        std::map<std::string, std::map<int, std::vector<int>>> _routingtable;
        std::map<std::string, std::map<int, std::vector<int>>> _primaryroutingtable;
        std::vector<int> _portsToConnectedNodes;
        void _printPrimaryRoutingTable();
        void _printRoutingTable();
        NetworkAdmin* _networkadmin;
        bool _active = true;
        bool _failactive = false;

    protected:
        virtual void initialize();
        virtual void handleMessage(cMessage *msg);

    public:
        bool isActive();
        void addRoutingInformation(std::string db, int portIn, int portOut);
        void addRoutingInformationNode(std::string db, int portIn);
        void addPrimaryRoutingInformation(std::string db, int portIn, int portOut);
        void addPortToNodeInformation(int port);
        std::vector<int> getEgrPortIndexList(uint16_t senderid, int ingPort);
        std::vector<int> getEgrPortIndexList(std::string routingId, int ingPort);
        std::vector<int> getPrimaryEgrPortIndexList(uint16_t senderid, int ingPort);
        std::vector<int> getEgrPortIndexListNode();
        std::vector<int> getEgrPortIndexListConnectedUnused(uint16_t senderid, int ingPort);
        std::vector<int> getPrimaryEgrPortIndexListConnectedUnused(uint16_t senderid, int ingPort);
        bool contains(std::vector<int> vec,int value);
        NetworkAdmin* getNetworkAdmin();
        void disableBridge();
        bool failactive();
        void setFailActive();
        bool isDistributingBridge(FabMsg* fm);
        bool isPrecheckingBridge(FabMsg* fm);
        bool isCheckingBridge(FabMsg* fm);
        uint32_t generateCrc32(FabMsg* fMsg, Data* dMsg);
    };

}
//namespace

#endif
