//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
// 

#include "ingressPort.h"
#include "../../packets/fabMsg_m.h"
#include "../../FabanSimLib.h"

namespace fabansimulator
::bridge {

    Define_Module(IngressPort);

    void IngressPort::initialize()
    {
        _bridgeadmin = dynamic_cast<BridgeAdmin*>(this->getParentModule()->getSubmodule("bridgeAdmin"));
    }

    void IngressPort::handleMessage(cMessage *msg)
    {
        if(_bridgeadmin->isActive()) {
            if(dynamic_cast<fabansimulator::FabMsg*>(msg)!=nullptr) {
                FabMsg* fm = dynamic_cast<fabansimulator::FabMsg*>(msg);
                fm->setRxPort(this->getIndex());
                if(_bridgeadmin->isDistributingBridge(fm) || _bridgeadmin->isCheckingBridge(fm)) {
                    uint32_t sig = fm->getSignature();
                    sig = rotateSignatureLeft(sig);
                    fm->setSignature(sig);
                }
                else if(_bridgeadmin->isPrecheckingBridge(fm) && fm->getHopCounter() < par("faultassumption").intValue()) {
                    uint32_t sig = fm->getSignature();
                    sig = rotateSignatureLeft(sig);
                    fm->setSignature(sig);
                }
            }
            else {
                throw cRuntimeError("Unknown message received!");
            }
            send(msg, "out");
        }
        else {
            delete(msg);
        }
    }

}
//namespace
