//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
// 

#ifndef __FABANSIMULATOR_ADMIN_H_
#define __FABANSIMULATOR_ADMIN_H_

#include <omnetpp.h>
#include "../../inc/json.hpp"
#include "../packets/fabMsg_m.h"

using json = nlohmann::json;

using namespace omnetpp;

namespace fabansimulator {

class NetworkAdmin : public cSimpleModule {
public:
    void distributePublicKeyPair(uint16_t id, uint32_t q, uint32_t r);
    std::string getDistributingBridgeName(uint16_t senderNode);
    int getMaxWavelength(uint16_t senderID);
    int getNumberOfBridges();
    int getMaxDataSize();
    bool isCheckingBridge(std::string db, std::string cb);
    uint32_t getDmask();
    uint32_t getCmask();
    uint32_t getPmask();
    uint32_t getRmask();
    std::string getRoutingIdentifier(FabMsg* fm);

protected:
    virtual void initialize();
    virtual void handleMessage(cMessage *msg);
    virtual void finish();

private:
    cTopology _topo;
    std::vector<cModule*> _nodelist, _bridgelist;
    cModule* _getModuleFromNode(std::string);
    uint32_t d_mask, c_mask, p_mask, r_mask;
    std::vector<std::string> _linkBlacklist;
    std::map<uint16_t, std::string> _senderIdBridgeMap;
    std::map<std::string, std::vector<std::string>> _checkingBridgeMap;
    std::map<std::string, int> _dbWavelengthMap;
    int _recursiveRouting(json jsarr, std::string db, std::string nodeFrom, std::string nodeTo);
    int _recursivePrimaryRouting(json jsarr, std::string db, std::string nodeFrom, std::string nodeTo);
    int _getInPort(std::string prevNode, std::string node, std::string nextNode);
    int _getOutPort(std::string prevNode, std::string node, std::string nextNode);
    void _addCheckingBridgeToMap(std::string db, std::string cb);

    void _printCheckingBridgeMap();
};

} //namespace

#endif
