//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
// 

#include "nodeAdmin.h"
#include "../../FabanSimLib.h"

namespace fabansimulator
::node {

    Define_Module(NodeAdmin);

    void NodeAdmin::initialize() {

        _id = (uint16_t) cSimulation::getActiveEnvir()->getUniqueNumber();

        _sgnKeyP = (uint32_t) (1 + 2*rand());
        _sgnKeyQ = (uint32_t) (1 + 2*rand());
        _sgnKeyR = oddMultiplication(_sgnKeyP, _sgnKeyQ);

        _networkadmin = dynamic_cast<NetworkAdmin*>(this->getParentModule()->getParentModule()->getSubmodule("networkadmin"));

        // Distribute own public keys to all other modules
        _networkadmin->distributePublicKeyPair(getId(), getSgnKeyQ(), getSgnKeyR());

        // Print private and public keys
        EV << "[" << this->getParentModule()->getName() << "] " << " P = " << std::hex << std::uppercase << getSgnKeyP();
        EV << "\t Q = " << std::hex << std::uppercase << getSgnKeyQ();
        EV << "\t R = " << std::hex << std::uppercase << getSgnKeyR() << endl;
    }

    void NodeAdmin::incSeqNo()
    {
        _seqNo = (_seqNo + 1) % 256;
    }

    void NodeAdmin::handleMessage(cMessage *msg) {

    }

    void NodeAdmin::addPublicSignatureKeys(uint16_t id, uint32_t q, uint32_t r) {
        std::pair<uint32_t, uint32_t> pair(q,r);
        std::pair<uint16_t, std::pair<uint32_t, uint32_t>> keyentry(id, pair);
        _publicSignatureKeys.insert(keyentry);
    }

    uint32_t NodeAdmin::generateCrc32(FabMsg* fMsg, Data* dMsg) {
        uint32_t crc = 0;

        // Serialize Packet
        uint8_t serializedPaket[fMsg->getByteLength() - (fMsg->getFieldSizeSignature()/8) + dMsg->getByteLength()];
        int i = 0;

        uint16_t id = fMsg->getSenderID();
        serializedPaket[i++] = *((uint8_t*) &id+1);
        serializedPaket[i++] = *((uint8_t*) &id);

        serializedPaket[i++] = fMsg->getSeqNo();

        double delTime = fMsg->getDeliveryTime();
        serializedPaket[i++] = *((uint8_t*) &delTime + 7);
        serializedPaket[i++] = *((uint8_t*) &delTime + 6);
        serializedPaket[i++] = *((uint8_t*) &delTime + 5);
        serializedPaket[i++] = *((uint8_t*) &delTime + 4);
        serializedPaket[i++] = *((uint8_t*) &delTime + 3);
        serializedPaket[i++] = *((uint8_t*) &delTime + 2);
        serializedPaket[i++] = *((uint8_t*) &delTime + 1);
        serializedPaket[i++] = *((uint8_t*) &delTime + 0);

        serializedPaket[i++] = fMsg->getHopCounter();

        for(uint8_t j = 0; j<dMsg->getDataArraySize(); j++) {
            serializedPaket[i++] = dMsg->getData(j);
        }

        for (int p = 0; p < fMsg->getByteLength() - (fMsg->getFieldSizeSignature()/8) + dMsg->getByteLength(); p++) {
            for (int i = 0; i < 7; i++) {
                if (((crc >> 31) & 1) != ((serializedPaket[p] >> i) & 1))
                crc = (crc << 1) ^ CRC32GENP;
                else
                crc = (crc << 1);
            }
        }
        return crc;
    }

    std::pair<uint32_t,uint32_t> NodeAdmin::getPublicKeys(uint16_t nodeId) {
        return _publicSignatureKeys.find(nodeId)->second;
    }

    uint16_t NodeAdmin::getId() const {
        return _id;
    }

    const std::map<uint16_t, std::pair<uint32_t, uint32_t> >& NodeAdmin::getPublicSignatureKeys() const {
        return _publicSignatureKeys;
    }

    uint8_t NodeAdmin::getSeqNo() const {
        return _seqNo;
    }

    uint32_t NodeAdmin::getSgnKeyP() const {
        return _sgnKeyP;
    }

    uint32_t NodeAdmin::getSgnKeyQ() const {
        return _sgnKeyQ;
    }

    uint32_t NodeAdmin::getSgnKeyR() const {
        return _sgnKeyR;
    }

    NetworkAdmin* NodeAdmin::getNetworkAdmin()
    {
        return _networkadmin;
    }

}

//namespace
