//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
// 

#ifndef __FABANSIMULATOR_NODEADMIN_H_
#define __FABANSIMULATOR_NODEADMIN_H_

#define CRC32GENP 0x04C11DB7

#include <omnetpp.h>
#include "../../packets/fabMsg_m.h"
#include "../../packets/data_m.h"
#include "../../networkadmin/networkadmin.h"

using namespace omnetpp;

namespace fabansimulator
::node {

    class NodeAdmin : public cSimpleModule {
    private:
        uint16_t _id = 0;
        uint8_t _seqNo = 0;
        uint32_t _sgnKeyP, _sgnKeyQ, _sgnKeyR;
        std::map<uint16_t, std::pair<uint32_t, uint32_t>> _publicSignatureKeys; // SenderID -> (q, r)
        fabansimulator::NetworkAdmin* _networkadmin;
    public:
        void addPublicSignatureKeys(uint16_t id, uint32_t q, uint32_t r);
        std::pair<uint32_t, uint32_t> getPublicKeys(uint16_t nodeId);
        uint32_t generateCrc32(FabMsg* fMsg, Data* dMsg);
        uint16_t getId() const;
        const std::map<uint16_t,std::pair<uint32_t,uint32_t> >& getPublicSignatureKeys() const;
        uint8_t getSeqNo() const;
        void incSeqNo();
        uint32_t getSgnKeyP() const;
        uint32_t getSgnKeyQ() const;
        uint32_t getSgnKeyR() const;
        NetworkAdmin* getNetworkAdmin();
    protected:
        virtual void initialize();
        virtual void handleMessage(cMessage *msg);
    };

}
//namespace

#endif
