/*
 * GlobalBpiLib.cpp
 *
 *      Author: Valentin Fitz
 *   Institute: University of Duisburg-Essen, Germany
 */

#include "GlobalBpiLib.hpp"
#include <cstring>
#include <sstream>
#include <ctime>
#include <iostream>
#include <fstream>
#include <csignal>
#include <climits>
#include <unistd.h>
#include <cstdarg>

using namespace std;
using json=nlohmann::json;

Configuration::Configuration() {

}

Configuration::~Configuration() {

}

/**
 * Reads and returns the value corresponding to the specified attribute from a local JSON file. The location
 * of the JSON file is specified in LocalServiceDefines.h.
 * @param jsonFilePath Path to the json file.
 * @param attribute Attribute of the attribute/value pair of interst.
 * @return Returns the value of the attribute/value specified by the parameter attribute. If the specified
 * attribute does not exist, an empty string will be returned.
 */
string Configuration::getConfig(string jsonFilePath, string attribute) {

	json _json = readJsonFile(jsonFilePath);
	if (_json == nullptr)
		return "";
	else {
		return _json.value(attribute, "");
	}
}

/**
 * (Over)writes the attribute/value pair, specified by the parameters, in the local JSON file. The location
 * of the JSON file is specified in LocalServiceDefines.h.
 * @param jsonFilePath Path to the json file.
 * @param attribute Attribute of the attribute/value pair of interst.
 * @param value Value of the attribute/value pair of interst.
 * @return
 */
bool Configuration::setConfig(string jsonFilePath, std::string attribute, std::string value) {
	json _json = readJsonFile(jsonFilePath);
	if (_json == nullptr)
		return false;
	else {
		_json[attribute] = value;
		writeJsonFile(_json, jsonFilePath);
		return true;
	}

}

json Configuration::readJsonFile(string pathToFile) {
	json _json;
	ifstream jsonfile(pathToFile);

	BpiOutStream bpiout(cout);
	BpiOutStream bpierr(cerr);

	if (!(bool) jsonfile || jsonfile.fail()) {
		bpierr << "Following file not found: \"" + string(pathToFile) << "\"" << endl;
		return nullptr;
	}

	jsonfile >> _json;
	jsonfile.close();

	return _json;
}

void Configuration::writeJsonFile(json json, string pathToFile) {
	ofstream jsonfile(pathToFile);

	BpiOutStream bpiout(cout);
	BpiOutStream bpierr(cerr);

	if (!(bool) jsonfile || jsonfile.fail()) {
		bpierr << "Could not write Json file" << endl;
		return;
	}
	jsonfile << setw(4) << json;
	jsonfile.close();
}

bool Configuration::deleteConfig(std::string pathToFile, std::string attribute) {
	json _json = readJsonFile(pathToFile);
	if (_json == nullptr)
		return false;
	else {
		_json.erase(attribute);
		writeJsonFile(_json, pathToFile);
		return true;
	}
}

BpiOutStream::BpiOutStream(ostream& os) :
		_os(os) {
}

BpiOutStream::~BpiOutStream() {
}

BpiOutStream& operator<<(BpiOutStream &bout, std::string s) {
	bout._localStringStream << s;
	return bout;
}

BpiOutStream& operator<<(BpiOutStream &bout, int i) {
	bout._localStringStream << i;
	return bout;
}

BpiOutStream& operator<<(BpiOutStream &bout, double d) {
	bout._localStringStream << d;
	return bout;
}

BpiOutStream& operator<<(BpiOutStream &bout, uint64_t d) {
	bout._localStringStream << d;
	return bout;
}

BpiOutStream& operator<<(BpiOutStream &bout, uint32_t d) {
	bout._localStringStream << d;
	return bout;
}

BpiOutStream& operator<<(BpiOutStream &bout, uint16_t d) {
	bout._localStringStream << d;
	return bout;
}

BpiOutStream& operator<<(BpiOutStream &bout, uint8_t d) {
	bout._localStringStream << d;
	return bout;
}

BpiOutStream& operator<<(BpiOutStream &bout, std::ostream& (*pf)(std::ostream&)) {
	istringstream iss(bout._localStringStream.str());
	bout._localStringStream.str("");
	stringstream ss;
	bool first = true;
	for (string line; getline(iss, line);) {
		if (first) {
			ss << (&bout._os == &cerr ? bpiGetNamestampErr(bpiName) : bpiGetNamestamp(bpiName)) << bpiGetTimestamp(Configuration::getConfig(JSON_CONFIGFILE,"clocksync")) << line;
			first = false;
		} else
			ss << endl << (&bout._os == &cerr ? bpiGetNamestampErr(bpiName) : bpiGetNamestamp(bpiName)) << bpiGetTimestamp(Configuration::getConfig(JSON_CONFIGFILE,"clocksync")) << line;
	}
	ss << pf;

	bout._os << ss.str();
	return bout;
}

DebugOutStream::DebugOutStream(ostream& os) :
		_os(os) {
}

DebugOutStream::~DebugOutStream() {
}

DebugOutStream& operator<<(DebugOutStream &dout, std::string s) {
	dout._localStringStream << s;
	return dout;
}

DebugOutStream& operator<<(DebugOutStream &dout, int i) {
	dout._localStringStream << i;
	return dout;
}

DebugOutStream& operator<<(DebugOutStream &dout, std::ostream& (*pf)(std::ostream&)) {
	istringstream iss(dout._localStringStream.str());
	dout._localStringStream.str("");
	stringstream ss;
	bool first = true;
	for (string line; getline(iss, line);) {
		if (first) {
			ss << bpiGetTimestamp(Configuration::getConfig(JSON_CONFIGFILE,"clocksync")) << line;
			first = false;
		} else
			ss << endl << bpiGetTimestamp(Configuration::getConfig(JSON_CONFIGFILE,"clocksync")) << line;
	}
	ss << pf;
	dout._os << ss.str();
	return dout;
}

void bpiSetName(std::string name) {
	bpiName = name;
}

std::string bpiGetName() {
	return bpiName;
}

Clock* getLocalClock() {
	return &localclock;
}

bool registerChild(int pid) {
	if (std::find(childs.begin(), childs.end(), pid) != childs.end()) {
		return false;
	} else {
		childs.push_back(pid);
		return true;
	}
}

bool deregisterChild(int pid) {
	if ((std::find(childs.begin(), childs.end(), pid)) != childs.end()) {
		childs.erase(std::find(childs.begin(), childs.end(), pid));
		return true;
	} else {
		return false;
	}
}

std::vector<int> getAllChilds() {
	return childs;
}

/**
 * Invokes a system call. The programm waits until the system call is finished and prints to stdout and stderr.
 * @param cmd Command to be executed by the system
 * @param printFct Pointer to a printer function
 * @param printErrFct Pointer to an error printer function
 */
void bpiSystemCall(std::string cmd, int (*printFct)(std::string), int (*printErrFct)(std::string)) {
	FILE *pipe_writer;

	BpiOutStream bpiout(cout);
	BpiOutStream bpierr(cerr);

	char puffer[PIPE_BUF];
	if ((pipe_writer = popen(cmd.c_str(), "r")) == NULL) {
		if (printErrFct == nullptr)
			bpierr << "System call \"" << cmd << "\" failed." << endl;
		else {
			std::stringstream ss;
			ss << "System call \"" << cmd << "\" failed." << endl;
			printErrFct(ss.str());
		}
	}
	signal(SIGCHLD, SIG_IGN);
	while (fgets(puffer, 255, pipe_writer) != NULL) {
		if (printFct != nullptr) {
			printFct(std::string(puffer));
		} else {
			bpiout << puffer << endl;
		}
	}
}

/**
 * Invokes a system call. The call runs concurrently to the application and prints to stdout and stderr.
 * @param cmd Command to be executed by the system
 * @param printFct Pointer to a printer function
 * @param printErrFct Pointer to an error printer function
 */
void bpiSystemCall2(std::string cmd, int (*printFct)(std::string), int (*printErrFct)(std::string), std::string cmdId) {
	FILE *pipe_writer;
	char puffer[PIPE_BUF];
	BpiOutStream bpiout(cout);
	BpiOutStream bpierr(cerr);

	if ((pipe_writer = popen(cmd.c_str(), "r")) == NULL) {
		if (printErrFct == nullptr)
			bpierr << "System call \"" << cmd << "\" failed." << endl;
		else {
			std::stringstream ss;
			ss << "System call \"" << cmd << "\" failed." << endl;
			printErrFct(ss.str());
		}
	}
	signal(SIGCHLD, SIG_IGN);

	pid_t pid;
	if ((pid = fork()) < 0) {
		if (printErrFct == nullptr)
			bpierr << "Error forking..." << endl;
		else
			printErrFct("Error forking...");
		return;
	} else if (pid == 0) {
		while (fgets(puffer, 255, pipe_writer) != NULL) {
			if (printFct != nullptr) {
				printFct(std::string(puffer));
			} else {
				bpiout << puffer << endl;
			}

		}
		if (cmdId.size() > 0)
			printFct(createStatCtrlMessage(2, cmdId.c_str(), "success"));
		_Exit(0);
	}
}
//
//std::string bpiTimestamp() {
//	std::stringstream ss;
//	time_t ltime = time(NULL);
//	char *charTime = asctime(localtime(&ltime));
//	charTime[strlen(charTime) - 1] = 0;
//	ss << "[" << string(charTime).substr(11, 8) << "]: ";
//	return ss.str();
//}

/**
 * Splits a tring into items, divided by a delimiter, and returns the set of items as a std::vector<strings>.
 * Consecutive delimiter characters will be trimmed to one single delimiter character.
 * @param str String to split into items.
 * @param delimiter Delimiter character.
 * @return std::vector<std::string> of items.
 */
vector<string> bpiSplitIntoItems(string str, char delimiter) {
	vector<string> items;
	size_t pos;
	str.erase(0, max((int) (str.find_first_not_of(delimiter)), 0));
	str.erase(str.find_last_not_of(delimiter) + 1);
	while ((pos = str.find(delimiter, 0)) != string::npos) {
		items.push_back(str.substr(0, pos));
		str = str.substr(pos + 1, string::npos);
		str.erase(0, max((int) (str.find_first_not_of(" ")), 0));
	}
	items.push_back(str);
	return items;
}

std::string bpiGetNamestamp(std::string name) {
	stringstream ss;
	ss << "[N<" << setw(STAMPLENGTH) << name.substr(0, STAMPLENGTH) << ">]";
	return ss.str();
}

std::string bpiGetNamestampErr(std::string name) {
	stringstream ss;
	ss << "\u2045N<" << setw(STAMPLENGTH) << name.substr(0, STAMPLENGTH) << ">\u2046";
	return ss.str();
}

std::string bpiGetTimestamp(std::string clocksyncmode) {
	std::stringstream ss;
	/* FIXME JN */
	if (strncmp(clocksyncmode.c_str(), "CUSTOM", 6)==0) {
		double time = localclock.gettime();
		uint64_t time_micro = (uint64_t) time;
		ss << "[T<" << time_micro << ">]";
	} else if (strncmp(clocksyncmode.c_str(), "SYSTEM", 6) == 0 || strncmp(clocksyncmode.c_str(), "PTP", 3) == 0 || strncmp(clocksyncmode.c_str(), "NTP", 6) == 0) {
//		timespec timeSp;
		clock_gettime(CLOCK_REALTIME, &globalTS);
		uint64_t localtimeMicro = (uint64_t) ((uint64_t) TEN_POWER_NINE * (uint64_t) globalTS.tv_sec + globalTS.tv_nsec) / 1000;
		ss << "[T<" << localtimeMicro << ">]";
	}

	/*END JN*/
	return ss.str();
}

uint64_t getLocalTime() {
	clock_gettime(CLOCK_REALTIME, &globalTS);
	uint64_t localtimeMicro = (uint64_t) ((uint64_t) TEN_POWER_NINE * (uint64_t) globalTS.tv_sec + globalTS.tv_nsec) / 1000;
	return localtimeMicro;
}

uint64_t getLocalTime48() {
	return (getLocalTime() << 16) >> 16;
}

std::string bpiConfirm(std::string command, std::string optional) {
	std::stringstream ss;
	ss << "<Q>" << command << " " << optional;
	ss.flush();
	return ss.str();
}

bool readInputFromStdin(std::string* s, int bufsize) {
	char puffer[bufsize];
	if (fgets(puffer, bufsize, stdin) == NULL)
		return false;
	else {
		*s = puffer;
		*s = s->substr(0, s->find_first_of("\n"));
		return s;
	}
}

int writeToFileDescriptor(int fileDescriptor, std::string data) {
	std::string s = data + "\n";
	return write(fileDescriptor, s.c_str(), s.length());
}

int readFromFileDescriptor(int fileDescriptor, std::string& data, int buffersize) {
	char b[buffersize];
	memset(&b[0], 0, sizeof(b));
	int n = -1;
	if ((n = read(fileDescriptor, b, buffersize - 1)) > 0) {
		b[n] = '\0';
		*&data = b;
	}
	return n;
}

void sigtermHandler(int signum) {
	std::vector<int> childs = getAllChilds();
	for (int pid : childs) {
		kill(pid, SIGTERM);
	}
	exit(signum);
}

/**
 * Creates a control message, consisting of argc many char* strings , separated each by a semicolon.
 * The result is enclosed by the control character DLE. A control message may not contain any line breaks.
 * @param argc Number of arguments in the argumentvector
 * @return Control message
 */
std::string createCtrlMessage(int argc, ...) {
	if (argc < 1)
		return "";
	std::stringstream ss;
	ss << DLE;
	va_list ptr;
	va_start(ptr, argc);
	ss << "\"" << va_arg(ptr, char*) << "\"";
	for (int i = 1; i < argc; i++) {
		ss << ";" << "\"" << va_arg(ptr, char*) << "\"";
	}
	ss << DLE;

	return ss.str();
}

/**
 * Creates a control message, consisting of argc many char* strings , separated each by a semicolon.
 * The result is enclosed by the control character DLE. A control message may not contain any line breaks.
 * @param argc Number of arguments in the argumentvector
 * @return Control message
 */
std::string createStatCtrlMessage(int argc, ...) {
	if (argc < 1)
		return "";
	std::stringstream ss;
	ss << DLE;
	va_list ptr;
	va_start(ptr, argc);
	ss << "\"" << va_arg(ptr, char*) << "\"";
	ss << ";\"S\"";
	for (int i = 1; i < argc; i++) {
		ss << ";" << "\"" << va_arg(ptr, char*) << "\"";
	}
	ss << DLE;

	return ss.str();
}

std::string macToString(int len, char* mac) {
	std::stringstream ss;
	for (int i = 0; i < len; i++) {
		ss << (i == 0 ? "" : ":");
		ss << std::setfill('0') << std::setw(2) << std::hex << std::uppercase << (uint16_t) mac[i];
	}
	return ss.str();
}

double getCurrentCpuUsage() {
	long double a[4], b[4], loadavg;
	FILE *fp;

	fp = fopen("/proc/stat", "r");
	if (fscanf(fp, "%*s %Lf %Lf %Lf %Lf", &a[0], &a[1], &a[2], &a[3]) < 0)
		return -1;
	fclose(fp);
	sleep(1);

	fp = fopen("/proc/stat", "r");
	if (fscanf(fp, "%*s %Lf %Lf %Lf %Lf", &b[0], &b[1], &b[2], &b[3]) < 0)
		return -1;
	fclose(fp);

	loadavg = ((b[0] + b[1] + b[2]) - (a[0] + a[1] + a[2])) / ((b[0] + b[1] + b[2] + b[3]) - (a[0] + a[1] + a[2] + a[3]));
	return loadavg;
}

Clock::Clock() :
		beg_(clock_::now()) {
}

void Clock::set(uint64_t us) {
//double d = std::chrono::duration_cast<std::chrono::microseconds>(clock_::now()-beg_).count();
	beg_ = clock_::now();
	beg_ = beg_ - std::chrono::microseconds(us);
}

void Clock::reset() {
	beg_ = clock_::now();
}

double Clock::gettime() const {
	return std::chrono::duration_cast<std::chrono::microseconds>(clock_::now() - beg_).count();
}
