/*
 * CommandHandler.cpp
 *
 *      Author: Valentin Fitz
 *   Institute: University of Duisburg-Essen, Germany
 */

#include <limits.h>
#include "BasicInputHandler.h"

using namespace std;

BasicInputHandler::BasicInputHandler() :
		bpiout(cout), bpierr(cerr) {
}

/**
 * Reads Commands from STDIN and initiates a command handling.
 */
void BasicInputHandler::runLoop() {
	protocol->start();
	char cmd[PIPE_BUF];

	while (_run) {
		if (fgets(cmd, PIPE_BUF, stdin) != NULL)
			basicHandleCommand(cmd);
	}
}

void BasicInputHandler::basicHandleCommand(char* cmdIn) {
	cmdIn[strlen(cmdIn) - 1] = 0;
	string input(cmdIn);

	if (strncmp(COMMAND_EXIT, cmdIn, strlen(COMMAND_EXIT)) == 0) {
		_CommandExit();
	} else if (strncmp(COMMAND_SYNC, cmdIn, strlen(COMMAND_SYNC)) == 0) {
		_CommandSync(cmdIn);
	} else if (strncmp(COMMAND_STATUS, cmdIn, strlen(COMMAND_STATUS)) == 0) {
		_CommandStatus(cmdIn);
	} else {
		handleCommandFI(cmdIn);
	}
}

void BasicInputHandler::handleCommand(char* cmdIn) {
	bpiout << "Override function \'handleCommand\' to implement command handling for custom commands" << endl;
}

void BasicInputHandler::_CommandStatus(std::string input) {
	// TODO
	bpiout << "Default Status function called" << endl;
}

void BasicInputHandler::_CommandSync(std::string input) {
	std::vector<std::string> items = bpiSplitIntoItems(input, ' ');
	long long int timepoint = std::stoull(items[1]);
	getLocalClock()->set(timepoint);
}


void BasicInputHandler::_CommandExit() {
	bpiout << bpiGetName() << " terminating" << endl;
	_run = false;
}

BasicInputHandler::~BasicInputHandler() {

}

void BasicInputHandler::handleCommandFI(char *cmdIn) {
	handleCommand(cmdIn);
}

