/*
 * CommandHandler.h
 *
 *      Author: Valentin Fitz
 *   Institute: University of Duisburg-Essen, Germany
 */

#ifndef BASICINPUTHANDLER_H_
#define BASICINPUTHANDLER_H_

#define COMMAND_STATUS "status"
#define COMMAND_SYNC "_sync"
#define COMMAND_EXIT "exit"


#include <string>
#include "../GlobalBpiLib.hpp"
#include "ProtocolLogic.h"

/**
 * \brief Basic input handler for protocol implementations.
 */
class BasicInputHandler {
public:
	BasicInputHandler();

	BpiOutStream bpiout; /**< Standard output printer. */
	BpiOutStream bpierr; /**< Error output printer. */

	void runLoop(); /**< Obtains and handles in an infinite loop incoming commands. */
	void basicHandleCommand(char *cmdIn); /**< Realizes basic command handling. */
	virtual void handleCommand(char *cmdIn); /**< Realizes extended command handling. */
	virtual ~BasicInputHandler();
	virtual void handleCommandFI(char *cmdIn); /**< Realizes FI command handling. */

protected:
	bool _run = true; /**< Indicates whether the InputHandler is active or not. */
	ProtocolLogic* protocol = nullptr; /**< Pointer to a ProtocolLogic. */
	virtual void _CommandStatus(std::string input); /**< Implementation of the command COMMAND_STATUS. */
	void _CommandSync(std::string input);
	void _CommandExit(); /**< Implementation of the command COMMAND_EXIT. */
};

#endif /* BASICINPUTHANDLER_H_ */
