/*
 * Port.cpp
 *
 *      Author: Valentin Fitz
 *   Institute: University of Duisburg-Essen, Germany
 */

#include "Port.h"

#include <unistd.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <linux/if_ether.h>
#include <netinet/in.h>
#include <net/if.h>

#include "EtherTypes.h"
#include "ProtocolLogic.h"

using namespace std;

Port::Port(int id, ProtocolLogic* proto) :
		bpiout(cout), bpierr(cerr),  statRx(0), statTx(0), _egrPort(-1), _ingPort(-1), _portId(id), _ackMsg(nullptr), protocol(proto) {
	pthread_mutex_init(&_mutex, NULL);
	pthread_cond_init(&_cond, NULL);
	pthread_mutex_init(&_ackMutex, NULL);
	pthread_cond_init(&_ackCond, NULL);
//	pthread_mutex_init(&_mutexBuf, NULL);
//	pthread_cond_init(&_bufCond, NULL);
	stringstream ss;
	ss << "eth0." << id;
	_devName = ss.str();
	_retrieveMacAddress();
}

Port::~Port() {
	pthread_mutex_destroy(&_mutex);
	pthread_cond_destroy(&_cond);
}

void Port::registerEthType(unsigned short eth_type) {
	_regEthTypes.push_back(eth_type);
}

void Port::start() {
	pthread_create(&_egrPort, NULL, _egrPortThreadEntryFunction, this);
	pthread_detach(_egrPort);
	pthread_create(&_ingPort, NULL, _ingPortThreadEntryFunction, this);
	pthread_detach(_ingPort);
//	pthread_create(&_ingPortFwd, NULL, _ingPortForwarderThreadEntryFunction, this);
//	pthread_detach(_ingPortFwd);
}

void Port::stop() {
	pthread_mutex_lock(&_mutex);
	active = false;
	pthread_cond_signal(&_cond);
	pthread_mutex_unlock(&_mutex);
}

char* Port::getRawMacAddress() {
	return source_mac;
}

void Port::setNeighbour(char* mac) {
	if (memcmp(mac, destination_mac, 6) == 0)
		return;
	pthread_mutex_lock(&_mutex);
	memcpy(destination_mac, mac, 6);
	pthread_mutex_unlock(&_mutex);
}

std::string Port::getLocalMacAddr() {
	return macToString(6, source_mac);
}

void Port::_retrieveMacAddress() {
	int fd;
	struct ifreq ifr;
	char name[IFNAMSIZ];

	fd = socket(AF_PACKET, SOCK_RAW, IPPROTO_RAW);

	strncpy(name, _devName.c_str(), IFNAMSIZ);
	strncpy(ifr.ifr_name, name, IFNAMSIZ);

	ioctl(fd, SIOCGIFHWADDR, &ifr);
	memcpy(source_mac, ifr.ifr_hwaddr.sa_data, 6);
	close(fd);
}

int Port::getPortId() {
	return _portId;
}

char* Port::getRawDestMacAddress() {
	return destination_mac;
}

void Port::setLinkSpeed(double d) {
	transmissionSpeed = d;
}

