/*
 * Port.cpp
 *
 *      Author: Valentin Fitz
 *   Institute: University of Duisburg-Essen, Germany
 */

#include "Port.h"

#include <unistd.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <linux/if_ether.h>
#include <netinet/in.h>
#include <net/if.h>

#include "EtherTypes.h"
#include "ProtocolLogic.h"

using namespace std;

void* Port::_ingPortThreadEntryFunction(void* This) {
	((Port *) This)->_startIngPortThread();
	return NULL;
}

void Port::_startIngPortThread() {
	BpiOutStream bpiout(cout);
	BpiOutStream bpierr(cerr);

	stringstream ss;
	ss << "ingPort_" << _portId;
	pthread_setname_np(pthread_self(), ss.str().c_str());

	_setupPcapListener();
	bpiout << "Port ing_" << _portId << " started" << endl;
	lasttval.tv_sec = -1;
	lasttval.tv_usec = -1;
	while (active) {
		pcap_dispatch(pcap_port, 1, Port::_packetHandlerCallback, (unsigned char*) this);
	}
	bpiout << "Port ing_" << _portId << " terminated" << endl;
}

struct thread_data {
	Port* This;
	EthernetMessage* ethMsg;
};


void Port::_packetHandlerCallback(u_char *user, const struct pcap_pkthdr* header, const u_char* pa) {
	((Port*) user)->_packetHandler(header, pa);
}


//void* Port::_packetHandlerThreadEntry( void *threadarg )
//{
//	thread_data *td;
//	td = (thread_data*) threadarg;
//	BpiOutStream bout(cout);
//	bout << "Thread executed" << std::endl;
//
//	((Port*)td->This)->_threadPacketHandler(td->ethMsg);
//
//}
//
//void Port::_threadPacketHandler(EthernetMessage* eth){
//	EthernetMessage* ethmsg = new EthernetMessage();
//	ethmsg = copyEthernetMessage(eth);
//	deleteMessage(eth);
//	if (ethmsg->ether_type == ETH_TYPE_ACK) {
//		pthread_mutex_lock(&_ackMutex);
//		_ackMsg = ethmsg;
//		pthread_cond_signal(&_ackCond);
//		pthread_mutex_unlock(&_ackMutex);
//	} else {
//		if (ACK == 1) {
//			for (uint8_t i = 0; i < _regEthTypes.size(); i++) {
//				if (_regEthTypes.at(i) == ethmsg->ether_type) {
//					_sendAck(ethmsg);
//				}
//			}
//		}
//		rxMsg++;
//		ethmsg->meta = new EthMsgMeta;
//		((EthMsgMeta*) ethmsg->meta)->port = _portId;
//		_handleIncomingPaketFI(ethmsg);
//		protocol->portInput(ethmsg);
//	}
//
//	pthread_exit(NULL);
//}

void Port::_packetHandler(const struct pcap_pkthdr* header, const u_char* pa) {
	if (pa != nullptr && (lasttval.tv_sec != header->ts.tv_sec || lasttval.tv_usec != header->ts.tv_usec)) {
		statRx++;
		EthernetMessage* ethmsg = deserializeEthernetMessage((const char*) pa, header->caplen);
		lasttval = header->ts;
		if (ethmsg->ether_type == ETH_TYPE_ACK) {
			pthread_mutex_lock(&_ackMutex);
			_ackMsg = ethmsg;
			pthread_cond_signal(&_ackCond);
			pthread_mutex_unlock(&_ackMutex);
		} else {
			if (ACK == 1) {
				for (uint8_t i = 0; i < _regEthTypes.size(); i++) {
					if (_regEthTypes.at(i) == ethmsg->ether_type) {
						_sendAck(ethmsg);
					}
				}
			}
			rxMsg++;
			ethmsg->meta = new EthMsgMeta;
			((EthMsgMeta*) ethmsg->meta)->port = _portId;
			_handleIncomingPaketFI(ethmsg);
			protocol->portInput(ethmsg);
			lasttval = header->ts;
		}
	} else if (pa != nullptr) {
		bpierr << "Timeval equal" << std::endl;
		bpierr << printMessage(deserializeEthernetMessage((const char*) pa, header->caplen)) << std::endl;
	}
}

void Port::_setupPcapListener() {
	BpiOutStream bpierr(cerr);

	pcap_port = pcap_open_live(_devName.c_str(), 1518, 1, 0, error_buffer);

//	pcap_port = pcap_create(_devName.c_str(), error_buffer);
//	if (pcap_port == nullptr) {
//		bpierr << error_buffer << std::endl;
//		active = false;
//		return;
//	}
//	if (pcap_set_immediate_mode(pcap_port, 1) != 0) {
//		bpierr << "Could not set immediate mode!" << std::endl;
//	}
//	if (pcap_set_promisc(pcap_port, 1) != 0) {
//		bpierr << "Could not set promisc mode!" << std::endl;
//	}
//	int act = pcap_activate(pcap_port);
//	if (act != 0) {
//		bpierr << "activate packet capture handle!" << std::endl;
//	}

	if (pcap_setdirection(pcap_port, PCAP_D_IN) == -1) {
		bpierr << "Could not specify port direction on device " << _devName << std::endl;
		bpierr << "This means outgoing traffic is treated as incoming traffic." << std::endl;
	}
}

void Port::_sendAck(EthernetMessage* msg) {
	Message* ack = serializeEthernetMessage(createEthernetMessage(msg->destination, source_mac, ETH_TYPE_ACK, createMessageData("ACK", 3, ETHDATAMIN)));
	sendto(raw_port, ack->data, ack->size, 0, (struct sockaddr*) &raw_port_address, sizeof(struct sockaddr_ll));
}

bool Port::_handleIncomingPaket(EthernetMessage* msg) {
	return true;
}

bool Port::_handleIncomingPaketFI(EthernetMessage*msg) {
	return _handleIncomingPaket(msg);
}

//void Port::_putInBuffer(struct pcapPacket p){
//	pthread_mutex_lock(&_mutexBuf);
//	_packetBuffer.push_back(p);
//	pthread_cond_signal(&_bufCond);
//	pthread_mutex_unlock(&_mutexBuf);
//	bpiout << "Put new packet" << std::endl;
//}
//
//struct pcapPacket Port::_getFromBuffer() {
//	pthread_mutex_lock(&_mutexBuf);
//	struct pcapPacket p = _packetBuffer.at(0);
//	_packetBuffer.erase(_packetBuffer.begin()+0);
//	pthread_mutex_unlock(&_mutexBuf);
//	return p;
//}
