/*
 * ProtocolLogic.h
 *
 *      Author: Valentin Fitz
 *   Institute: University of Duisburg-Essen, Germany
 */

#ifndef PROTOCOLS_PROTOCOLLOGIC_H_
#define PROTOCOLS_PROTOCOLLOGIC_H_

#include <pthread.h>
#include <vector>
#include <deque>
#include <string>
#include "Port.h"
#include "../GlobalBpiLib.hpp"
#include "MessageTypes.h"

/**
 * \brief Default implementation of the logic of a protocol implementation.
 */
class ProtocolLogic {
protected:
	Port **ports;
	std::vector<EthernetMessage*> protocol_messages;
	std::deque<std::string> control_messages;
	pthread_t protologic;
	pthread_mutex_t _mutex;
	pthread_cond_t _cond;
	bool run = true;
	void _runProtoLogic();
	void _initializeProtocol();
	void _sendMessage(EthernetMessage* msg, int outPorts); /**< DO NOT CALL THIS FUNCTION! Implementation of the sending routine which shall not be modified or derived from. */
	virtual Port* _createPort(int i);
	virtual void _executeProtocolLogic();
	virtual void _handleControlMessages();
	virtual void _handleControlMessage(std::string cmdId, std::string cmd);
	virtual void _handleControlMessageFI(std::string cmdId, std::string cmd);
	BpiOutStream bpiout; /**< Standard output printer. */
	BpiOutStream bpierr; /**< Error output printer. */
	char source_mac[6];
public:
	ProtocolLogic();
	virtual ~ProtocolLogic();
	static void * protoLogicEntry(void * This);
	void start();
	void stop();
	void portInput(EthernetMessage* input);
	void controlInput(std::string input);
	virtual void sendMessage(EthernetMessage* msg, int outPorts); /**< Sends msg (adds msg to the coresponding port queues) via ports specified by the binary representation of the parameter ports. */
};

#endif /* PROTOCOLS_PROTOCOLLOGIC_H_ */
