/*
 * FIProtocolLogic.cpp
 *
 *  Created on: Mar 5, 2019
 *      Author: jubuntu
 */

#include "FIProtocolLogic.h"

#include <pthread.h>
#include <cmath>
#include <cstdlib>
#include <cstring>
#include <random>
#include <regex>
#include <sstream>
#include <string>
#include <thread>
#include <vector>

#include <bitset>
#include "../GlobalBpiLib.hpp"
#include "../Protocols/MessageTypes.h"

struct thread_args {
	FIProtocolLogic *This;
	int delayT;
	EthernetMEssage *msgT;
	int outPortsT;
};

FIProtocolLogic::FIProtocolLogic() {
	srand(time(NULL));
}

FIProtocolLogic::~FIProtocolLogic() {

}
void FIProtocolLogic::_handleControlMessageFI(std::string cmdId, std::string cmd) {

	// Split received control message (command from ConfigTool)
	std::vector<std::string> items = bpiSplitIntoItems(cmd, ' ');
	int itemsSize = items.size();

	std::regex regex0_100("((^([0-9]{1,2}){1}(\\.[0-9]{1,2})?$)|100|100\\.[0]+)");	//number between 00-100.00
	std::regex regexRange("(([0-9]+)|[0-9]+-[0-9]+)"); 								//number Range in form of either single number or 2 numbers with -
	std::regex regexMean("([0-9]+)");												//general number
	std::regex regexDouble("(([0-9]+)|[0-9]+\\.[0-9]+)");						//general number or number with a dot
	std::regex regex0_15("[0-9]|10|11|12|13|14|15");
	// Perform command handling here

	//###########   Bit Flips  #############// 			done!
	if (strncmp(items[0].c_str(), "fi_jn_bf", 8) == 0) { //amount of parameters determined by the chosen probabilityDistribution

		this->bfProbabilitiyDistribution = StringToProbDist(items[1]);

		switch (bfProbabilitiyDistribution) {

		case discrete_equal_distribution: //needs 2 Parameters Percentage & Range
		{
			if (itemsSize == 4) { //number of Parameters
				if (std::regex_match(items[2], regex0_100)) { //Percentage between 0-100
					if (std::regex_match(items[3], regexRange)) { //Range is valid

//								bpiout<< "BF Percentage set to "<< items[2] <<  std::endl;
						this->bfPercentage = std::stod(items[2]) / 100;

						std::vector<std::string> bfRange = bpiSplitIntoItems(items[3], '-');
						if (bfRange.size() == 1) {
//										bpiout<< "min bpisplit = " << bfRange[0]<< " max bpiSplit  = " << bfRange[0]<< std::endl;
							this->bfRangeMin = std::atoi(bfRange[0].c_str());
							this->bfRangeMax = std::atoi(bfRange[0].c_str());
							bpiout << "FI_BF : p=" << this->bfPercentage << " / NUM~U([" << this->bfRangeMin << "," << this->bfRangeMax << "])" << std::endl;
						} else {
//										bpiout<< "min bpisplit = " << bfRange[0]<< " max bpiSplit  = " << bfRange[1]<< std::endl;
							this->bfRangeMin = std::atoi(bfRange[0].c_str());
							this->bfRangeMax = std::atoi(bfRange[1].c_str());
							bpiout << "FI_BF : p=" << this->bfPercentage << " / NUM~U([" << this->bfRangeMin << "," << this->bfRangeMax << "])" << std::endl;
						}

					} else {
						bpierr << "Please provide a valid Range" << std::endl;
					}
				} else {
					bpierr << "Please provide a percentage value between 00-100.00" << std::endl;
				}
			} else {
				bpierr << "Please provide the correct amount of Parameters for the BF Probabilitydistribution" << std::endl;
			}
		}
			break;

		case Gaussian_distribution: //needs 4 parameters Percentage, Range, mean(Erwartungswert) & standard deviation (Standardabweichung)
		{
			if (itemsSize == 6) { //number of Parameters
				if (std::regex_match(items[2], regex0_100)) { //Percentage between 0-100
					if (std::regex_match(items[3], regexRange)) { //Range is valid
						if (std::regex_match(items[4], regexMean)) { //Mean is valid
							if (std::regex_match(items[5], regexDouble)) { //Std Deviation is valid
//										bpiout << "BF Percentage set to " << items[2] << std::endl;
								this->bfPercentage = std::stod(items[2]) / 100;
//										bpiout << "BF Mean set to " << items[4] << std::endl;
								this->bfMean = std::atoi(items[4].c_str());
//										bpiout << "BF Standard deviation set to " << items[5] << std::endl;
								this->bfStandardDeviation = std::stod(items[5]);

								//setting range is influenced by either giving a single number or a range
								std::vector<std::string> bfRange = bpiSplitIntoItems(items[3], '-');
								if (bfRange.size() == 1) {
//												bpiout << "min bpisplit = " << bfRange[0] << " max bpiSplit  = " << bfRange[0] << std::endl;
									this->bfRangeMin = std::atoi(bfRange[0].c_str());
									this->bfRangeMax = std::atoi(bfRange[0].c_str());
								} else {
//												bpiout << "min bpisplit = " << bfRange[0]<< " max bpiSplit  = " << bfRange[1]	<< std::endl;
									this->bfRangeMin = std::atoi(bfRange[0].c_str());
									this->bfRangeMax = std::atoi(bfRange[1].c_str());
								}
								bpiout << "FI_BF : p=" << this->bfPercentage << " / NUM~N(\u03bc=" << this->bfMean << ",\u03c3=" << this->bfStandardDeviation << ") / Min=" << this->bfRangeMin
										<< " / Max=" << this->bfRangeMax << std::endl;
							} else {
								bpierr << "Please provide a valid Standard deviation" << std::endl;
							}
						} else {
							bpierr << "Please provide a valid Mean" << std::endl;
						}
					} else {
						bpierr << "Please provide a valid Range" << std::endl;
					}
				} else {
					bpierr << "Please provide a percentage value between 00-100.00" << std::endl;
				}
			} else {
				bpierr << "Please provide the correct amount of Parameters for the BF Probabilitydistribution" << std::endl;
			}
		}
			break;
		case Exponential_distribution: //needs 3 Parameters Percentage & lamda & max value
		{
			if (itemsSize == 5) { //number of Parameters
				if (std::regex_match(items[2], regex0_100)) { //Percentage between 0-100
					if (std::regex_match(items[3], regexDouble)) { //Range is valid
						if (std::regex_match(items[4], regexRange)) { //Range is valid

//												bpiout<< "BF Percentage set to "<< items[2] <<  std::endl;
							this->bfPercentage = std::stod(items[2]) / 100;
							this->bfLambda = std::stod(items[3]);
							std::vector<std::string> bfRange = bpiSplitIntoItems(items[4], '-');
							if (bfRange.size() == 1) {
//														bpiout<< "min bpisplit = " << bfRange[0]<< " max bpiSplit  = " << bfRange[0]<< std::endl;
								this->bfRangeMin = std::atoi(bfRange[0].c_str());
								this->bfRangeMax = std::atoi(bfRange[0].c_str());
							} else {
//														bpiout<< "min bpisplit = " << bfRange[0]<< " max bpiSplit  = " << bfRange[1]<< std::endl;
								this->bfRangeMin = std::atoi(bfRange[0].c_str());
								this->bfRangeMax = std::atoi(bfRange[1].c_str());
							}
							bpiout << "FI_BF : p=" << this->bfPercentage << " / NUM~Exp(\u03bb=" << this->bfLambda << ") / Min=" << this->bfRangeMin << " / Max=" << this->bfRangeMax << std::endl;

						} else {
							bpierr << "Please provide a valid Range" << std::endl;
						}
					} else {
						bpierr << "Please provide a valid value for lambda" << std::endl;
					}
				} else {
					bpierr << "Please provide a percentage value between 00-100.00" << std::endl;
				}
			} else {
				bpierr << "Please provide the correct amount of Parameters for the BF Exponential_distribution" << std::endl;
			}
		}
			break;

		default:
			break;
		}
	}

	//###########   Wrong Forwarding  #############// done!

	else if (strncmp(items[0].c_str(), "fi_jn_wf", 8) == 0) {

		if (itemsSize == 3) { 							//number of Parameters needs 2 Parameters Percentage and WFPortSet
			if (std::regex_match(items[1], regex0_100)) { 	//Percentage between 0-100
				if (std::regex_match(items[2], regex0_15)) {
					this->wfPercentage = std::stod(items[1]);
					this->wfPortSet = std::atoi(items[2].c_str());
					bpiout << "FI_WF : p=" << this->wfPercentage << " / PossiblePortTargets=" << this->wfPortSet << std::endl;
				} else {
					bpierr << "Please provide a number for the ports between 0-15!" << std::endl;
				}
			} else {
				bpierr << "Please provide a percentage value between 00-100.00" << std::endl;
			}
		} else {
			bpierr << "Please provide the correct amount of Parameters for the Wrong Forwarding" << std::endl;
		}
	}

	//###########	Duplication		##############// done!

	else if (strncmp(items[0].c_str(), "fi_jn_duplication", 17) == 0) {
		this->duplicationProbabilitiyDistribution = StringToProbDist(items[1]);

		switch (duplicationProbabilitiyDistribution) {

		case discrete_equal_distribution: //needs 2 Parameters Percentage & Range
		{
			if (itemsSize == 4) { //number of Parameters
				if (std::regex_match(items[2], regex0_100)) { //Percentage between 0-100
					if (std::regex_match(items[3], regexRange)) { //Range is valid

//										bpiout<< "Duplication Percentage set to "<< items[2] <<  std::endl;
						this->duplicationPercentage = std::stod(items[2]) / 100;

						std::vector<std::string> duplicationRange = bpiSplitIntoItems(items[3], '-');
						if (duplicationRange.size() == 1) {
//												bpiout<< "min bpisplit = " << duplicationRange[0]<< " max bpiSplit  = " << duplicationRange[0]<< std::endl;
							this->duplicationRangeMin = std::atoi(duplicationRange[0].c_str());
							this->duplicationRangeMax = std::atoi(duplicationRange[0].c_str());
						} else {
//												bpiout<< "min bpisplit = " << duplicationRange[0]<< " max bpiSplit  = " << duplicationRange[1]<< std::endl;
							this->duplicationRangeMin = std::atoi(duplicationRange[0].c_str());
							this->duplicationRangeMax = std::atoi(duplicationRange[1].c_str());
						}
						bpiout << "FI_DUP: p=" << this->duplicationPercentage << " / NUM~U([" << this->duplicationRangeMin << "," << this->duplicationRangeMax << "])" << std::endl;
					} else {
						bpierr << "Please provide a valid Range" << std::endl;
					}
				} else {
					bpierr << "Please provide a percentage value between 00-100.00" << std::endl;
				}
			} else {
				bpierr << "Please provide the correct amount of Parameters for the Duplication Probabilitydistribution" << std::endl;
			}
		}
			break;

		case Gaussian_distribution: //needs 4 parameters Percentage, Range, mean(Erwartungswert) & standard deviation (Standardabweichung)
		{
			if (itemsSize == 6) { //number of Parameters
				if (std::regex_match(items[2], regex0_100)) { //Percentage between 0-100
					if (std::regex_match(items[3], regexRange)) { //Range is valid
						if (std::regex_match(items[4], regexMean)) { //Mean is valid
							if (std::regex_match(items[5], regexDouble)) { //Std Deviation is valid
//												bpiout << "Duplication Percentage set to " << items[2] << std::endl;
								this->duplicationPercentage = std::stod(items[2]) / 100;
//												bpiout << "Duplication Mean set to " << items[4] << std::endl;
								this->duplicationMean = std::atoi(items[4].c_str());
//												bpiout << "Duplication Standard deviation set to " << items[5] << std::endl;
								this->duplicationStandardDeviation = std::stod(items[5]);

								//setting range is influenced by either giving a single number or a range
								std::vector<std::string> duplicationRange = bpiSplitIntoItems(items[3], '-');
								if (duplicationRange.size() == 1) {
//														bpiout << "min bpisplit = " << duplicationRange[0] << " max bpiSplit  = " << duplicationRange[0] << std::endl;
									this->duplicationRangeMin = std::atoi(duplicationRange[0].c_str());
									this->duplicationRangeMax = std::atoi(duplicationRange[0].c_str());
								} else {
//														bpiout << "min bpisplit = " << duplicationRange[0]<< " max bpiSplit  = " << duplicationRange[1]	<< std::endl;
									this->duplicationRangeMin = std::atoi(duplicationRange[0].c_str());
									this->duplicationRangeMax = std::atoi(duplicationRange[1].c_str());
								}
								bpiout << "FI_DUP: p=" << this->duplicationPercentage << " / NUM~N([\u03bc=" << this->duplicationMean << ",\u03c3=" << this->duplicationStandardDeviation << "]) / Min="
										<< this->duplicationRangeMin << " / Max=" << this->duplicationRangeMax << std::endl;
							} else {
								bpierr << "Please provide a valid Standard deviation" << std::endl;
							}
						} else {
							bpierr << "Please provide a valid Mean" << std::endl;
						}
					} else {
						bpierr << "Please provide a valid Range" << std::endl;
					}
				} else {
					bpierr << "Please provide a percentage value between 00-100.00" << std::endl;
				}
			} else {
				bpierr << "Please provide the correct amount of Parameters for the Duplication Probabilitydistribution" << std::endl;
			}
		}
			break;

		case Exponential_distribution: //needs 3 Parameters Percentage & lamda & max value
		{
			if (itemsSize == 5) { //number of Parameters
				if (std::regex_match(items[2], regex0_100)) { //Percentage between 0-100
					if (std::regex_match(items[3], regexDouble)) { //Range is valid
						if (std::regex_match(items[4], regexRange)) { //Range is valid

//														bpiout<< "Duplication Percentage set to "<< items[2] <<  std::endl;
							this->duplicationPercentage = std::stod(items[2]) / 100;
							this->duplicationLambda = std::stod(items[3]);
							std::vector<std::string> duplicationRange = bpiSplitIntoItems(items[4], '-');
							if (duplicationRange.size() == 1) {
//																bpiout<< "min bpisplit = " << duplicationRange[0]<< " max bpiSplit  = " << duplicationRange[0]<< std::endl;
								this->duplicationRangeMin = std::atoi(duplicationRange[0].c_str());
								this->duplicationRangeMax = std::atoi(duplicationRange[0].c_str());
							} else {
//																bpiout<< "min bpisplit = " << duplicationRange[0]<< " max bpiSplit  = " << duplicationRange[1]<< std::endl;
								this->duplicationRangeMin = std::atoi(duplicationRange[0].c_str());
								this->duplicationRangeMax = std::atoi(duplicationRange[1].c_str());
							}
							bpiout << "FI_DUP: p=" << this->duplicationPercentage << " / NUM~Exp(\u03bb=" << this->duplicationLambda << ") / Min=" << this->duplicationRangeMin << " / Max="
									<< this->duplicationRangeMax << std::endl;
						} else {
							bpierr << "Please provide a valid Range" << std::endl;
						}
					} else {
						bpierr << "Please provide a valid value for lambda" << std::endl;
					}
				} else {
					bpierr << "Please provide a percentage value between 00-100.00" << std::endl;
				}
			} else {
				bpierr << "Please provide the correct amount of Parameters for the Duplication Exponential_distribution" << std::endl;
			}
		}
			break;

		default:
			break;
		}

	}

	//###########   	Omission		#############//		done!
	else if ((strncmp(items[0].c_str(), "fi_jn_omissionPercentage", 24) == 0) && (itemsSize == 2)) {

		if (std::regex_match(items[1], regex0_100)) {
			//			solution with values on the Ports
			//			for (int i = 0; i < 4; i++) {
			//				dynamic_cast<FIPort*>(ports[i])->setmessageDropPercentage(d_stod);
			//			}
//			bpiout << "OmissionPercentage changed to " << items[1] <<  std::endl;
			this->omissionPercentage = (std::stod(items[1])) / 100;
			bpiout << "FI_OM : p=" << this->omissionPercentage << std::endl;

		} else {
			bpierr << "Please provide a percentage value between 00-100.00" << std::endl;
		}
	}

	//###########   ByzantineBehavior  #############//		done!
	else if ((strncmp(items[0].c_str(), "fi_jn_byzantineBehavior", 23) == 0) && (itemsSize == 2)) {
		if (std::regex_match(items[1], std::regex("true|false"))) {
			this->ByzantineBehavior = strncmp(items[1].c_str(), "false", 5); //strncmp returns 0 if the strings match otherwise 1
			bpiout << "FI_BYZ: " << items[1] << std::endl;
		} else {
			bpierr << "Please provide the argument true or false!" << std::endl;
		}
	}

	//###########   Crash Percentage  #############//			done!
	else if ((strncmp(items[0].c_str(), "fi_jn_crashPercentage", 21) == 0) && (itemsSize == 2)) {
		if (std::regex_match(items[1], regex0_100)) {
//			bpiout << "crashPercentage changed to " << items[1] << std::endl;
			this->crashPercentage = (std::stod(items[1])) / 100;
			bpiout << "FI_C  : p=" << this->crashPercentage << std::endl;
		} else {
			bpierr << "Please provide a percentage value between 00-100.00" << std::endl;
		}
	}

	//###########   set Crash false  #############//			done!
	else if ((strncmp(items[0].c_str(), "fi_jn_crashedFalse", 18) == 0) && (itemsSize == 1)) {
		this->crashed = false;
		bpiout << "crashed set to false!" << std::endl;
	}

	//###########   DELAY  #############// 			done!
	else if (strncmp(items[0].c_str(), "fi_jn_delay", 10) == 0) { //amount of parameters determined by the chosen probabilityDistribution

		this->delayProbabilitiyDistribution = StringToProbDist(items[1]);

		switch (delayProbabilitiyDistribution) {

		case discrete_equal_distribution: //needs 2 Parameters Percentage & Range
		{
			if (itemsSize == 4) { //number of Parameters
				if (std::regex_match(items[2], regex0_100)) { //Percentage between 0-100
					if (std::regex_match(items[3], regexRange)) { //Range is valid

//							bpiout<< "delay Percentage set to "<< items[2] <<  std::endl;
						this->delayPercentage = std::stod(items[2]) / 100;

						std::vector<std::string> delayRange = bpiSplitIntoItems(items[3], '-');
						if (delayRange.size() == 1) {
//									bpiout<< "min bpisplit = " << delayRange[0]<< " max bpiSplit  = " << delayRange[0]<< std::endl;
							this->delayRangeMin = std::atoi(delayRange[0].c_str());
							this->delayRangeMax = std::atoi(delayRange[0].c_str());
						} else {
//									bpiout<< "min bpisplit = " << delayRange[0]<< " max bpiSplit  = " << delayRange[1]<< std::endl;
							this->delayRangeMin = std::atoi(delayRange[0].c_str());
							this->delayRangeMax = std::atoi(delayRange[1].c_str());
						}
						bpiout << "FI_DEL: p=" << this->delayPercentage << " / NUM~U([" << this->delayRangeMin << "," << this->delayRangeMax << "])" << std::endl;

					} else {
						bpierr << "Please provide a valid Range" << std::endl;
					}
				} else {
					bpierr << "Please provide a percentage value between 00-100.00" << std::endl;
				}
			} else {
				bpierr << "Please provide the correct amount of Parameters for the delay Probabilitydistribution" << std::endl;
			}
		}
			break;

		case Gaussian_distribution: //needs 4 parameters Percentage, Range, mean(Erwartungswert) & standard deviation (Standardabweichung)
		{
			if (itemsSize == 6) { //number of Parameters
				if (std::regex_match(items[2], regex0_100)) { //Percentage between 0-100
					if (std::regex_match(items[3], regexRange)) { //Range is valid
						if (std::regex_match(items[4], regexMean)) { //Mean is valid
							if (std::regex_match(items[5], regexDouble)) { //Std Deviation is valid
//									bpiout << "delay Percentage set to " << items[2] << std::endl;
								this->delayPercentage = std::stod(items[2]) / 100;
//									bpiout << "delay Mean set to " << items[4] << std::endl;
								this->delayMean = std::atoi(items[4].c_str());
//									bpiout << "delay Standard deviation set to " << items[5] << std::endl;
								this->delayStandardDeviation = std::stod(items[5]);

								//setting range is influenced by either giving a single number or a range
								std::vector<std::string> delayRange = bpiSplitIntoItems(items[3], '-');
								if (delayRange.size() == 1) {
//											bpiout << "min bpisplit = " << delayRange[0] << " max bpiSplit  = " << delayRange[0] << std::endl;
									this->delayRangeMin = std::atoi(delayRange[0].c_str());
									this->delayRangeMax = std::atoi(delayRange[0].c_str());
								} else {
//											bpiout << "min bpisplit = " << delayRange[0]<< " max bpiSplit  = " << delayRange[1]	<< std::endl;
									this->delayRangeMin = std::atoi(delayRange[0].c_str());
									this->delayRangeMax = std::atoi(delayRange[1].c_str());
								}
								bpiout << "FI_DEL: p=" << this->delayPercentage << " / NUM~N([\u03bc=" << this->delayMean << ",\u03c3=" << this->delayStandardDeviation << "]) / Min="
										<< this->delayRangeMin << " / Max=" << this->delayRangeMax << std::endl;
							} else {
								bpierr << "Please provide a valid Standard deviation" << std::endl;
							}
						} else {
							bpierr << "Please provide a valid Mean" << std::endl;
						}
					} else {
						bpierr << "Please provide a valid Range" << std::endl;
					}
				} else {
					bpierr << "Please provide a percentage value between 00-100.00" << std::endl;
				}
			} else {
				bpierr << "Please provide the correct amount of Parameters for the delay Probabilitydistribution" << std::endl;
			}
		}
			break;

		case Exponential_distribution: //needs 3 Parameters Percentage & lamda & max value
		{
			if (itemsSize == 5) { //number of Parameters
				if (std::regex_match(items[2], regex0_100)) { //Percentage between 0-100
					if (std::regex_match(items[3], regexDouble)) { //Range is valid
						if (std::regex_match(items[4], regexRange)) { //Range is valid

//											bpiout<< "BF Percentage set to "<< items[2] <<  std::endl;
							this->delayPercentage = std::stod(items[2]) / 100;
							this->delayLambda = std::stod(items[3]);
							std::vector<std::string> delayRange = bpiSplitIntoItems(items[4], '-');
							if (delayRange.size() == 1) {
//													bpiout<< "min bpisplit = " << delayRange[0]<< " max bpiSplit  = " << delayRange[0]<< std::endl;
								this->delayRangeMin = std::atoi(delayRange[0].c_str());
								this->delayRangeMax = std::atoi(delayRange[0].c_str());
							} else {
//													bpiout<< "min bpisplit = " << delayRange[0]<< " max bpiSplit  = " << delayRange[1]<< std::endl;
								this->delayRangeMin = std::atoi(delayRange[0].c_str());
								this->delayRangeMax = std::atoi(delayRange[1].c_str());
							}
							bpiout << "FI_DEL: p=" << this->delayPercentage << " / NUM~Exp(\u03bb=" << this->delayLambda << ") / Min=" << this->delayRangeMin << " / Max=" << this->delayRangeMax
									<< std::endl;

						} else {
							bpierr << "Please provide a valid Range" << std::endl;
						}
					} else {
						bpierr << "Please provide a valid value for lambda" << std::endl;
					}
				} else {
					bpierr << "Please provide a percentage value between 00-100.00" << std::endl;
				}
			} else {
				bpierr << "Please provide the correct amount of Parameters for the BF Exponential_distribution" << std::endl;
			}
		}
			break;

		default:
			break;
		}
	}

	//###########   Output  #############//
	else if (strncmp(items[0].c_str(), "fi_jn_output", 12) == 0) {

		std::stringstream ss;

		ss << " The current parameter values are: " << std::endl;

		ss << "bfPercentage 						= " << bfPercentage << std::endl;
		ss << "bfProbabilitiyDistribution 			= " << bfProbabilitiyDistribution << std::endl;
		ss << "bfRangeMin 							= " << bfRangeMin << std::endl;
		ss << "bfRangeMax 							= " << bfRangeMax << std::endl;
		ss << "bfMean 								= " << bfMean << std::endl;
		ss << "bfStandardDeviation 					= " << bfStandardDeviation << std::endl;
		ss << "bfLambda 							= " << bfLambda << std::endl;

		ss << "wfPercentage 						= " << wfPercentage << std::endl;
		ss << "wfPortSet 							= " << wfPortSet << std::endl;

		ss << "duplicationPercentage 				= " << duplicationPercentage << std::endl;
		ss << "duplicationProbabilitiyDistribution	= " << duplicationProbabilitiyDistribution << std::endl;
		ss << "duplicationRangeMin 					= " << duplicationRangeMin << std::endl;
		ss << "duplicationRangeMax 					= " << duplicationRangeMax << std::endl;
		ss << "duplicationMean 						= " << duplicationMean << std::endl;
		ss << "duplicationStandardDeviation			= " << duplicationStandardDeviation << std::endl;
		ss << "duplicationLambda					= " << duplicationLambda << std::endl;

		ss << "omissionPercentage 					= " << omissionPercentage << std::endl;

		ss << "crashed 								= " << crashed << std::endl;
		ss << "crashPercentage 						= " << crashPercentage << std::endl;

		ss << "delayPercentage 						= " << delayPercentage << std::endl;
		ss << "delayProbabilitiyDistribution		= " << delayProbabilitiyDistribution << std::endl;
		ss << "delayRangeMin						= " << delayRangeMin << std::endl;
		ss << "delayRangeMax						= " << delayRangeMax << std::endl;
		ss << "delayMean							= " << delayMean << std::endl;
		ss << "delayStandardDeviation				= " << delayStandardDeviation << std::endl;
		ss << "delayLambda 							= " << delayLambda << std::endl;

		ss << "ByzantineBehavior 					= " << ByzantineBehavior << std::endl;

		bpiout << ss.str() << std::endl;

	}

	else {
		_handleControlMessage(cmdId, cmd);
	}
}

void FIProtocolLogic::sendMessage(EthernetMessage *msg, int outPorts, bool checkForDuplication) {
	//Random engine
	unsigned seed = std::chrono::steady_clock::now().time_since_epoch().count();
	std::default_random_engine eng(seed);

	int actualNumberOfBFArray[4] = { };
	int actualNumberOfDuplicates = 0;
	int actualSendingDelayArray[4] = { };

	//first check if the entire node has crashed or is crashing
	//###########   Crash		  #############//
	if (((double) rand() / RAND_MAX) < crashPercentage) {
		bpiout << "Node has crashed" << std::endl;
		crashed = true;
	}

	if (crashed) {
		//Do nothing because Node has crashed
		bpiout << "Node broken" << std::endl;
	}

	else {
		//Node has not crashed

		//###########   Duplication		  #############//
		if (duplicationRangeMax > 0 && checkForDuplication) {
			if (((double) rand() / RAND_MAX) < duplicationPercentage) {
				switch (duplicationProbabilitiyDistribution) {

				case discrete_equal_distribution: {
					std::uniform_int_distribution<int> distro(duplicationRangeMin, duplicationRangeMax);
					actualNumberOfDuplicates = distro(eng);
					bpiout << "sending  " << actualNumberOfDuplicates << " duplicates" << std::endl;
				}
					break;

				case Gaussian_distribution: {
					std::normal_distribution<double> distro(duplicationMean, duplicationStandardDeviation);
					do {
						actualNumberOfDuplicates = (int) ((distro(eng)) + 0.5);
					} while ((duplicationRangeMin > actualNumberOfDuplicates) || (actualNumberOfDuplicates > duplicationRangeMax));
					bpiout << "sending " << actualNumberOfDuplicates << " duplicates" << std::endl;
				}
					break;
				case Exponential_distribution: {
					std::exponential_distribution<double> distro(duplicationLambda);
					do {
						actualNumberOfDuplicates = (int) ((distro(eng)) + 0.5);
					} while (actualNumberOfDuplicates > duplicationRangeMax);
					bpiout << "sending " << actualNumberOfDuplicates << " duplicates" << std::endl;
				}
					break;
				}
				// sending with flag to not check for duplication again!
				if (actualNumberOfDuplicates > 0) {
					for (int var = 0; var < actualNumberOfDuplicates; var++) {
						EthernetMessage *msgduplicate = copyEthernetMessage(msg);
						sendMessage(msgduplicate, outPorts, false);
					}
				}
			}	//else --> nothing to do because not triggered
		}	//else --> nothing to do because values provided in duplicationRangeMax or checkforDuplication is set to false

		//###########   Wrong Forwarding  #############//
		if (((double) rand() / RAND_MAX) < wfPercentage) {
			if (wfPortSet == outPorts) { //check if the PortSet is identical to the addressed Ports if so... no change can be made.
			} else {
				int wfOutPorts = 0;
				do {
					int p = 0;
					int tempwfPortSet = wfPortSet;

					for (int i = 0; i < 4; i++) {
						p = tempwfPortSet % 2;
						tempwfPortSet = tempwfPortSet / 2;
						if (p == 1) {
							if (((double) rand() / RAND_MAX) < 0.5) {
								bpiout << "not sending on Port " << i + 1 << std::endl;
							} else {
								wfOutPorts += std::pow(2, i);
							}
						}
					}
				} while (outPorts == wfOutPorts);
				outPorts = wfOutPorts;
			}
		} //else --> change nothing, wfPercentage not triggered

		//###########   Omission Percentage  #############//
		if (!ByzantineBehavior) {
			double value = (double) rand() / RAND_MAX;
//			std::stringstream ss;
//			ss <<value<<std::endl;
//			bpiout << ss.str() << std::endl;
//			bpiout<< value<<std::endl;
			if (value < omissionPercentage) {
				//bpiout<<"All Messages dropped the value is "<< value  << std::endl;
				_sendMessage(copyEthernetMessage(msg), 0); //All Messages are dropped so the function can be called here
				return; //No further faults can occur because the Node is not sending any Messages anyway
			}
			//else --> change nothing no messages are being dropped
		}
		//Messages can be dropped on some Ports and on some still be sent
		else {
			int p = 0;
			int omissionOutPorts = 0;
			int tempOutPorts = outPorts;
			for (int i = 0; i < 4; i++) {
				p = tempOutPorts % 2;
				tempOutPorts = tempOutPorts / 2;
				if (p == 1) {
					if (((double) rand() / RAND_MAX) < omissionPercentage) {
						bpiout << "Messages dropped on Port " << i + 1 << std::endl;
					} else {
						omissionOutPorts += std::pow(2, i);
					}
				}
			}
			outPorts = omissionOutPorts; //The message is going to be send to those Ports which shall not be dropped
		}

		//###########   	Bitflip		#############//
		if (bfRangeMax > 0) {
//			if(((double) rand() / RAND_MAX) < bfPercentage){
			int bfLoopCounter = ByzantineBehavior ? 4 : 1;
			switch (bfProbabilitiyDistribution) {

			case discrete_equal_distribution: {
				std::uniform_int_distribution<int> distro(bfRangeMin, bfRangeMax);
				for (int i = 0; i < bfLoopCounter; i++) {
					if (((double) rand() / RAND_MAX) < bfPercentage) {
						actualNumberOfBFArray[i] = distro(eng);
						bpiout << "number of Bitflips " << actualNumberOfBFArray[i] << std::endl;
					}
				}

			}
				break;

			case Gaussian_distribution: {
				std::normal_distribution<double> distro(bfMean, bfStandardDeviation);
				for (int i = 0; i < bfLoopCounter; i++) {
					if (((double) rand() / RAND_MAX) < bfPercentage) {
						do {
							actualNumberOfBFArray[i] = (int) (distro(eng) + 0.5);
						} while ((bfRangeMin > actualNumberOfBFArray[i]) || (actualNumberOfBFArray[i] > bfRangeMax));
						bpiout << "number of Bitflips " << actualNumberOfBFArray[i] << std::endl;
					}
				}
			}
				break;

			case Exponential_distribution: {
				std::exponential_distribution<double> distro(bfLambda);
				for (int i = 0; i < bfLoopCounter; i++) {
					if (((double) rand() / RAND_MAX) < bfPercentage) {
						do {
							actualNumberOfBFArray[i] = (int) (distro(eng) + 0.5);
						} while (actualNumberOfBFArray[i] > bfRangeMax);
						bpiout << "number of Bitflips " << actualNumberOfBFArray[i] << std::endl;
					}
				}
			}
				break;
			}
//			}
			//else--> change nothing, BFPercentage not triggered
		}			//else --> change nothing, bfRangeMax not greater than 0

		//###########   Delay Percentage  #############//
		if (delayRangeMax > 0) {
//			if(((double) rand() / RAND_MAX) < delayPercentage){
			int delayLoopCounter = ByzantineBehavior ? 4 : 1;
			switch (delayProbabilitiyDistribution) {

			case discrete_equal_distribution: {
				std::uniform_int_distribution<int> distro(delayRangeMin, delayRangeMax);
				for (int i = 0; i < delayLoopCounter; i++) {
					if (((double) rand() / RAND_MAX) < delayPercentage) {
						actualSendingDelayArray[i] = distro(eng); //This means that the values provided represents milliseconds
						bpiout << "waiting for " << actualSendingDelayArray[i] << "milliseconds" << std::endl;
					}
				}
			}
				break;

			case Gaussian_distribution: {
				std::normal_distribution<double> distro(delayMean, delayStandardDeviation);
				for (int i = 0; i < delayLoopCounter; i++) {
					if (((double) rand() / RAND_MAX) < delayPercentage) {
						do {
							actualSendingDelayArray[i] = (int) ((distro(eng)) + 0.5);
						} while ((delayRangeMin > actualSendingDelayArray[i]) || (actualSendingDelayArray[i] > delayRangeMax));
						bpiout << "waiting for " << actualSendingDelayArray[i] << "milliseconds" << std::endl;
					}
				}
			}
				break;

			case Exponential_distribution: {
				std::exponential_distribution<double> distro(delayLambda);
				for (int i = 0; i < delayLoopCounter; i++) {
					if (((double) rand() / RAND_MAX) < delayPercentage) {
						do {
							actualSendingDelayArray[i] = (int) ((distro(eng)) + 0.5);
						} while (actualSendingDelayArray[i] > delayRangeMax);
						bpiout << "waiting for " << actualSendingDelayArray[i] << "milliseconds" << std::endl;
					}
				}
			}
				break;

			}
//			}
		}

		if (ByzantineBehavior) {
			int a = 0;
			for (int i = 0; i < 4; i++) {
				a = outPorts % 2;
				outPorts = outPorts / 2;
				if (a == 1) {
					if (actualSendingDelayArray[i] > 0) {
						bpiout << "create new sending Thread" << std::endl;
						thread_args *sendingargs = new thread_args();
						sendingargs->This = this;
						sendingargs->delayT = actualSendingDelayArray[i];
						sendingargs->msgT = (actualNumberOfBFArray[i] > 0) ? createBFmsg(msg, actualNumberOfBFArray[i]) : copyEthernetMessage(msg); //use BF message if actual number of BF is greater 0
						sendingargs->outPortsT = std::pow(2, i);
						pthread_t thread;
						bpiout << "create Thread" << std::endl;
						pthread_create(&thread, NULL, _threadEntryFunction, sendingargs);
						bpiout << "Thread created" << std::endl;
						pthread_detach(thread);

					} else {
						_sendMessage((actualNumberOfBFArray[i] > 0) ? createBFmsg(msg, actualNumberOfBFArray[i]) : copyEthernetMessage(msg), std::pow(2, i));
					}
				}
			}

		} else {
			//		###Sending in a different Thread because of delay###
			if (actualSendingDelayArray[0] > 0) {
				bpiout << "create new sending Thread" << std::endl;
				thread_args *sendingargs = new thread_args();
				sendingargs->This = this;
				sendingargs->delayT = actualSendingDelayArray[0];
				sendingargs->msgT = (actualNumberOfBFArray[0] > 0) ? createBFmsg(msg, actualNumberOfBFArray[0]) : copyEthernetMessage(msg); //use BF message if actual number of BF is greater 0
				sendingargs->outPortsT = outPorts;
				pthread_t thread;
				bpiout << "create Thread" << std::endl;
				pthread_create(&thread, NULL, _threadEntryFunction, sendingargs);
				bpiout << "Thread created" << std::endl;
				pthread_detach(thread);
			} else {
				_sendMessage((actualNumberOfBFArray[0] > 0) ? createBFmsg(msg, actualNumberOfBFArray[0]) : copyEthernetMessage(msg), outPorts); //use BF message if actual number of BF is greater 0
			}
		}
	}
	deleteMessage(msg);
}

void* FIProtocolLogic::_threadEntryFunction(void *voidArgs) {
	thread_args *theseargs = (thread_args*) voidArgs;
	((FIProtocolLogic*) theseargs->This)->sendingThreadTask(theseargs->msgT, theseargs->outPortsT, theseargs->delayT);
	return NULL;
}

void FIProtocolLogic::sendingThreadTask(EthernetMessage *msg, int outPorts, int Delay) {
	bpiout << "In sendingThreadTask" << std::endl;
	std::this_thread::sleep_for(std::chrono::milliseconds(Delay));
	_sendMessage(msg, outPorts);

}

void FIProtocolLogic::_threadTask(int a) {
	bpiout << "in Thread task number = " << a << std::endl;
	bpiout << "Thread end" << std::endl;
}

FIProtocolLogic::ProbabilityDistribution FIProtocolLogic::StringToProbDist(std::string StringEnum) {
	if (strncmp(StringEnum.c_str(), "discrete_equal_distribution", 27) == 0) {
//		bpiout << "Probdist discrete_equal_distribution" << std::endl;
		return discrete_equal_distribution;
	} else if (strncmp(StringEnum.c_str(), "Gaussian_distribution", 21) == 0) {
//		bpiout << "Probdist Gaussian_distribution" << std::endl;
		return Gaussian_distribution;
	} else if (strncmp(StringEnum.c_str(), "Exponential_distribution", 24) == 0) {
//		bpiout << "Probdist Exponential_distribution" << std::endl;
		return Exponential_distribution;
	} else {
//		bpiout << "invalid ProbabilityDistribution. Value is set to default(discrete_equal_distribution)" << std::endl;
		return discrete_equal_distribution;
	}
}

EthernetMessage* FIProtocolLogic::createBFmsg(EthernetMessage *msg, int actualNumberOfBF) {
	//BF in entire Message inc. address space etc.
//	For Documentation purpose
//	char destination[6]; /* 6 octets */
//	char source[6]; /* 6 octets */
//	unsigned short ether_type; /* 2 octets */
//	Message* data; /* Data - 46..1500 octets */
//	uint32_t fcs; /* 4 octets */

	size_t sizeOfMsg = 6 /*destination*/+ 6 /*source*/+ 2/*ether_type*/+ strlen(msg->data->data) + 4 /*fcs*/;
	std::bitset<8> msgBitArray[sizeOfMsg];

	std::string BFMsgDestination;
	std::string BFMsgSource;
	unsigned short BFMsgType = 0;
	std::string BFMsgData;
	uint32_t BFMsgfcs = 0;

	size_t sizeOfMsgData = strlen(msg->data->data);
	std::string BFMsgDataOnly;
	unsigned seed = std::chrono::steady_clock::now().time_since_epoch().count();
	std::default_random_engine eng(seed);
	std::uniform_int_distribution<int> distroData(0, sizeOfMsgData - 1);
	std::uniform_int_distribution<int> distroMsg(0, sizeOfMsg - 1);
	std::uniform_int_distribution<int> distroByte(0, 7);

	//read the message as Bits
	//BFMsgType & BFMsgfcs is read into the array in different order. least significant Byte is stored in [0] most significant Byte stored in higher number
	for (size_t i = 0; i < sizeOfMsg; i++) {
		if (i < 6) {
			msgBitArray[i] = std::bitset<8>(msg->destination[i]);
		} else if (i >= 6 && i < 12) {
			msgBitArray[i] = std::bitset<8>(msg->source[i - 6]);
		} else if (i >= 12 && i < 14) {
			msgBitArray[i] = std::bitset<8>(msg->ether_type >> 8 * (i - 12));
		} else if (i >= 14 && i < 14 + sizeOfMsgData) {
			msgBitArray[i] = std::bitset<8>(msg->data->data[i - 14]);
		} else if (i >= 14 + sizeOfMsgData && i < 14 + sizeOfMsgData + 4) {
			msgBitArray[i] = std::bitset<8>((msg->fcs) >> 8 * (i - (14 + sizeOfMsgData)));
		}
	}

	//Flip the passed amount of Bits
	for (int i = 0; i < actualNumberOfBF; ++i) {
		msgBitArray[distroMsg(eng)].flip(distroByte(eng));
	}

	//create String for Message Data
	for (size_t i = 0; i < sizeOfMsg; i++) {
		if (i < 6) {
			BFMsgDestination += msgBitArray[i].to_ulong();
		} else if (i >= 6 && i < 12) {
			BFMsgSource += msgBitArray[i].to_ulong();
		} else if (i >= 12 && i < 14) {
			BFMsgType += (std::bitset<16>((msgBitArray[i].to_ulong()) << 8 * (i - 12))).to_ulong();
		} else if (i >= 14 && i < 14 + sizeOfMsgData) {
			BFMsgData += msgBitArray[i].to_ulong();
		} else if (i >= 14 + sizeOfMsgData && i < 14 + sizeOfMsgData + 4) {
			BFMsgfcs += (std::bitset<32>((msgBitArray[i].to_ulong()) << 8 * (i - (sizeOfMsgData + 14)))).to_ulong();
		}
	}

	Message *BFData = createMessageData(BFMsgData.c_str(), sizeOfMsgData, ETHDATAMIN);
	auto em = createEthernetMessage((char*) BFMsgDestination.c_str(), (char*) BFMsgSource.c_str(), BFMsgType, BFData);

	em->fcs = BFMsgfcs;
	return em;
}
