#ifndef LSINTERFACE_H_
#define LSINTERFACE_H_

#define BUFFERSIZE 4096

#include "LocalBpiLib.h"
#include <string>
#include <pthread.h>
#include <limits.h>

/**
 * \brief Implementation of a communication interface between the TcpInterface and an instance of the LocalService.
 *
 * This class represents the control and IO interface for an instance of the LocalService.
 */
class LsInterface {
public:
	LsInterface(std::string uid, std::string runPath); /**< Creates a ProtocolInterface with specified uid and runPath. */
	virtual ~LsInterface();

	void start(); /**< Starts the ProtocolInterface by invoking a thread. */
	void stopLocalService(); /**< Stops the running thread. */
	int sendToLocalService(std::string s); /**< Sends a message to the local service. */
	bool isActive(); /**< Checks whether the ProtocolInterface ist still active or was terminated already. */
	std::string getUid(); /**< Returns the unique identifier of the Local Service. */
	int* getFdLsOutput(); /**< Retrieves a pointer to the fd which receives normal output from the LocalService. */
	int* getFdLsOutputErr(); /**< Retrieves a pointer to the fd which receives error output from the LocalService. */

private:
	std::string _uid; /**< Unique identifier of the protocol. */
	std::string _runPath; /**< Absolute path to the executeable of the protocol. */
	pid_t _pid; /**< Process identifier of the forked child process. */
	bool _active; /**< Indicates whether the interface ist still active or not. */
	pthread_t _piThread; /**< Thread object. */
	int _fdChildIn; /**< Ingoing file descriptor: LsInterface -> Ls. */
	int _fdChildOut; /**< Outgoing file descriptor: Ls -> LsInterface. */
	int _fdChildErr; /**< Outgoing file descriptor for errors: Ls -> LsInterface.*/

	static void * _threadEntryFunction(void * This); /**< Entry function for thread execution. */
	void _startThread(); /**< Core of the thread. */
	int _executeProtocol(); /**< Creates pipes, forks the thread and executes the child process. */
};

#endif /* LSINTERFACE_H_ */
