#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <iostream>
#include <string.h>
#include <stdio.h>
#include <algorithm>
#include <signal.h>
#include <limits.h>
#include <ctime>
#include "TcpInterface.h"
#include <sys/select.h>

using namespace std;
using json=nlohmann::json;

TcpInterface::TcpInterface(std::string execPath) :
		_lsIntfce(nullptr), _piThread(-1), _csThread(-1), _fdTcpOut(nullptr), _fdTcpErr(nullptr), _execPath(execPath) {
	_buffersize = stoi(Configuration::getConfig(JSON_CONFIGFILE, "buffer_size"));
	dout << "Buffersize set to " << _buffersize << endl;
	std::string ackmode = Configuration::getConfig(JSON_CONFIGFILE, "ack_mode");
	if (ackmode.compare("hashes") == 0) {
		_ackmode = ACKMODE_ACKHASHWRAP;
		dout << "ACKMODE set to ACKMODE_ACKHASHWRAP" << endl;
	} else if (ackmode.compare("none") == 0) {
		_ackmode = ACKMODE_NONE;
		dout << "ACKMODE set to ACKMODE_NONE" << endl;
	} else {
		_ackmode = ACKMODE_ACKWRAP;
		dout << "ACKMODE set to ACKMODE_ACKWRAP" << endl;
	}

	int fd_local[2];
	if (pipe(fd_local) < 0) {
		derr << "Could not create control pipe." << endl;
	}
	_fdCtrlIn = fd_local[1];
	_fdCtrlOut = fd_local[0];

	for (int i = 0; i < CLOCKSYNCVALUECOUNT; i++) {
		_clockSyncTxTimepoint[i][0] = -1;
		_clockSyncTxTimepoint[i][1] = -1;
	}

	pthread_mutex_init(&_mutex, NULL);
	pthread_cond_init(&_cond, NULL);
}

TcpInterface::~TcpInterface() {
}

/**
 * Starts the TcpInterface as separate thread, listening for data from
 * and to a tcp client and an instance of the LocalService
 */
void TcpInterface::start() {
	pthread_create(&_piThread, NULL, _threadEntryFunction, this);
	pthread_detach(_piThread);
}

bool TcpInterface::setLsInterface(LsInterface *__lsIntfce) {
	_lsIntfce = __lsIntfce;
	dout << "TcpInterface linked to LsInterface" << endl;
	return _lsIntfce != nullptr;
}

void TcpInterface::repeatFdListening() {
	writeToFileDescriptor(_fdCtrlIn, " ");
}

void* TcpInterface::_threadEntryFunction(void *This) {
	((TcpInterface*) This)->_startThread();
	return NULL;
}

void* TcpInterface::_clocksyncThreadEntryFunction(void *This) {
	((TcpInterface*) This)->_clockSync();
	return NULL;
}

void TcpInterface::_startThread() {
	pthread_setname_np(pthread_self(), "TcpInterface");

	string strBuffer;
	fd_set readfds;

	if (_lsIntfce != nullptr) {
		_fdTcpOut = _lsIntfce->getFdLsOutput();
		_fdTcpErr = _lsIntfce->getFdLsOutputErr();
		dout << "Connected to LocalService Output stream successfully" << endl;
	}

	_clocksyncmode = Configuration::getConfig(JSON_CONFIGFILE, "clocksync");
	if (_clocksyncmode.size() == 0)
		_clocksyncmode = "NONE";

	pthread_create(&_csThread, NULL, _clocksyncThreadEntryFunction, this);
	pthread_detach(_csThread);

	signal(SIGPIPE, SIG_IGN);

	while (active) {

		FD_ZERO(&readfds);
		FD_SET(_fdTcpBd, &readfds);
		FD_SET(_fdCtrlOut, &readfds);
		FD_SET(*_fdTcpOut, &readfds);
		FD_SET(*_fdTcpErr, &readfds);

		if (select(max(max(max(_fdTcpBd, *_fdTcpOut), _fdCtrlOut), *_fdTcpErr) + 1, &readfds,
		NULL, NULL, NULL) < 0) {
			derr << "Failed reading file descriptors" << endl;
			continue;
		}

		// Handle incoming messages, received over the tcp connection from the ConfigTool
		if (FD_ISSET(_fdTcpBd, &readfds)) {
			if (readFromFileDescriptor(_fdTcpBd, strBuffer, _buffersize) > 0) {
				std::vector<std::string> messages = splitIntoMessages(strBuffer);
				for (std::string msg : messages) {
					if (msg.c_str()[0] != '\u0002') {
						msg = _partialMsgBuffer.append(msg);
						_partialMsgBuffer = "";
						_handleInput(msg.substr(1, msg.length() - 2));
					} else if (msg.c_str()[msg.length() - 1] != '\u0003') {
						_partialMsgBuffer = msg;
					} else {
						_handleInput(msg.substr(1, msg.length() - 2));
					}

				}
			} else {
				if (send(_fdTcpBd, "hello", 5, 0) < 0) {
					dout << "Connection to tcp server seems to be lost" << endl;
					connected = false;
					pthread_mutex_lock(&_mutex);
					pthread_cond_signal(&_cond);
					pthread_mutex_unlock(&_mutex);
				}
			}
		}

		// Handle incoming messages, which have to be send over the tcp connection to the ConfigTool.
		if (FD_ISSET(*_fdTcpOut, &readfds)) {
			if (readFromFileDescriptor(*_fdTcpOut, strBuffer, _buffersize) > 0) {
				if (!SUPPRESS_LOCALMESSAGEPRINT)
					_PrintOutgoingMessage((strBuffer));
				if (connected) {
					string out = _partialMsgChildBuffer;
					_partialMsgChildBuffer = "";
					out.append(strBuffer);

					if (out.at(out.length() - 1) != '\n') {
						int pos = out.find_last_of("\n");
						if (pos >= 0) {
							_partialMsgChildBuffer = out.substr(pos, out.length() - pos);
							out = out.substr(0, pos);
						} else {
							_partialMsgChildBuffer = out;
							out = "";
						}
					} else {
						out = out.substr(0, out.length() - 1);
					}
					if (!out.empty())
						_sendTcpMessage(out);
				}
			} else {
				if (!_lsIntfce->isActive())
					*_fdTcpOut = -1;
			}
		}

		if (FD_ISSET(*_fdTcpErr, &readfds)) {
			if (readFromFileDescriptor(*_fdTcpErr, strBuffer, _buffersize) > 0) {
				if (!SUPPRESS_LOCALMESSAGEPRINT)
					_PrintOutgoingMessage((strBuffer));
				if (connected) {
					string out = _partialMsgChildBufferErr;
					_partialMsgChildBufferErr = "";
					out.append(strBuffer);

					if (out.at(out.length() - 1) != '\n') {
						int pos = out.find_last_of("\n");
						if (pos >= 0) {
							_partialMsgChildBufferErr = out.substr(pos, out.length() - pos);
							out = out.substr(0, pos);
						} else {
							_partialMsgChildBufferErr = out;
							out = "";
						}
					} else {
						out = out.substr(0, out.length() - 1);
					}
					if (!out.empty())
						_sendTcpMessage(out);
				}
			} else {
				if (!_lsIntfce->isActive())
					*_fdTcpErr = -1;
			}
		}

		// Handle control messages. Used at the moment only for interrupting the socket select statement
		if (FD_ISSET(_fdCtrlOut, &readfds)) {
			if (readFromFileDescriptor(_fdCtrlOut, strBuffer, _buffersize) > 0) {
			}
		}
	}
}

void TcpInterface::_clockSync() {
	pthread_setname_np(pthread_self(), "ClockSync");

	std::stringstream out;
	dout << "Clock Syncronization Mode: " << _clocksyncmode << std::endl;

	if (strncmp(_clocksyncmode.c_str(), "NONE", 4) == 0) {
		return;
	} else if (strncmp(_clocksyncmode.c_str(), "CUSTOM", 4) == 0) {
		while (active) {
			if (_fdTcpBd > 0) {
				_old = (long long int) getLocalClock()->gettime();
				printToTcpConnection("<csreq>");
			}
			usleep(1 * 1000000);
		}
	} //FIXME JN
	else if (strncmp(_clocksyncmode.c_str(), "SYSTEM", 6) == 0) {
		while (active) {
			if (_fdTcpBd > 0) {
				//_old = (long long int) getLocalClock()->gettime();
				clock_gettime(CLOCK_REALTIME, &_oldTP);
				printToTcpConnection("<csreq>");
				_oldNano = (long long int) TEN_POWER_NINE * _oldTP.tv_sec + _oldTP.tv_nsec;
			}
			usleep(3 * 1000000);
		}
	} //END JN
	else {
		dout << "Unknown Clock Synchronization Mode. Clock synchronization thread terminated" << std::endl;
	}
}

/**
 * Detects which command has to be started and initiate corresponding function calls. All known commands are
 * defined in InputHandler.h. For a command "exampleCommand" there should exist a
 * <b>\#define COMMAND_EXAMPLECOMMAND exampleCommand</b> which helps to identify known commands. After recognizing
 * a command, a function call with the naming convention <b>_CommandExampleCommand(input)</b> should be the only
 * operation to be done in this function. If a command could not be recognized, the function checks whether this
 * could be a command for an active Protocol. If this is true, the function sends the command to the corresponding
 * Protocol. Otherwise the function results in an error output.
 * @param cmdIn Command to be handled as char sequence
 */
void TcpInterface::_handleInput(std::string input) {

	// ALIVE MSG
	if (strncmp(COMMAND_ALIVE, input.c_str(), strlen(COMMAND_ALIVE)) == 0) {
		if (!SUPPRESS_LOCALMESSAGEPRINT_ALIVE && !SUPPRESS_LOCALMESSAGEPRINT)
			_PrintIncomingMessage(input);
		std::stringstream ss;
		ss << input << " " << (int) (getCurrentCpuUsage() * 100);
		_sendTcpMessage(ss.str());
		if (!SUPPRESS_LOCALMESSAGEPRINT_ALIVE && !SUPPRESS_LOCALMESSAGEPRINT)
			_PrintOutgoingMessage(input);
	}
	// CLOCKSYNC MSG
	else if (strncmp(COMMAND_SYNC, input.c_str(), strlen(COMMAND_SYNC)) == 0) {
//		std::string time = input.substr(strlen(COMMAND_SYNC));
//		long long int newsynctime = std::stoull(time);
//		long long int localtime = (long long int) getLocalClock()->gettime();
//		long long int offset = localtime - _old;
//		long long int diff = newsynctime - localtime;
//
//		// Save new Clock Syny Timepoint
//		_clockSyncTxTimepoint[_csPointer][0] = offset;
//		_clockSyncTxTimepoint[_csPointer][1] = diff;
//		_csPointer = (_csPointer + 1) % CLOCKSYNCVALUECOUNT;
//
//		// Check if Clock Sync initialization phase is finished
//		if (!csInitFinished) {
//			csInitFinished = true;
//			for (int i = 0; i < CLOCKSYNCVALUECOUNT; i++) {
//				if (_clockSyncTxTimepoint[i][0] < 0)
//					csInitFinished = false;
//			}
//		}
//
//		// Calculate correct value by removing the one with the
//		long long int corr = 0;
//		if (csInitFinished) {
//			int max = 0;
//			for (int i = 1; i < CLOCKSYNCVALUECOUNT; i++) {
//				if (abs(_clockSyncTxTimepoint[i][0]) > abs(_clockSyncTxTimepoint[max][0]))
//					max = i;
//			}
//			int max2 = (max == 0 ? 1 : 0);
//			for (int i = 0; i < CLOCKSYNCVALUECOUNT; i++) {
//				if (i != max && abs(_clockSyncTxTimepoint[i][0]) > abs(_clockSyncTxTimepoint[max2][0]))
//					max2 = i;
//			}
//			cout << "max2=" << max2 << "\tmax=" << max << std::endl;
//
//			for (int i = 0; i < CLOCKSYNCVALUECOUNT; i++) {
//				if (i != max2 && i != max) {
//					corr += _clockSyncTxTimepoint[i][1] + _clockSyncTxTimepoint[i][0] / 2;
//				}
//				_clockSyncTxTimepoint[i][0] = -1;
//			}
//
//			corr /= CLOCKSYNCVALUECOUNT - 2;
//			csInitFinished = false;
//		}
//
//		if (corr != 0) {
//			getLocalClock()->set(getLocalClock()->gettime() + corr);
//			std::stringstream sa;
//			sa << COMMAND_SYNC << " " << (long long int) getLocalClock()->gettime() << " IDxxxx";
//			if (_lsIntfce->isActive())
//				_lsIntfce->sendToLocalService(sa.str());
//		}

		std::string time = input.substr(strlen(COMMAND_SYNC));
		long long int newsynctime = std::stoull(time);
		long long int localtime = 0;
		long long int offset = 0;
		long long int diff = 0;
		long long int temp = 0;

		if (strncmp(_clocksyncmode.c_str(), "CUSTOM", 6) == 0) {
			localtime = (long long int) getLocalClock()->gettime();
			offset = (localtime - _old) * 1000;
			diff = newsynctime - (localtime * 1000);

		} else if (strncmp(_clocksyncmode.c_str(), "SYSTEM", 6) == 0) {
			clock_gettime(CLOCK_REALTIME, &localtimeTP);
			localtimeNano = TEN_POWER_NINE * (long long int) localtimeTP.tv_sec + localtimeTP.tv_nsec;
			offset = localtimeNano - _oldNano;
			diff = newsynctime - localtimeNano;

		}
		// Save new Clock Syny Timepoint
		_clockSyncTxTimepoint[_csPointer][0] = offset;
		_clockSyncTxTimepoint[_csPointer][1] = diff;
		_csPointer = (_csPointer + 1) % CLOCKSYNCVALUECOUNT;

		// Check if Clock Sync initialization phase is finished
		if (!csInitFinished) {
			csInitFinished = true;
			for (int i = 0; i < CLOCKSYNCVALUECOUNT; i++) {
				if (_clockSyncTxTimepoint[i][0] < 0)
					csInitFinished = false;
			}
		}

		// Calculate correct value by removing the one with the
		long long int corr = 0;
		if (csInitFinished) {
			int max = 0;
			for (int i = 1; i < CLOCKSYNCVALUECOUNT; i++) {
				if (abs(_clockSyncTxTimepoint[i][0]) > abs(_clockSyncTxTimepoint[max][0]))
					max = i;
			}
			int max2 = (max == 0 ? 1 : 0);
			for (int i = 0; i < CLOCKSYNCVALUECOUNT; i++) {
				if (i != max && abs(_clockSyncTxTimepoint[i][0]) > abs(_clockSyncTxTimepoint[max2][0]))
					max2 = i;
			}
//			cout << "max2=" << max2 << "\tmax=" << max << std::endl;

			for (int i = 0; i < CLOCKSYNCVALUECOUNT; i++) {
				if (i != max2 && i != max) {
					corr += _clockSyncTxTimepoint[i][1] + _clockSyncTxTimepoint[i][0] / 2;
				}
				_clockSyncTxTimepoint[i][0] = -1;
			}

			corr /= CLOCKSYNCVALUECOUNT - 2;
			csInitFinished = false;
		}
		if (corr != 0) {
			std::stringstream sa;
			if (strncmp(_clocksyncmode.c_str(), "CUSTOM", 6) == 0) {
				corr /= 1000;
				getLocalClock()->set(getLocalClock()->gettime() + corr);
				cout << "Clockadjust=" << corr << endl;
				sa << COMMAND_SYNC << " " << (long long int) getLocalClock()->gettime() << " IDxxxx";
				if (_lsIntfce->isActive())
					_lsIntfce->sendToLocalService(sa.str());
			} else if (strncmp(_clocksyncmode.c_str(), "SYSTEM", 6) == 0) {
				clock_gettime(CLOCK_REALTIME, &corrTP);
				temp = ((long long int) corrTP.tv_sec * TEN_POWER_NINE) + corrTP.tv_nsec;
				if (corr >= TEN_POWER_NINE)
					corrTP.tv_sec = corrTP.tv_sec + corr / TEN_POWER_NINE; //falls corr größer als 1 sek werden sek sofort angepasst
				corrTP.tv_nsec += corr;
				if (corrTP.tv_nsec > TEN_POWER_NINE) {
					++corrTP.tv_sec;  //check if sum of nano causes the seconds to go up
					corrTP.tv_nsec = corrTP.tv_nsec % TEN_POWER_NINE;
				}
				clock_settime(CLOCK_REALTIME, &corrTP);
			}
		}

	}
	// ACK MSG
	else if (input.c_str()[0] == '\u0006') {
		if (!SUPPRESS_LOCALMESSAGEPRINT && !SUPPRESS_LOCALACKMSGPRINT)
			_PrintIncomingMessage(input);
	}
	// Arbitrary Incoming Message
	else {
		if (!SUPPRESS_LOCALMESSAGEPRINT)
			_PrintIncomingMessage(input);

		switch (_ackmode) {
		case ACKMODE_NONE:

			break;
		case ACKMODE_ACKWRAP:
			if (!SUPPRESS_LOCALMESSAGEPRINT && !SUPPRESS_LOCALACKMSGPRINT)
				_PrintOutgoingMessage(ackMessage(input));
			_sendTcpMessage(ackMessage(input));
			break;
		case ACKMODE_ACKHASHWRAP:
			if (!SUPPRESS_LOCALMESSAGEPRINT && !SUPPRESS_LOCALACKMSGPRINT)
				_PrintOutgoingMessage((createHash(input)));
			_sendTcpMessage((createHash(input)));
			break;
		default:
			break;
		}
		std::vector<std::string> items = bpiSplitIntoItems(input, ' ');
		string cmdId = "";
		if (items.size() < 2) {
			std::stringstream cmdss;
			cmdss << createStatCtrlMessage(2, cmdId.c_str(), "error", "command got no id");
			_sendTcpMessageWithNamestamp(cmdss.str());
			return;
		}
		cmdId = items.back();
		items.pop_back();

		std::stringstream ss;
		for (auto item = items.begin(); item != items.end(); item++) {
			ss << *item << " ";
		}
		std::string cmd = ss.str().substr(0, ss.str().size() - 1);

		stringstream cmdss;
		if (strncmp(COMMAND_EXIT, input.c_str(), strlen(COMMAND_EXIT)) == 0) {
			stringstream cmdss;
			cmdss << createStatCtrlMessage(2, cmdId.c_str(), "success");
			_sendTcpMessageWithNamestamp(cmdss.str());
			_CommandExit();
		} else if (strncmp(COMMAND_ADIEU, input.c_str(), strlen(COMMAND_ADIEU)) == 0) {
			stringstream cmdss;
			cmdss << createStatCtrlMessage(2, cmdId.c_str(), "success");
			_sendTcpMessageWithNamestamp(cmdss.str());
			_CommandAdieu(cmd);
		} else if (strncmp(COMMAND_SHUTDOWN, input.c_str(), strlen(COMMAND_SHUTDOWN)) == 0) {
			stringstream cmdss;
			cmdss << createStatCtrlMessage(2, cmdId.c_str(), "success");
			_sendTcpMessageWithNamestamp(cmdss.str());
			_CommandShutdown(cmdId, cmd);
		} else if (strncmp(COMMAND_STATUS, input.c_str(), strlen(COMMAND_STATUS)) == 0) {
			_CommandStatus(cmdId, cmd);
		} else if (strncmp(COMMAND_GETCONFIG, input.c_str(), strlen(COMMAND_GETCONFIG)) == 0) {
			_CommandGetConfig(cmdId, cmd);
		} else if (strncmp(COMMAND_SETCONFIG, input.c_str(), strlen(COMMAND_SETCONFIG)) == 0) {
			_CommandSetConfig(cmdId, cmd);
		} else if (strncmp(COMMAND_COMPILELIB, input.c_str(), strlen(COMMAND_COMPILELIB)) == 0) {
			_CommandCompilelib(cmdId, cmd);
		} else if (strncmp(COMMAND_COMPILE, input.c_str(), strlen(COMMAND_COMPILE)) == 0) {
			_CommandCompile(cmdId, cmd);
		} else if (strncmp(COMMAND_CLEANLIB, input.c_str(), strlen(COMMAND_CLEANLIB)) == 0) {
			_CommandCleanlib(cmdId, cmd);
		} else if (strncmp(COMMAND_CLEAN, input.c_str(), strlen(COMMAND_CLEAN)) == 0) {
			_CommandClean(cmdId, cmd);
		} else if (strncmp(COMMAND_START, input.c_str(), strlen(COMMAND_START)) == 0) {
			_CommandStart(cmdId, cmd);
		} else if (strncmp(COMMAND_STOP, input.c_str(), strlen(COMMAND_STOP)) == 0) {
			_CommandStop(cmdId, cmd);
		} else if (strncmp(COMMAND_SSH, input.c_str(), strlen(COMMAND_SSH)) == 0) {
			_CommandSsh(cmdId, cmd);
		} else {
			if (_lsIntfce->isActive())
				if (input.length() <= MAXINPUTLENGTH) {
					_lsIntfce->sendToLocalService(input);
				} else {
					stringstream ss;
					ss << "Received too large message (command) with length " << input.size() << ". Forwarding to LocalService suppressed." << flush;
					_sendTcpMessageWithNamestamp(ss.str());
					cmdss << createStatCtrlMessage(3, cmdId.c_str(), "fail", "message too large");
				}
			else {
				std::stringstream ss;
				ss << "Received unknown command \"" << input << "\" which can't be forwarded to the LocalService due to inactivity!" << flush;
				printErrToTcpConnection(ss.str());
				cmdss << createStatCtrlMessage(3, cmdId.c_str(), "fail", "LocalService inactive");
			}
		}
		std::string msg = cmdss.str();
		if (msg.length() > 0)
			_sendTcpMessageWithNamestamp(msg);
	}
}

void TcpInterface::_PrintIncomingMessage(std::string s) {
	dout << " => MSG: \"" << s << "\"" << endl;
}

void TcpInterface::_PrintOutgoingMessage(std::string s) {
	int pos = s.find_last_of("\n");
	if (pos < 0)
		pos = s.length();
	dout << " <= MSG: \"" << s.substr(0, pos) << "\"" << endl;
}

int TcpInterface::_sendTcpMessage(std::string s) {
	string out = wrapInStxEtx(s).append("\n");
	return write(_fdTcpBd, out.c_str(), out.length());
}

int TcpInterface::_sendTcpMessageWithNamestamp(std::string s) {
	if (!SUPPRESS_LOCALMESSAGEPRINT) {
		_PrintOutgoingMessage(bpiGetNamestamp(NAME).append(s));
	}
	struct stat t_stat;
	stat(_execPath.c_str(), &t_stat);
	struct tm *timeinfo = localtime(&t_stat.st_ctime);
	return writeToFileDescriptor(_fdTcpBd, wrapInStxEtx(std::string(bpiGetNamestamp(NAME)).append(s).append(asctime(timeinfo))));
}

void TcpInterface::_CommandGetConfig(std::string id, std::string input) {
	stringstream cmdss;
	std::vector<std::string> items = bpiSplitIntoItems(input, ' ');
	if (items.size() < 2) {
		json j = Configuration::readJsonFile(JSON_CONFIGFILE);
		stringstream tmpSs;
		for (json::iterator it = j.begin(); it != j.end(); ++it) {
			tmpSs.str("");
			tmpSs << "\"" << it.key() << "\" : " << it.value() << endl;
			printToTcpConnection(tmpSs.str());
		}
		cmdss << createStatCtrlMessage(2, id.c_str(), "success");
	} else if (items.size() > 2) {
		printToTcpConnection(std::string("Too many parameters: \"").append(input).append("\""));
		printToTcpConnection(std::string("Specifiy no parameter to get an output of the complete config file or specify exactly one parameter of interest."));
		cmdss << createStatCtrlMessage(3, id.c_str(), "error", "invalid parameter request");
	} else {
		std::stringstream ss;
		ss << "\"" << items.at(1) << "\": \"" << Configuration::getConfig(JSON_CONFIGFILE, items.at(1)) << "\"" << endl;
		printToTcpConnection(ss.str());
		cmdss << createStatCtrlMessage(2, id.c_str(), "success");
	}
	_sendTcpMessageWithNamestamp(cmdss.str());
}

void TcpInterface::_CommandSetConfig(std::string id, std::string input) {
	stringstream cmdss;
	std::vector<std::string> items = bpiSplitIntoItems(input, ' ');
	if (items.size() < 3) {
		printErrToTcpConnection(std::string("Invalid key/value assignment: \"").append(input).append("\""));
		cmdss << createStatCtrlMessage(3, id.c_str(), "error", "Invalid key/value assignment");
	} else {
		bool result = Configuration::setConfig(JSON_CONFIGFILE, items.at(1), items.at(2));
		if (result) {
			std::stringstream ss;
			ss << "Successfully set configuration: \"" << items.at(1) << "\": \"" << items.at(2) << "\"" << endl;
			printToTcpConnection(ss.str());
			cmdss << createStatCtrlMessage(2, id.c_str(), "success");
		} else {
			printErrToTcpConnection(std::string("Configuration could not be set: \"").append(input).append("\""));
			cmdss << createStatCtrlMessage(3, id.c_str(), "error", "Configuration could not be set");
		}
	}
	_sendTcpMessageWithNamestamp(cmdss.str());
}

void TcpInterface::_CommandStatus(std::string id, std::string input) {
	stringstream cmdss;
	struct stat t_stat;
	stat(_execPath.c_str(), &t_stat);
	struct tm *timeinfo = localtime(&t_stat.st_ctime);
	printToTcpConnection(std::string("IoService [ACTIVE] -- ").append(asctime(timeinfo)));

	if (_lsIntfce->isActive()) {
		cmdss << "pstatus" << " " << id;
		_lsIntfce->sendToLocalService(cmdss.str());
	} else {
		printToTcpConnection(std::string("  \u255A\u2550 Local Service [INACTIVE]"));
		cmdss << createStatCtrlMessage(3, id.c_str(), "success", "LocalService inactive");
		_sendTcpMessageWithNamestamp(cmdss.str());
	}
}

void TcpInterface::_CommandCompile(std::string id, std::string input) {
	stringstream cmdss;
	if (_lsIntfce->isActive()) {
		printErrToTcpConnection("LocalService running. Stop before compiling.");
		cmdss << createStatCtrlMessage(3, id.c_str(), "error", "LocalService running");
		_sendTcpMessageWithNamestamp(cmdss.str());
	} else {
		stringstream ss;
		string path = Configuration::getConfig(JSON_CONFIGFILE, "source_dir");
		if (path.empty()) {
			ss << "Compilation failed. \"source_dir\" is not set in the configuration." << endl;
			printErrToTcpConnection(ss.str());
			cmdss << createCtrlMessage(3, id.c_str(), "error", "source_dir not set");
			_sendTcpMessageWithNamestamp(cmdss.str());
			return;
		}
		if (chdir(path.c_str()) != 0) {
			ss << "Compilation failed. \"source_dir\" is not a valid path to the LocalService" << endl;
			printErrToTcpConnection(ss.str());
			cmdss << createStatCtrlMessage(3, id.c_str(), "error", "source_dir not valid");
			_sendTcpMessageWithNamestamp(cmdss.str());
			return;
		}
		bpiSystemCall2("make all", &printToTcpConnection, &printErrToTcpConnection, id);
	}
}

void TcpInterface::_CommandCompilelib(std::string id, std::string input) {
	stringstream cmdss;
	stringstream ss;
	string path = Configuration::getConfig(JSON_CONFIGFILE, "libpath");
	if (path.empty()) {
		ss << "Compilation failed. \"libpath\" is not set in the configuration." << endl;
		printErrToTcpConnection(ss.str());
		cmdss << createStatCtrlMessage(3, id.c_str(), "error", "libpath not set");
		_sendTcpMessageWithNamestamp(cmdss.str());
		return;
	}
	if (chdir(path.c_str()) != 0) {
		ss << "Compilation failed. \"libpath\" is not a valid path to the GlobalBpiLib" << endl;
		printErrToTcpConnection(ss.str());
		cmdss << createCtrlMessage(3, id.c_str(), "error", "libpath not valid");
		_sendTcpMessageWithNamestamp(cmdss.str());
		return;
	}
	bpiSystemCall2("cd Debug && make all && cd ../Release && make all", &printToTcpConnection, &printErrToTcpConnection, id);
}

void TcpInterface::_CommandClean(std::string id, std::string input) {
	stringstream cmdss;
	if (_lsIntfce->isActive()) {
		printErrToTcpConnection("LocalService running. Stop before cleaning.");
		cmdss << createStatCtrlMessage(3, id.c_str(), "error", "LocalService running");
		_sendTcpMessageWithNamestamp(cmdss.str());
	} else {
		stringstream ss;
		string path = Configuration::getConfig(JSON_CONFIGFILE, "source_dir");
		if (path.empty()) {
			ss << "Clean failed. \"source_dir\" is not set in the configuration." << endl;
			printErrToTcpConnection(ss.str());
			cmdss << createStatCtrlMessage(3, id.c_str(), "error", "source_dir not set");
			_sendTcpMessageWithNamestamp(cmdss.str());
			return;
		}
		if (chdir(path.c_str()) != 0) {
			ss << "Clean failed. \"source_dir\" is not a valid path to the LocalService" << endl;
			printErrToTcpConnection(ss.str());
			cmdss << createCtrlMessage(3, id.c_str(), "error", "source_dir not valid");
			_sendTcpMessageWithNamestamp(cmdss.str());
			return;
		}
		bpiSystemCall2("make clean", &printToTcpConnection, &printErrToTcpConnection, id);
	}
}

void TcpInterface::_CommandCleanlib(std::string id, std::string input) {
	stringstream cmdss;
	stringstream ss;
	string path = Configuration::getConfig(JSON_CONFIGFILE, "libpath");
	if (path.empty()) {
		ss << "Clean failed. Could not find a path to the the GlobalBpiLib" << endl;
		printErrToTcpConnection(ss.str());
		cmdss << createStatCtrlMessage(3, id.c_str(), "error", "libpath not set");
		_sendTcpMessageWithNamestamp(cmdss.str());
		return;
	}
	if (chdir(path.c_str()) != 0) {
		ss << "Compilation failed. \"source_dir\" is not a valid path to the GlobalBpiLib" << endl;
		printErrToTcpConnection(ss.str());
		cmdss << createCtrlMessage(3, id.c_str(), "error", "libpath not valid");
		_sendTcpMessageWithNamestamp(cmdss.str());
		return;
	}
	bpiSystemCall2("cd Debug && make clean && cd ../Release && make clean", &printToTcpConnection, &printErrToTcpConnection, id);
}

void TcpInterface::_CommandStart(std::string id, std::string input) {
	stringstream cmdss;
	if (!SUPPRESS_LOCALCOMMANDDEBUG)
		dout << "Command \"" << input << "\" received." << endl;
	_lsIntfce->start();
	cmdss << createStatCtrlMessage(2, id.c_str(), "success");
	_sendTcpMessageWithNamestamp(cmdss.str());
}

void TcpInterface::_CommandStop(std::string id, std::string input) {
	stringstream cmdss;
	_lsIntfce->stopLocalService();
	cmdss << createStatCtrlMessage(2, id.c_str(), "success");
	_sendTcpMessageWithNamestamp(cmdss.str());
}

void TcpInterface::_CommandAdieu(std::string input) {
	dout << "Tcp disconnect" << endl;
	connected = false;
	pthread_mutex_lock(&_mutex);
	pthread_cond_signal(&_cond);
	pthread_mutex_unlock(&_mutex);
}

void TcpInterface::_CommandShutdown(std::string id, std::string input) {
	bpiSystemCall2("shutdown now", &printToTcpConnection, &printErrToTcpConnection, id);
}

void TcpInterface::_CommandExit() {
	dout << "Tcp disconnect" << endl;
	connected = false;
	active = false;
	pthread_mutex_lock(&_mutex);
	pthread_cond_signal(&_cond);
	pthread_mutex_unlock(&_mutex);
}

void TcpInterface::_CommandSsh(std::string id, std::string input) {
	stringstream cmdss;
	cmdss << input.substr(4);
//	printToTcpConnection(cmdss.str().c_str());
	bpiSystemCall2(cmdss.str(), &printToTcpConnection, &printErrToTcpConnection, id);

}
