/*
 * UdpListener.cpp
 *
 *  Created on: Jul 6, 2018
 *      Author: valentin
 */

#include "UdpConnector.h"

#include <pthread.h>

using namespace std;

UdpConnector::UdpConnector(TcpInterface* tcpIntfce) {
	_Buffersize = stoi(Configuration::getConfig(JSON_CONFIGFILE, "buffer_size"));
	_tcpIntfce = tcpIntfce;
	_UdpPort = stoi(Configuration::getConfig(JSON_CONFIGFILE, "port"));
}

UdpConnector::~UdpConnector() {
}

void UdpConnector::enterConnectionLoop() {

	while (_tcpIntfce->active) {
		_InitializeUdpListener();
		_WaitForConnectionRequest();
		_tcpIntfce->repeatFdListening();
		pthread_mutex_lock(&_tcpIntfce->_mutex);
		dout << "UDP Socket closed" << endl;
		close(_UdpSocket);
		pthread_cond_wait(&_tcpIntfce->_cond, &_tcpIntfce->_mutex);
		pthread_mutex_unlock(&_tcpIntfce->_mutex);

		if (_tcpIntfce->_fdTcpBd > 0) {
			close(_tcpIntfce->_fdTcpBd);
			_tcpIntfce->_fdTcpBd = -1;
		}
		if (_tcpIntfce->connected) {
			dout << "UDP notified" << endl;
		}
	}
}

void UdpConnector::_InitializeUdpListener() {
	_UdpBuffer = new char[_Buffersize]();

	if ((_UdpSocket = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
		cerr << "Error creating udp socket" << endl;
		exit(-1);
	}
	memset((char *) &_saUdp, 0, sizeof(_saUdp));

	_saUdp.sin_family = AF_INET;
	_saUdp.sin_port = htons(_UdpPort);
	_saUdp.sin_addr.s_addr = htonl(INADDR_ANY);

	if (bind(_UdpSocket, (struct sockaddr*) &_saUdp, sizeof(_saUdp)) == -1) {
		dout << "Error binding socket" << endl;
		exit(-1);
	}
}

void UdpConnector::_WaitForConnectionRequest() {

	setFd(nullptr);
	dout << "Not connected to ConfigTool. Waiting for UDP Broadcast..." << endl;

	int slen = sizeof(_saServer);

	if (recvfrom(_UdpSocket, _UdpBuffer, 1500, 0, (struct sockaddr *) &_saServer, (socklen_t*) &slen) == -1) {
		cerr << "Error receiving udp packet" << endl;
	} else {
		_ConnectToConfigTool();
	}

}

void UdpConnector::_ConnectToConfigTool() {
	if ((_tcpIntfce->_fdTcpBd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
		cerr << "Error creating tcp socket" << endl;
		return;
	}

	struct sockaddr_in server_address;
	server_address.sin_family = PF_INET;
	server_address.sin_port = htons(_UdpPort);
	server_address.sin_addr.s_addr = _saServer.sin_addr.s_addr;

	if (connect(_tcpIntfce->_fdTcpBd, (sockaddr *) &server_address, sizeof(server_address)) != 0) {
		dout << "Error creating TCP connection" << endl;
		return;
	}

	char str[INET_ADDRSTRLEN];
	inet_ntop(AF_INET, &(_saServer.sin_addr), str, INET_ADDRSTRLEN);

	dout << "Connected sucessfully to " << str << endl;
	setFd(&(_tcpIntfce->_fdTcpBd));
	_tcpIntfce->connected = true;
}
