/*
 * CommandHandler.h
 *
 *  Created on: Jul 28, 2017
 *      Author: valentin
 */

#ifndef INPUTHANDLER_H_
#define INPUTHANDLER_H_

#define COMMAND_GETCONFIG "getc"
#define COMMAND_SETCONFIG "setc"
#define COMMAND_DELETECONFIG "deletec"
#define COMMAND_COMPILE "pcompile"
#define COMMAND_START "pstart"
#define COMMAND_STOP "pstop"
#define COMMAND_RESTART "prestart"
#define COMAMND_STATUS "pstatus"
#define COMMAND_SYNC "_sync"
#define COMMAND_EXIT "locexit"

#define COMMAND_SETSYSTEMCLOCK "setsystemclock"

#include "LocalBpiLib.h"
#include "ProtocolInterface.h"
#include <string>
#include <map>
#include <vector>

/**
 * \brief Handles user input.
 *
 * This class represents the core functionality of the local service. It runs in an infinite loop until the user
 * terminates the aplication and retains and handles all incoming commands from SDTIN. This makes it possible to
 * use the Local Service as standalone application or by executing it in a process chain from another application,
 * i.e. the I/O Service. Output from the Local Service itself or from executed protocols as forked child processes
 * is written to STDOUT, so that this can be either seen by the user or read by the executing parent process.
 */
class InputHandler {
public:
	InputHandler(std::string execPath);

	BpiOutStream bpiout; /**< Standard output printer. */
	BpiOutStream bpierr; /**< Error output printer. */

	void runLoop(); /**< Obtains and handles in an infinite loop incoming commands. */
	void handleCommand(char *cmdIn); /**< Realizes command handling. */

private:
	std::string _execPath;
	bool _run = true; /**< Indicates whether the InputHandler is active or not. */
	std::map<std::string, ProtocolInterface*> _activeProtocols; /**< Map of all active/running Protocols. */

	void _CommandGetConfig(std::string id, std::string input); /**< Implementation of the command COMMAND_GETCONFIG. */
	void _CommandSetConfig(std::string id, std::string input); /**< Implementation of the command COMMAND_SETCONFIG. */
	void _CommandDeleteConfig(std::string id, std::string input); /**< Implementation of the command COMMAND_DELETECONFIG. */
	void _CommandCompile(std::string id, std::string input); /**< Implementation of the command COMMAND_COMPILE. */
	void _CommandStart(std::string id, std::string input); /**< Implementation of the command COMMAND_START. */
	void _CommandStop(std::string id, std::string input); /**< Implementation of the command COMMAND_STOP. */
	void _CommandRestart(std::string id, std::string input); /**< Implementation of the command COMMAND_RESTART. */
	void _CommandStatus(std::string id, std::string input); /**< Implementation of the command COMMAND_STATUS. */
	void _CommandSyncClock(std::string id, std::string input);
	void _CommandExit(std::string id); /**< Implementation of the command COMMAND_EXIT. */
	void _CommandSetSystemClock(std::string id, std::string input);
};

#endif /* INPUTHANDLER_H_ */
