#include "DemoLogic.h"

#include <Protocols/EtherTypes.h>
#include <Protocols/Port.h>
#include <Protocols/MessageTypes.h>

#include "DemoPort.h"

CustomLogic::CustomLogic() {
}

CustomLogic::~CustomLogic() {
}

Port* CustomLogic::_createPort(int i) {
	// Create objects (ports) from a port-derived class, which
	// override ingoing and outgoing message handling
	return new CustomPort(i, this);
}

void CustomLogic::_executeProtocolLogic() {
	// Iterate over all (Ethernet) messages received from other clients
	for (auto msgit = protocol_messages.begin(); msgit != protocol_messages.end(); msgit = protocol_messages.erase(msgit)) {
		EthernetMessage* message = *msgit;

		// Message handling depending on Ether Type
//		switch ((message)->ether_type) {
//		case ETH_TYPE_DEMO:
		if (out)
			bpiout << "RX:   " << printMessage(message) << std::endl;
		if (duration == -1) {
			timer.reset();
		}
		duration = timer.elapsed();
		rxTotal++;
		totalsize += message->size;
		deleteMessage(message);

//			break;
//		default:
//			break;
//		}
	}
}

void CustomLogic::_handleControlMessage(std::string cmdId, std::string cmd) {
	// Split received control message (command from ConfigTool)
	std::vector<std::string> items = bpiSplitIntoItems(cmd, ' ');

	// Perform command handling here
	if (strncmp(items[0].c_str(), "sayhi", 5) == 0) {
		// Send a message with data "hi" over all ports

		for (int i = 0; i < 4; i++) {
			Message* data = createMessageData("hi", ETHDATAMIN);
			auto em = createEthernetMessage(_bc_mac, source_mac, ETH_TYPE_DEMO, data);
			bpiout << "Message sent on port " << i + 1 << ": " << printMessage(em) << std::endl;
			ports[i]->addMessage(em);
		}
	} else if (strncmp(items[0].c_str(), "show", 4) == 0) {
		std::stringstream ss;
		ss << "Rx Total: " << rxTotal << std::endl;
		ss << "Total Size: " << totalsize << " byte" << std::endl;
		ss << "Duration: " << duration / 1000000.0 << " sec" << std::endl;
		ss << "Speed: " << totalsize * 8.0 / duration << " MBit/s" << std::endl;
		rxTotal = 0;
		duration = -1;
		totalsize = 0;
		ss << "Resetted";
		bpiout << ss.str() << std::endl;
	} else if (strncmp(items[0].c_str(), "reset", 5) == 0) {
		rxTotal = 0;
		duration = -1;
		totalsize = 0;
		bpiout << "Resetted" << std::endl;
	} else if (strncmp(items[0].c_str(), "setoutoff", 9) == 0) {
		out = false;
	} else if (strncmp(items[0].c_str(), "setouton", 8) == 0) {
		out = true;
	}
}

