#include "FabanLogic.h"

#include <Protocols/EtherTypes.h>
#include <Protocols/Port.h>
#include <Protocols/MessageTypes.h>
#include <string>
#include <unistd.h>
#include <stdint.h>
#include <bitset>

#include "FabanPort.h"

FabanLogic::FabanLogic() {
	_nodeId = -1;
	_seqNo = 0;
	_startTime = std::chrono::system_clock::now();
	_sgnKeyP = _dMask = _cMask = _rMask = 0;
	for (int i = 0; i < 6; i++)
		_bc_mac[i] = (char) 255;
	senderNumMsg = 0;
	senderDelayMin = 0;
	senderDelayMax = 0;
	statRx = 0;
//	statRxDup = 0;
	statTx = 0;
	statRxFaulty = 0;
	statDeltaTd = 0;
	statDeltaTdSum = 0;
	statR=0;
	statD = 0;
	statL = 0;
	for(int i = 0; i< 256; i++){
		seqCounter1[i] = 255;
		seqCounter2[i] = 255;
		x[i]=-1;
		y[i]=-1;
//		std::vector<uint8_t> temp;
//		lc.push_back(temp);
	}

//	std::cout << getLocalTime() << std::endl;
}

FabanLogic::~FabanLogic() {
}

Port* FabanLogic::_createPort(int i) {
	return new FabanPort(i, this);
}

void FabanLogic::_executeProtocolLogic() {
	// Iterate over all (Ethernet) messages received from other clients
	for (auto msgit = protocol_messages.begin();
			msgit != protocol_messages.end();) {
		EthernetMessage *message = *msgit;
		msgit = protocol_messages.erase(msgit);
		// Message handling depending on Ether Type
		switch ((message)->ether_type) {
		case ETH_TYPE_FABAN_H0: {
			// Act as distributing bridge
		}
			break;
		case ETH_TYPE_FABAN_H1: {
			// Act as checking bridge
			FabanMessage *fm = deserializeFabanMessage(message->data);
//			bpiout << "**CB** " << printMessage(message) << " -> " << printMessage(fm) << " -> " << printMessage(fm->data, true) << std::endl;
			if (_deliveryDeadlineCheck(fm)) {
				for (auto src : _routingtable) {
					if (src.first == fm->senderId)
						for (auto portIn : src.second) {
							if (portIn.first
									== ((EthMsgMeta*) message->meta)->port)
								for (auto portOut : portIn.second) {
//									bpiout << "\tDeliveryCheck ok. Forwarding msg to port " << portOut << std::endl;
//									ports[portOut - 1]->addMessage(copyEthernetMessage(message));
									sendMessage(copyEthernetMessage(message), (int) pow(2, portOut-1));
								}
						}
				}
			} else {
				bpiout << "Delivery Deadline check failed for Faban Message:"
						<< std::endl;
				bpiout << "\t" << printMessage(fm) << std::endl;
				bpiout << "Suppress message" << std::endl;
			}
			deleteMessage(fm);
//			message->ether_type = ETH_TYPE_FABAN_H2;
			_deliverMessage(message);
		}
			break;
		case ETH_TYPE_FABAN_H2: {
			FabanMessage *fm = deserializeFabanMessage(message->data);
//			bpiout << "**FB** " << printMessage(message) << " -> " << printMessage(fm) << " -> " << printMessage(fm->data, true) << std::endl;
			for (auto src : _routingtable) {
				if (src.first == fm->senderId)
					for (auto portIn : src.second) {
						if (portIn.first == ((EthMsgMeta*) message->meta)->port)
							for (auto portOut : portIn.second) {
//								bpiout << "\tForwarding msg to port " << portOut << std::endl;
//								ports[portOut - 1]->addMessage(
//										copyEthernetMessage(message));
								sendMessage(copyEthernetMessage(message), (int) pow(2, portOut-1));

							}
					}
			}
			deleteMessage(fm);
			_deliverMessage(message);
		}
			break;
		default:
			break;
		}
	}
}

void FabanLogic::_handleControlMessage(std::string cmdId, std::string cmd) {
	// Split received control message (command from ConfigTool)
	std::vector<std::string> items = bpiSplitIntoItems(cmd, ' ');

	// Perform command handling here
	if (items[0].compare("status") == 0) {
		if (_routingtable.empty())
			bpiout
					<< createCtrlMessage(3, cmdId.c_str(), "active",
							"unconfigured") << std::endl;
		else
			bpiout
					<< createCtrlMessage(3, cmdId.c_str(), "active",
							"configured") << std::endl;

	}

	// RESETROUTING
	else if (items[0].compare("resetrouting") == 0) {
		_routingtable.clear();
		_publicsignaturekeys.clear();
		bpiout << createCtrlMessage(2, cmdId.c_str(), "resetted") << std::endl;
	}

	// ADDROUTINGINFO
	else if (items[0].compare("addroutinginfo") == 0) {
		if (items.size() == 4) {
			uint8_t src = std::stoi(items[1]);
			int portIn = atoi(items[2].c_str());
			std::string portOutStr = items[3];
			std::vector<int> portOut;
			for (uint8_t i = 0; i < portOutStr.size(); i++) {
				portOut.push_back(portOutStr.at(i) - '0');
			}
			std::pair<int, std::vector<int>> x =
					std::pair<int, std::vector<int>>(portIn, portOut);
			auto it = _routingtable.find(src);
			if (it == _routingtable.end()) {
				std::map<int, std::vector<int>> newSrcMap;
				newSrcMap.insert(x);
				_routingtable.insert(
						std::pair<uint8_t, std::map<int, std::vector<int>>>(src,
								newSrcMap));
			} else {
				it->second.insert(x);
			}
		}
	}

	/**
	 * SHOWCONFIGURATION
	 */
	else if (items[0].compare("showconfiguration") == 0) {
		// Connection Information
		std::stringstream cd;
		cd << "Connected devices:\n";
		for (int i = 0; i < 4; i++) {
			cd << "\tPort_" << i << " -> "
					<< macToString(6, ports[i]->getRawDestMacAddress())
					<< std::endl;
		}
		bpiout << cd.str() << std::endl;

		// NodeID
		if (_nodeId == -1)
			bpiout << "NodeID: \n\t(not configured)" << std::endl;
		else
			bpiout << "NodeID: \n\t" << _nodeId << std::endl;

		//Routing
		bpiout << "Routing:" << std::endl;
		if (_routingtable.empty())
			bpiout << "\t(empty)" << std::endl;
		for (auto src : _routingtable) {
			for (auto portIn : src.second) {
				std::stringstream ss;
				ss << "[";
				for (auto portOut : portIn.second) {
					ss << portOut;
				}
				ss << "]";
				bpiout << "\t" << src.first << ": "
						<< (portIn.first == -1 ?
								"*" : std::to_string(portIn.first).c_str())
						<< " -> " << ss.str().c_str() << std::endl;
			}
		}

		// Signature Keys
		std::stringstream p;
		if (_sgnKeyP == 0)
			p << "(not set)";
		else
			p << "p = " << std::hex << std::uppercase << std::setfill('0')
					<< std::setw(8) << _sgnKeyP;
		bpiout << "Private Signature Key: \n\t" << p.str() << std::endl;
		bpiout << "Public Signature Key Table:" << std::endl;
		if (_publicsignaturekeys.empty())
			bpiout << "\t(empty)" << std::endl;
		for (auto src : _publicsignaturekeys) {
			std::stringstream ss;
			ss << "\t[" << (int) src.first << "]\tq = " << std::hex
					<< std::uppercase << std::setfill('0') << std::setw(8)
					<< (uint32_t) src.second.first << "\tr = "
					<< (uint32_t) src.second.second;
			bpiout << ss.str() << std::endl;
		}

		// Signature Masks
		std::stringstream ss;
		ss << "Faban Signature Masks:" << std::endl;
		ss << "\tD_Mask = " << std::hex << std::uppercase << std::setfill('0')
				<< std::setw(8) << _dMask << std::endl;
		ss << "\tC_Mask = " << std::hex << std::uppercase << std::setfill('0')
				<< std::setw(8) << _cMask << std::endl;
		ss << "\tR_Mask = " << std::hex << std::uppercase << std::setfill('0')
				<< std::setw(8) << _rMask;
		bpiout << ss.str() << std::endl;
	}

	/**
	 * INIT
	 */
	else if (items[0].compare("init") == 0) {
		if (items.size() > 1) {
			_nodeId = atoi(items[1].c_str());
		}
		auto em = createEthernetMessage(_bc_mac, source_mac,
		ETH_TYPE_FABAN_INIT, createMessageData("init", 4));
		sendMessage(em, 0b1111);
		bpiout << "Init message sent on all ports" << std::endl;
	}

	/**
	 * PING
	 */
	else if (items[0].compare("ping") == 0) {
		for (int i = 0; i < 4; i++) {
			Message *data;
			if (items.size() > 1)
				data = createMessageData(items[1].c_str(), items[1].size());
			else
				data = createMessageData("NULL", 4);
			auto em = createEthernetMessage(ports[i]->getRawDestMacAddress(),
					source_mac, ETH_TYPE_FABAN, data);
			ports[i]->addMessage(em);
		}
	}

	/**
	 * SENDFABANMESSAGE
	 */
	else if (items[0].compare("sendfabanmessage") == 0) {
		std::string d = "NULL";
		if (items.size() > 1)
			d = items[1];

		Message *data = createMessageData(d.c_str(), d.size(), FABDATAMIN);
		FabanMessage *fabMsg = createFabanMessage(_nodeId, _seqNo++, 0xABCDEF,
				data);
		EthernetMessage *em = createEthernetMessage(_bc_mac, source_mac,
		ETH_TYPE_FABAN_H0, serializeFabanMessage(fabMsg));
//		std::stringstream ss;
//		ss << "****** " << std::setw(2) << std::hex << std::uppercase << em->fcs << " changed to " << _sign(em)->fcs;
		_sign(em)->fcs;
//		bpiout << "**BC** " << printMessage(em) << " -> "
//				<< printMessage(fabMsg) << " -> " << printMessage(data, true)
//				<< std::endl;

		// TODO Simulate here Ingress Port handling until EndStations are implemented
		uint32_t sig = em->fcs;
		if ((sig & (1 << 31)) != 0) {
			sig = (sig << 1) + 1;
		} else {
			sig = (sig << 1);
		}
		em->fcs = sig;
//		bpiout << "**RL** " << printMessage(em) << std::endl;
		// END

		for (auto src : _routingtable) {
			if (src.first == _nodeId)
				for (auto portIn : src.second) {
					if (portIn.first == -1)
						for (auto portOut : portIn.second) {
//							bpiout << "Sending msg via port " << portOut << std::endl;
							// TODO Sending to checking bridges
							ports[portOut - 1]->addMessage(
									copyEthernetMessage(em));
						}
				}
		}
		deleteMessage(fabMsg);
		deleteMessage(em);
	}

	/**
	 * SENDER
	 */
	else if (items[0].compare("sender") == 0) {
		if (items.size() <= 1) {
			bpiout << "No Sender specific parameter specified" << std::endl;
		} else if (items[1].compare("set") == 0) {
			if (items.size() % 2 != 0) {
				bpiout << "Invalid parameters" << std::endl;
			} else {
				for (size_t i = 2; i < items.size(); i += 2) {
					if (items[i].compare("-n") == 0) {
						senderNumMsg = std::stoi(items[i+1]);
						bpiout << "SET: n="<< senderNumMsg << std::endl;
					} else if (items[i].compare("-d") == 0) {
						senderDelayMin = std::stod(items[i+1]);
						senderDelayMax = std::stod(items[i + 1]);
					}else if (items[i].compare("-dmin") == 0) {
						senderDelayMin = std::stod(items[i+1]);
					}else if (items[i].compare("-dmax") == 0) {
						senderDelayMax = std::stod(items[i+1]);
					}
				}
			}

		} else if (items[1].compare("start") == 0) {
			pthread_t thread;
			pthread_create(&thread, NULL, _senderThreadEntryFunction, this);
			pthread_detach(thread);
		}
	}

	/**
	 * SHOWSTATS
	 */
	else if (items[0].compare("showstats") == 0) {
		bpiout << "Statistics:\n\tRX = " << statRx
				<< "\n\tTX = " << statTx
				<< "\n\tRX(UNIQUE) = " << statR
				<< "\n\tRX(Dup) = " << statD
				<< "\n\tRX(Faulty) = " << statRxFaulty
				<< "\n\tLost = " << statL
				<< "\n\tPortTX = " << this->ports[0]->statTx << " / " << this->ports[1]->statTx << " / " << this->ports[2]->statTx << " / " << this->ports[3]->statTx
				<< "\n\tPortRX = "<< this->ports[0]->statRx << " / " << this->ports[1]->statRx << " / " << this->ports[2]->statRx << " / " << this->ports[3]->statRx
				<< "\n\tDeltaTD = " << statDeltaTd << "us  = " <<statDeltaTd/1000.0 << "ms = " << statDeltaTd/1000000.0 << "s"
				<< "\n\tDeltaTDMean = " << statDeltaTd/((double)statRx) << "us  = " <<statDeltaTd/((double)statRx)/1000.0 << "ms = " << statDeltaTd/((double)statRx)/1000000.0 << "s"
				<< std::endl;
	}

	/**
	 * RESETSTATS
	 */
	else if (items[0].compare("resetstats") == 0) {
		statRx = 0;
		statTx = 0;
		statR=0;statD=0;statL=0;
		for (int i = 0; i < 4; i++) {
			this->ports[i]->statRx = 0;
			this->ports[i]->statTx = 0;
		}
		statRxFaulty = 0;
		statDeltaTd = 0;
		statDeltaTdSum = 0;

//		for(int i = 0; i< 256; i++){
//			seqCounter[i] = 0;
//		}
	}

	/**
	 * ADDSGNPUBLIC
	 */
	else if (items[0].compare("addsgnpublic") == 0) {
		if (items.size() != 4)
			bpierr << "Invalid command received: " << cmd << std::endl;
		else {
			uint32_t q, r;
			try {
				q = std::stoul(items[2].c_str());
				r = std::stoul(items[3].c_str());
			} catch (const std::out_of_range &e) {
				bpiout << "Out of Range error: " << items[2].c_str() << " "
						<< items[3].c_str() << std::endl;
			}

			_publicsignaturekeys.insert(
					std::pair<uint8_t, std::pair<unsigned int, unsigned int>>(
							(uint8_t) std::stoi(items[1]),
							std::pair<uint32_t, uint32_t>(q, r)));
		}
	}

	/**
	 * ADDSGNPRIVATE
	 */
	else if (items[0].compare("addsgnprivate") == 0) {
		if (items.size() != 2)
			bpierr << "Invalid command received: " << cmd << std::endl;
		else
			try {
				_sgnKeyP = std::stoul(items[1]);
			} catch (const std::out_of_range &e) {
				bpiout << "Out of Range error: Private" << std::endl;
			}
	}

	/**
	 * SETMASKS
	 */
	else if (items[0].compare("setmasks") == 0) {
		if (items.size() != 4)
			bpierr << "Invalid command received: " << cmd << std::endl;
		else {
			try {
				_dMask = (uint32_t) std::stoul(items[1]);
				_cMask = (uint32_t) std::stoul(items[2]);
				_rMask = (uint32_t) std::stoul(items[3]);
				for (int i = 0; i < 4; i++) {
					((FabanPort*) ports[i])->setMaskD(_dMask);
					((FabanPort*) ports[i])->setMaskC(_cMask);
				}
			} catch (const std::out_of_range &e) {
				bpiout << "Out of Range error: Masks => " << items[1] << " "
						<< items[2] << " " << items[3] << std::endl;
			}

		}
	} else if(items[0].compare("setlinkspeed") == 0){
		if (items.size() == 2) {
			double ls = std::stod(items[1]);
			for (int i = 0; i < 4; i++) {
				ports[i]->setLinkSpeed(ls);
			}
			bpiout << "Link Speed set to " << ls << std::endl;
		} else {
			bpiout << "Wrong command format" << std::endl;
		}
	}
}

EthernetMessage* FabanLogic::_sign(EthernetMessage *msg) {
	if (msg == nullptr)
		return nullptr;

	uint32_t sgn = _oddMultiplication(msg->fcs, _sgnKeyP);
	msg->fcs = sgn;
	return msg;
}

bool FabanLogic::_signaturetest(EthernetMessage *msg) {
	FabanMessage *fm = deserializeFabanMessage(msg->data);
	std::map<uint8_t, std::pair<uint32_t, uint32_t>>::iterator it =
			_publicsignaturekeys.find(fm->senderId);
	if (it == _publicsignaturekeys.end())
		return false;

	uint32_t q = it->second.first;
	uint32_t r = it->second.second;
	uint32_t u = msg->fcs;
	msg->fcs = 0;
//	bpiout << "**S*** " << printMessage(msg) << std::endl;
	uint32_t crc = generateCrc32(msg);
	msg->fcs = u;
	std::stringstream ss;
//	ss << "\t  u ** q = " << std::setw(2) << std::hex << std::uppercase << u << " ** " << q << " = " << _oddMultiplication(u, q) << "\n\tcrc ** r = " << crc << " ** " << r << " = "
//			<< _oddMultiplication(crc, r);
//	bpiout << ss.str() << std::endl;
	return _oddMultiplication(u, q) == _oddMultiplication(crc, r);
}

uint32_t FabanLogic::_oddMultiplication(uint32_t x, uint32_t y) {
	uint64_t z = ((uint64_t) x) * ((uint64_t) y);
	z = 2 * ((z << 32) >> 32) + (uint64_t) x + (uint64_t) y;
	return (uint32_t) z;
}

bool FabanLogic::_deliveryDeadlineCheck(FabanMessage *msg) {

	return true;
}

void FabanLogic::_deliverMessage(EthernetMessage *msg) {
	// Simulate here the hop between the CB and connected Endstations, i.e. perform egrPort operations.
	if (msg->ether_type == ETH_TYPE_FABAN_H1) {
		msg->fcs = msg->fcs ^ _cMask;
		msg->ether_type = ETH_TYPE_FABAN_H2;
	}

	// Simulate reception in an endstation.
	msg->ether_type = ETH_TYPE_FABAN_H0;
	_rotateSignatureRight(msg);
	_rotateSignatureRight(msg);
	msg->fcs = msg->fcs ^ _rMask;
//	bpiout << "**RM** " << printMessage(msg) << std::endl;
	FabanMessage *fm = deserializeFabanMessage(msg->data);
	bool sigtest = _signaturetest(msg);
//	bpiout << "SEQ = " << fm->sequenceNumber << std::endl;
//	bpiout << getLocalTime48() << " - " << (double) fm->deliveryTime  <<  " = " << getLocalTime48() -  (double) fm->deliveryTime << "us" << std::endl;
	statDeltaTd = (getLocalTime48() -  (double) fm->deliveryTime > statDeltaTd ? getLocalTime48() -  (double) fm->deliveryTime : statDeltaTd);
	statDeltaTdSum += getLocalTime48() -  (double) fm->deliveryTime;
	statRx++;
//	bpiout << "**RX** " << printMessage(msg) << " -> " << printMessage(fm)
//			<< " -> " << printMessage(fm->data, true) << " --> SgnTest = "
//			<< (sigtest ? "OK" : "FAILED") << std::endl;
	if (sigtest) {
//		if(seqCounter1[fm->senderId]==fm->sequenceNumber || seqCounter2[fm->senderId]==fm->sequenceNumber ){
//			statRxDup++;
//		}else{
//			statRx++;
//		}
//
//		if(seqCounter1[fm->senderId]+1==fm->sequenceNumber){
//			seqCounter1[fm->senderId] = fm->sequenceNumber;
//		}else if(seqCounter2[fm->senderId]+1 == fm->sequenceNumber){
//			seqCounter2[fm->senderId] = fm->sequenceNumber;
//		}else if(fm->sequenceNumber - seqCounter1[fm->senderId] < fm->sequenceNumber - seqCounter2[fm->senderId]){
//			seqCounter1[fm->senderId] = fm->sequenceNumber;
//		}else{
//			seqCounter2[fm->senderId] = fm->sequenceNumber;
//		}

		uint8_t seq = fm->sequenceNumber;
		uint8_t id = fm->senderId;
		int lcPos = _contains(lc[id], seq);
		if(lcPos>=0){
			statR++;
			statL--;
//			bpiout << "[" << (int)id << "] " << (int)seq << " found at pos " << (int) lcPos << std::endl;
			lc[id].erase(lc[id].begin()+lcPos);
		}else
		if(_dist(x[id],seq) >0 && _dist(x[id],seq) <=100 && _dist(x[id],seq)<=_dist(y[id],seq)){
//			for(int n = x[id]+1; n<seq; n++){
//				lc.at(id).push_back(n);
//			}

			int old = x[id];
			int newLost = _dist(x[id],seq)-1;;
			statL += _dist(x[id],seq)-1;
			x[id]=seq;
			if (newLost> 0)
//				bpiout << "[" << (int)id << "] " <<"xy = " << old << "/" << y[id] << " -> " << x[id] << "/" << y[id] << "\tLOST += " << newLost << std::endl;
			for(int i = old+1; i<= old + newLost; i++){
//				bpiout <<"[" << (int)id << "] " <<  "LC.add(" <<(int)(i%256) << ")" << std::endl;
				lc[id].push_back((uint8_t)(i%256));
			}
			statR++;
		}else if(_dist(y[id],seq) <=100){
			statD++;
			y[id]=seq;
		}
//		else if(x[id]>= y[id] && x[id] - seq>=200){
//			statL += (seq>x[id] ? seq-x[id]-1 : 255-x[id] + seq);
//			x[id] = seq;
//			statR++;
//		}
//		else if(x[id]< y[id] && y[id] - seq>=200){
//			statD++;
//			y[id] = seq;
//		}
		else{
			statD++;
			_cleanLc(id, seq);
//			bpiout << "[" << (int)id << "] " << "DUP: " << (int)seq << std::endl;
		}
//		bpiout << "xy= " << x[id] << " / " << y[id] << std::endl;
	} else{
		statRxFaulty++;
	}
	deleteMessage(fm);
	deleteMessage(msg);
}

int FabanLogic::_contains(std::vector<uint8_t> vec, uint8_t el){
	for(int i = 0; i<vec.size(); i++){
		if (vec[i] == el)
			return i;
	}
	return -1;
}
void FabanLogic::_cleanLc(uint8_t id, uint8_t seq){
	if(lc[id].empty()) return;
	int num = 0;
	for(int i = 0; i<lc[id].size(); i++){
		if(_dist(lc[id].at(i), seq)<=20){
			lc[id].erase(lc[id].begin() + i);
			num++;
		}
//		bpiout << "Removed " << num << " from lc list" << std::endl;
	}

}

int FabanLogic::_dist(int a, int b){
	if (a <= b)
		return b - a;
	return 256 - a + b;
}

EthernetMessage* FabanLogic::_rotateSignatureRight(EthernetMessage *msg) {
	uint32_t sig = msg->fcs;
	if ((sig & 1) != 0) {
		sig = (sig >> 1) + (1 << 31);
	} else {
		sig = (sig >> 1);
	}
	msg->fcs = sig;
//	bpiout << "**RR** " << printMessage(msg) << std::endl;
	return msg;
}

void* FabanLogic::_senderThreadEntryFunction(void *This) {
	((FabanLogic*) This)->_senderThreadTask();
	return NULL;
}

void FabanLogic::_senderThreadTask() {
	bpiout << "SENDER(" << senderNumMsg << ") started..." << std::endl;
	for(uint32_t i = 0; i<senderNumMsg; i++){
//		bpiout << i << std::endl;
		double delay = ((double)rand()/RAND_MAX)*(senderDelayMax-senderDelayMin) + senderDelayMin;
//		std::stringstream ss;
//		ss << delay << std::endl;
//		bpiout << "Sleep for " << ss.str() << std::endl;
		_sendFabanMessage();
		usleep(1000000.0*delay);
	}
	bpiout << "SENDER(" << senderNumMsg << ") finished..." << std::endl;
}

void FabanLogic::_sendFabanMessage() {
	std::string d = "NULL";
//			if (items.size() > 1)
//				d = items[1];

	Message *data = createMessageData(d.c_str(), d.size(), FABDATAMIN);
//	uint64_t creationTime = (getLocalTime() << 16) >> 16;
//	uint64_t test = 0xFFFFFFFFFFFF;
//	std::string str = std::bitset<64>(test).to_string();
//	std::string str2 = std::bitset<64>(creationTime).to_string();
//	bpiout << str << std::endl;
//	bpiout << str2 << std::endl;
	FabanMessage *fabMsg = createFabanMessage(_nodeId, _seqNo++, getLocalTime48(), data);
//	bpiout << printMessage(fabMsg) << std::endl;
	EthernetMessage *em = createEthernetMessage(_bc_mac, source_mac,
	ETH_TYPE_FABAN_H0, serializeFabanMessage(fabMsg));
	//		std::stringstream ss;
	//		ss << "****** " << std::setw(2) << std::hex << std::uppercase << em->fcs << " changed to " << _sign(em)->fcs;
	_sign(em)->fcs;
//			bpiout << "**BC** " << printMessage(em) << " -> "
//					<< printMessage(fabMsg) << " -> " << printMessage(data, true)
//					<< std::endl;

	// TODO Simulate here Ingress Port handling until EndStations are implemented
	uint32_t sig = em->fcs;
	if ((sig & (1 << 31)) != 0) {
		sig = (sig << 1) + 1;
	} else {
		sig = (sig << 1);
	}
	em->fcs = sig;
	//		bpiout << "**RL** " << printMessage(em) << std::endl;
	// END

	for (auto src : _routingtable) {
		if (src.first == _nodeId)
			for (auto portIn : src.second) {
				if (portIn.first == -1)
					for (auto portOut : portIn.second) {
						//							bpiout << "Sending msg via port " << portOut << std::endl;
						// TODO Sending to checking bridges
//								ports[portOut - 1]->addMessage(copyEthernetMessage(em));
						sendMessage(copyEthernetMessage(em), (int) pow(2, portOut - 1));
					}
			}
	}
	deleteMessage(fabMsg);
	deleteMessage(em);
	statTx++;
}
