#include "FabanPort.h"

#include <Protocols/Port.h>
#include <Protocols/EtherTypes.h>

FabanPort::FabanPort(int i, ProtocolLogic* logic) :
		FIPort(i, logic) {
	_dmask = 0;
	_cmask = 0;
}

FabanPort::~FabanPort() {
}

bool FabanPort::_handleIncomingPaket(EthernetMessage* msg) {
	// Manipulation of incoming EthernetMessage
	if (msg->ether_type == ETH_TYPE_FABAN_INIT) {
		setNeighbour(msg->source);
	} else if (msg->ether_type == ETH_TYPE_FABAN_H0) {
		_rotateSignatureLeft(msg);
	} else if (msg->ether_type == ETH_TYPE_FABAN_H1) {
		_rotateSignatureLeft(msg);
	}
	return true;
}

bool FabanPort::_handleOutgoingPaket(EthernetMessage* msg) {
	// Manipulation of outgoing EthernetMessage
	if (msg->ether_type == ETH_TYPE_FABAN_H0) {
		msg->ether_type = ETH_TYPE_FABAN_H1;
		msg->fcs = msg->fcs ^ _dmask;
//		bpiout << "**DM** " << printMessage(msg) << std::endl;
	} else if (msg->ether_type == ETH_TYPE_FABAN_H1) {
		msg->ether_type = ETH_TYPE_FABAN_H2;
		msg->fcs = msg->fcs ^ _cmask;
//		bpiout << "**CM** " << printMessage(msg) << std::endl;
	}
	return true;
}

EthernetMessage* FabanPort::_rotateSignatureLeft(EthernetMessage* msg) {
	uint32_t sig = msg->fcs;
	if ((sig & (1 << 31)) != 0) {
		sig = (sig << 1) + 1;
	} else {
		sig = (sig << 1);
	}
	msg->fcs = sig;
//	bpiout << "**RL** " << printMessage(msg) << std::endl;
	return msg;
}

void FabanPort::setMaskD(uint32_t dmask) {
	_dmask = dmask;
}

void FabanPort::setMaskC(uint32_t cmask) {
	_cmask = cmask;
}
