#include "FaultyContentGeneratorLogic.h"

#include <Protocols/EtherTypes.h>
#include <Protocols/Port.h>
#include <Protocols/MessageTypes.h>
#include <string>
#include <iostream>
#include <unistd.h>
#include <random>

FaultyContentGeneratorLogic::FaultyContentGeneratorLogic() :
		_mu(1), _sigma(0.5), _pfab(0.5), _active(true) {
}

void FaultyContentGeneratorLogic::_executeProtocolLogic() {
}

void FaultyContentGeneratorLogic::_handleControlMessage(std::string id, std::string cmd) {
	this->cmdId = id;
	std::vector<std::string> items = bpiSplitIntoItems(cmd, ' ');
	if (strncmp("sendrandomfabanmessage", cmd.c_str(), 22) == 0) {
		std::string d = "NULL";
		Message *data = createMessageData(d.c_str(), d.size(), FABDATAMIN);
		FabanMessage *fabMsg = createFabanMessage(0, 0, 0xABCDEF, data);
		EthernetMessage *em = createEthernetMessage(_bc_mac, source_mac, ETH_TYPE_FABAN_H0, serializeFabanMessage(fabMsg));

		ports[0]->addMessage(copyEthernetMessage(em));
		ports[1]->addMessage(copyEthernetMessage(em));
		ports[2]->addMessage(copyEthernetMessage(em));
		ports[3]->addMessage(copyEthernetMessage(em));
		bpiout << "Sending on all ports" << std::endl;
		deleteMessage(fabMsg);
		deleteMessage(em);
	} else if (strncmp("setdist", items[0].c_str(), 7) == 0) {
		if (items.size() != 3) {
			bpiout << "Wrong number of arguments. Specify mu and sigma. " << std::endl;
		} else {
			_mu = std::atof(items[1].c_str());
			_sigma = std::atof(items[2].c_str());
			bpiout << "Parameters set to mu=" << _mu << ", sigma=" << _sigma << std::endl;
		}
	} else if (strncmp("setpfab", items[0].c_str(), 7) == 0) {
		if (items.size() != 2) {
			bpiout << "Wrong number of arguments. Specify p_fab. " << std::endl;
		} else {
			_pfab = std::atof(items[1].c_str());
			bpiout << "Parameter set to p_fab=" << _pfab << std::endl;
		}
	} else if (strncmp("start", items[0].c_str(), 5) == 0) {
		_active = true;
		pthread_t thread;
		pthread_create(&thread, NULL, _senderThreadEntryFunction, this);
		pthread_detach(thread);
	} else if (strncmp("stop", items[0].c_str(), 4) == 0) {
		_active = false;
	}
}

FaultyContentGeneratorLogic::~FaultyContentGeneratorLogic() {
}

void* FaultyContentGeneratorLogic::_senderThreadEntryFunction(void *This) {
	((FaultyContentGeneratorLogic*) This)->_senderThreadTask();
	return NULL;
}

void FaultyContentGeneratorLogic::_senderThreadTask() {
	bpiout << "FaultyGenerator started with mu=" << _mu << ", sigma=" << _sigma << ", p_fab=" << _pfab << std::endl;
	std::default_random_engine generator;
	std::normal_distribution<double> distribution(_mu, _sigma);
	while (_active) {
		double sleepfor = distribution(generator);
		_sendFabanMessage();
		usleep(sleepfor * 1000000.0);
	}
}

void FaultyContentGeneratorLogic::_sendFabanMessage() {
	std::string d = "NULL";
	std::string invalidmsg = "INVALIDMESSAGE";

	Message *data = createMessageData(d.c_str(), d.size(), FABDATAMIN);
	FabanMessage *fabMsg = createFabanMessage(0, 0, getLocalTime48(), data);
	EthernetMessage *em;
	if (((double) rand() / RAND_MAX) < _pfab) {
		em = createEthernetMessage(_bc_mac, source_mac, ETH_TYPE_FABAN_H0, serializeFabanMessage(fabMsg));
	} else {
		em = createEthernetMessage(_bc_mac, source_mac, ETH_TYPE_FABAN_H0, createMessageData(invalidmsg.c_str(), invalidmsg.size(), FABDATAMIN));
	}

	ports[0]->addMessage(copyEthernetMessage(em));
	ports[1]->addMessage(copyEthernetMessage(em));
	ports[2]->addMessage(copyEthernetMessage(em));
	ports[3]->addMessage(copyEthernetMessage(em));
	deleteMessage(fabMsg);
	deleteMessage(em);

}
