#include <iostream>
#include <sys/socket.h>
#include <netinet/in.h>
#include <net/if.h>
#include <cstring>
#include <linux/if_packet.h>
#include <linux/if_ether.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <cstdlib>
#include <stdlib.h>
#include <unistd.h>
#include <chrono>

using namespace std;

class Timer
{
public:
    Timer() : beg_(clock_::now()) {}
    void reset() { beg_ = clock_::now(); }
    double elapsed() const { 
        return std::chrono::duration_cast<std::chrono::microseconds>
            (clock_::now() - beg_).count(); }

private:
    typedef std::chrono::high_resolution_clock clock_;
    typedef std::chrono::duration<double, std::ratio<1> > second_;
    std::chrono::time_point<clock_> beg_;
};

int main(int argc, char** argv) {

	string dev_name = "eth0.1";
	int len = 125;
	int n = 1;
	double speed = 100;
	bool wait = true;

	if(argc % 2 == 0) {
		cout << "Invalid parameters" << endl;
		exit(-1);
	}

	for(int i = 1; i<argc; i=i+2){
		if(argv[i][0]!='-') {
			cout << "Invalid parameter: " << argv[i] << endl;
			exit(-1);
		}
		if(strncmp(argv[i],"-dev",4)==0){
			dev_name = argv[i+1];
		}
		if(strncmp(argv[i],"-n",2)==0){
			n = atoi(argv[i+1]);
		}
		if(strncmp(argv[i],"-l",2)==0){
			len = atoi(argv[i+1]);
		}
		if(strncmp(argv[i],"-N",2)==0){
			wait = false;
		}
		if(strncmp(argv[i],"-s",2)==0){
			speed = stod(argv[i+1]);
		}

	}

	cout << "RAW Socket on interface " << dev_name << " created. Press to continue..." << endl;
	if(wait==true){	
		cin.ignore();
	}
	int raw_port = -1;
	struct sockaddr_ll raw_port_address;

	if ((raw_port = socket(AF_PACKET, SOCK_RAW, IPPROTO_RAW)) == -1) {
		cerr << "Error creating raw socket" << endl;
		return -1;
	}
	struct ifreq if_idx;
	memset(&if_idx, 0, sizeof(struct ifreq));
	strncpy(if_idx.ifr_name, dev_name.c_str(), IFNAMSIZ - 1);
	if (ioctl(raw_port, SIOCGIFINDEX, &if_idx) < 0) {
		cerr << "Error getting device index (egress thread)" << endl;
		return -1;
	}
	raw_port_address.sll_ifindex = if_idx.ifr_ifindex;
	raw_port_address.sll_halen = ETH_ALEN;
	
	char data[len];
	for (int i = 0; i < len; i++) {
		if (i < 6)
			data[i] = 0xFF;
		else if (i < 12)
			data[i] = 0x00;
		else if (i < 14)
			data[i] = 0xAA;
		else
			data[i] = 'A';
	}

	long long int i = 0;

        //Timer compl;
        //Timer t;
        Timer t2;
        t2.reset();
        double elms = 0;  
        for (i = 0; i < n; i++) {
                //t.reset();
                int s = sendto(raw_port, data, len, 0, (struct sockaddr*) &raw_port_address, sizeof(struct sockaddr_ll));
                //cout << ((i+1)*8.0*len)/speed << "\tDELTA=" << elms << endl;
                if(speed>0){
                        elms = ((i+1)*8.0*len)/speed - t2.elapsed();
                        if(elms>0)
                                usleep(elms);
                }
        }

        double endtime = t2.elapsed();

	cout << "Msg Sent   = " << i << endl;
	cout << "Msg Size   = " << len << " bytes"<<endl;
	cout << "Duration   = " << endtime << " µs"<< endl;
	cout << "Total      = " << i*len*8 << " bits" << endl; 
	cout << "Speed      = " << (n*8*len/endtime) << " Mbit/s" << endl;
	
	return 0;
}

