package topology;

public class FabanWaveTool {

	FabanDescriptionFileHandler fd;

	public FabanWaveTool() {
	}

	public FabanWaveTool(Topology network) {
		fd = new FabanDescriptionFileHandler(network);
	}
	
	public FabanWaveTool(Topology network, String selectionMode) {
		fd = new FabanDescriptionFileHandler(network);
		fd.setSelectionMode(selectionMode);
	}

	public boolean loadJsonFile(String inputpath) {
		fd = new FabanDescriptionFileHandler(inputpath);
		if (fd.isSucessfullyLoaded())
			if (fd.readAndValidateTopology())
				return true;
			else {
				fd = null;
				return false;
			}
		else {
			fd = null;
			return false;
		}
	}

	public boolean parseTopology() {
		if (fd != null) {
			return fd.createNetworkFromTopologyDescription();
		}
		return false;
	}

	public boolean parseFabanConfiguration() {
		if (fd != null) {
			return fd.readFabanConfiguration();
		} else
			return false;
	}

	public boolean validateFabanConfiguration() {
		if (fd != null) {
			return fd.validateFabanConfiguration();
		} else {
			return false;
		}
	}

	public boolean wavesExists() {
		if (fd != null) {
			return fd.generateWavesSilent();
		} else {
			return false;
		}
	}

	public boolean generateWaves() {
		if (fd != null) {
			if (fd.generateWaves()) {
				fd.createJsonWaveObject();
				return true;
			} else {
				return false;
			}
		} else {
			return false;
		}
	}

	public boolean saveJsonFile(String outputfile) {
		if (fd != null) {
			fd.save(outputfile);
			return true;
		} else
			return false;
	}

	public Topology getNetwork() {
		if (fd != null) {
			return fd.getNetwork();
		} else {
			return null;
		}
	}

}
