package topology;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
public class Topology {

	private ArrayList<Bridge> bridges = new ArrayList<Bridge>();
	private ArrayList<EndStation> endstations = new ArrayList<EndStation>();
	private ArrayList<UndirectedLink> links = new ArrayList<UndirectedLink>();

	private Map<Bridge, FabanMetaInformation> waves = new HashMap<>();

	public Topology() {

	}

	public ArrayList<Bridge> getBridges() {
		return bridges;
	}

	public ArrayList<EndStation> getEndstations() {
		return endstations;
	}

	public ArrayList<UndirectedLink> getLinks() {
		return links;
	}

	public ArrayList<UndirectedLink> getBridgeLinks() {
		ArrayList<UndirectedLink> linkset = new ArrayList<>();
		for (UndirectedLink l : links) {
			if (l.a instanceof Bridge && l.b instanceof Bridge)
				linkset.add(l);
		}
		return linkset;
	}

	public Map<Bridge, FabanMetaInformation> getWaves() {
		return waves;
	}

	public Bridge getBridge(String name) {
		for (Bridge b : bridges) {
			if (b.name.equals(name))
				return b;
		}
		return null;
	}

	public boolean addBridge(String name) {
		bridges.add(new Bridge(name));
		return true;
	}

	public boolean addEndStation(String name) {
		endstations.add(new EndStation(name));
		return true;
	}

	public boolean addLink(String a, String b) {
		Component cA = getComponent(a);
		Component cB = getComponent(b);
		if (cA == null || cB == null || (cA instanceof EndStation && cB instanceof EndStation)) {
			return false;
		}
		links.add(new UndirectedLink(cA, cB));
		return true;
	}

	public boolean addFabanMetaInformation(Bridge d, FabanMetaInformation fmi) {
		if (waves.containsKey(d)) {
			return false;
		} else {
			waves.put(d, fmi);
			return true;
		}
	}

	public boolean hasWave(Bridge d) {
		return waves.containsKey(d);
	}

	public String printWaveInfo() {
		return ("Number of Waves: " + waves.size());
	}

	public boolean isConnected() {
		ArrayList<Bridge> conBridges = new ArrayList<>();
		Component temp;
		if (bridges.isEmpty())
			return false;
		else {
			conBridges.add(bridges.get(0));
			for (int i = 0; i < bridges.size(); i++) {
				Bridge b = bridges.get(i);
				for (Link l : links) {
					if ((temp = l.getConnectedComponent(b)) != null) {
						if (temp instanceof Bridge && !conBridges.contains(temp))
							conBridges.add((Bridge) temp);
					}
				}
			}
		}
		if (conBridges.size() == bridges.size())
			return true;
		else
			return false;
	}

	public ArrayList<Bridge> getNeighbours(Bridge b) {
		ArrayList<Bridge> neighbours = new ArrayList<>();
		Component temp;
		for (Link l : links) {
			if ((temp = l.getConnectedComponent(b)) != null) {
				if (temp instanceof Bridge && !neighbours.contains(temp))
					neighbours.add((Bridge) temp);
			}
		}
		return neighbours;
	}

	private Component getComponent(String name) {
		for (Component c : endstations)
			if (c.name.equals(name))
				return c;
		for (Component c : bridges)
			if (c.name.equals(name))
				return c;
		return null;
	}
}
